"
 COPYRIGHT (c) 1991 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#Image
	instanceVariableNames:'bytes width height bitsPerPixel depth colorMap maxPixelValue
		rowByteSize bitsPerSample samplesPerPixel photometric device
		deviceForm monoDeviceForm fullColorDeviceForm mask
		maskedPixelsAre0 fileName imageSequence'
	classVariableNames:'Lobby DitherAlgorithm NumberOfDitherColors
		CollectGarbageWhenRunningOutOfColors ImageNotFoundQuerySignal
		BadImageFormatQuerySignal ImageSaveErrorSignal
		FileCreationErrorSignal CannotRepresentImageSignal
		InformationLostQuerySignal'
	poolDictionaries:''
	category:'Graphics-Images'
!

!Image class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1991 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    this class provides (some time in the future) representation
    for all kinds of images (monochrome, greyscale and color)
    and will finally replace Form. 
    For now (only ;-) depths of 1, 2, 4, 8, 16 and 24 are supported.

    An Image keeps all of its information in a device independent way, 
    but may be associated to a device. 
    The data held keeps all information which was originally present, 
    even if the display-device has lower resolution.
    Therefore, it is possible to process and manipulate images without loosing 
    color information.

    Images may be created manually (by providing a pixel array),
    by screen capture or by reading a file (using an ImageReader).
    This gives a device independent image.
    For display, a device representation is required, which can be
    aquired by sending the 
        'onDevice:aDevice' 
    message to the image. This will create a (possibly dithered) device-form, 
    representing the image using the currently available colors.

    In rare cases, an explicit monochrome representation of the image is needed 
    (X servers take monochrome icons only), this can be created by sending
    it the message 
        'monochromeOn:aDevice'.

    Also, it is planned to generate another hi-color resolution version,
    which uses its own colormap and allows the use of all 256 colors on an 8bit display 
    (this is not currently implemented).

    An images pixel interpretation is controlled by the photometric instance variable
    and (if that is #palette) a colorMap.
    This is in contrast to the ST-80 way, where all of this info is in the colorMap
    (and using specialized colorMaps for b&w / greyScale images).
    This may change in future versions for more application compatibility.

    To convert pictures from/to external file-formats, image readers are used
    which have the file-format knowledge built in. 
    There are readers for most common formats available 
    (see ImageReader and especially subclasses such as TIFFReader, GIFReader etc.).

    File formats are handled by subclasses of ImageReader, which understand
    a specific format. You can add more readers, by adding an association
    such as ('.jpg' -> JPEGReader) to the class variable 'FileFormats' 
    (see the classes #initialize method, which sets up some default, and the
    patches/display.rc files, which add more).

    Some algorithms used here (especially dithering & color allocation) are
    experimental and far from being perfect (some are very slow). 
    For now, the most common cases have been optimized and perform reasonably
    fast - however, with uncommon depth/visualType combinations, you may
    run into slower fallback methods ...

    Much more work is needed if you plan to do heavy image processing and will
    (may) be done in the future if there is a demand for it ...
    Dithering can be controlled by the DitherAlgorithm classVariable:

       DitherAlgorithm:

       nil                  a simple threshold algorithm
                            (i.e. for mono, p<0.5 -> black, p>=0.5 -> white)

       #pattern             patterned dither
                            (for p, take dithered color to fill pixel;
                             uses dithering in color-class)

       #error               error diffusion dither (Floyd-Steinberg)
                            planned - not yet implemented.



    [instance variables:]

        width               <Integer>       the width in pixels
        height              <Integer>       the height in pixels
        bytes               <ByteArray>     the full image information
        photometric         <Symbol>        #rgb, #palette, #blackIs0 or #whiteIs0
        samplesPerPixel     <Integer>       the number of planes
        bitsPerSample       <Array>         the number of bits per plane

        colorMap            <Array>         only if photometric is #palette;
                                            maps pixel values to r/g/b values.

        device              <Workstation>   the device on which deviceForm,
                                            monoDeviceForm and lowResDeviceForm are

        deviceForm          <Form>          the device form which gives the best
                                            possible aproximation of the image on
                                            device using standard colors.

        monoDeviceForm      <Form>          the device form which gives a monochrome
                                            aproximation of the image on device.

        fullColorDeviceForm <Form>          the device form which gives the best
                                            possible aproximation of the image on
                                            device using private colors.
                                            (not yet implemented)

        mask                <ImageMask>     an optional mask;
                                            if non-nil, only pixels for which the
                                            corresponding mask bit is non-zero
                                            are drawn.

        maskedPixelsAre0    <Boolean>       a hint for image processors and drawers
                                            if true, masked pixels are known to be
                                            zero in the pixel bytes.

        fileName            <String>        the name of the file from which the
                                            image was loaded - nil otherwise.
                                            Useful for image save functions
                                            and for the UIPainter utility.

        imageSequence                       the imageSequence, of which the
                                            instance is a frame or nil, 
                                            if its not part of a sequence.

        bitsPerPixel                        not used in ST/X
        depth                               - these have been added in instVar-slots
        maxPixelValue                       - according to the ST-80's image class.
        rowByteSize                         - to allow loading of st-80 images
                                            - (which are stored as instVarAt:put: expressions)

    [class variables:]

        Lobby               <Registry>      keeps track of known images
                                            (for resource freeing with garbage collector)

        DitherAlgorithm     <Symbol>        defines how to dither

        NumberOfDitherColors <Integer>      defines, how many dither colors to use

        FileFormats         <Dictionary>    associates filename extensions to
                                            image reader classes (now set-up in startup-file)

        CollectGarbageWhenRunningOutOfColors
                            <Boolean>       if true, and we run out of available
                                            device colors during creation of a
                                            device image, collect garbage for
                                            possible image reclamation.
                                            If false, proceed immediately.
                                            Default is true.

        ImageNotFoundQuerySignal
                            <QuerySignal>   raised, if an image could not be loaded
                                            from a file. The parameter is the images
                                            fileName. 
                                            A handler may return a replacement
                                            image or proceed with nil.
                                            If unhandled, a nil is returned from the
                                            image creation.

        BadImageFormatQuerySignal
                            <QuerySignal>   raised, if an image could not be loaded
                                            from a file due to a file error or
                                            unsupported format. 
                                            A handler may return a replacement
                                            image or proceed with nil.
                                            If unhandled, a nil is returned from the
                                            image creation.

        ImageSaveErrorSignal
                            <Signal>        parent of errors below.

        FileCreationErrorSignal
                            <Signal>        file could not be created when saving an
                                            image.

        CannotRepresentImageSignal
                            <Signal>        the specified ImageReader cannot represent
                                            the given image.

        InformationLostQuerySignal
                            <Signal>        the specified ImageReader can represent
                                            the given image, but some information
                                            (typically color resolution) is lost.
                                            If unhandled, the save proceeds as usual.


    todo:
        support alpha masks
        cleanup the dithering & conversion code

    [See also:]
        Form Icon ImageReader

    [author:]
        Claus Gittinger
"

!

examples
"
    reading from a file (many formats are supported):
                                                                        [exBegin]
        (Image fromFile:'bitmaps/gifImages/claus.gif') inspect
                                                                        [exEnd]
                                                                        [exBegin]
        (Image fromFile:'bitmaps/SmalltalkX.xbm') inspect
                                                                        [exEnd]
                                                                        [exBegin]
        (Image fromFile:'bitmaps/winBitmaps/okSmily_up.bmp') inspect
                                                                        [exEnd]

    The following examples demonstrate various depth
    and colorMap variations ...

    inline image:
      default: depth=1 & #blackIs0
                                                                        [exBegin]
        (Image width:8 height:8
               fromArray:#( 2r11111111
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r11111111 )) inspect
                                                                        [exEnd]

      with #whiteIs0 photometric
                                                                        [exBegin]
        ((Image width:8 height:8
               fromArray:#( 2r11111111
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r11111111 ))
            photometric:#whiteIs0)
                 inspect
                                                                        [exEnd]

      with a colorMap
                                                                        [exBegin]
        ((Image width:8 height:8
               fromArray:#( 2r11111111
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r10000001
                            2r11111111 ))
            colorMap:(Array with:(Color red)
                            with:(Color yellow)))
                 inspect
                                                                        [exEnd]

      a depth4 greyScale image:
      (default photometric is #blackIs0)
                                                                        [exBegin]
        (Depth4Image
             width:8 
             height:4
             fromArray:#[ 
                            16r00 16r11 16r22 16r33
                            16r44 16r55 16r66 16r77
                            16r88 16r99 16raa 16rbb
                            16rcc 16rdd 16ree 16rff 
                        ]) inspect
                                                                        [exEnd]
        ((Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ])
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]
      the following has the same effect:
                                                                        [exBegin]
        ((Image
             width:4 
             height:4
             depth:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ])
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]
      with reverse grey-interpretation:
                                                                        [exBegin]
        ((Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ])
            photometric:#whiteIs0;
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]

      with 1-bit-per-pixel rgb interpretation:
                                                                        [exBegin]
        ((Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ])
            photometric:#rgb;
            samplesPerPixel:3;
            bitsPerSample:#(1 1 1);
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]

      with 1/2/1 rgb interpretation:
                                                                        [exBegin]
        ((Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ])
            photometric:#rgb;
            samplesPerPixel:3;
            bitsPerSample:#(1 2 1);
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]

      a 2/2/0 rgb image (i.e. no blue):
                                                                        [exBegin]
         |i|

         i := Depth4Image
                    width:4
                    height:4
                    fromArray:#[ 16r01 16r23
                                 16r45 16r67
                                 16r89 16rab
                                 16rcd 16ref ].
         i photometric:#rgb.
         i samplesPerPixel:3.
         i bitsPerSample:#(2 2 0).

         i := i magnifiedBy:30.
         i inspect.
                                                                        [exEnd]


      a 0/0/4 rgb image (i.e. no red or green):
                                                                        [exBegin]
         |i|

         i := Depth4Image
                    width:4
                    height:4
                    fromArray:#[ 16r01 16r23
                                 16r45 16r67
                                 16r89 16rab
                                 16rcd 16ref ].
         i photometric:#rgb.
         i samplesPerPixel:3.
         i bitsPerSample:#(0 0 4).

         i := i magnifiedBy:30.
         i inspect.
                                                                        [exEnd]


      a 2plane greyscale image:
                                                                        [exBegin]
        ((Depth2Image
             width:4 
             height:4
             fromArray:#[ 
                            4r0123
                            4r1230
                            4r2301
                            4r3012 
                        ])
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]

      with colors:
                                                                        [exBegin]
        ((Depth2Image
             width:4 
             height:4
             fromArray:#[ 
                            4r0123
                            4r1230
                            4r2301
                            4r3012 
                        ])
            colorMap:(Array with:(Color black)
                            with:(Color red)
                            with:(Color green)
                            with:(Color blue));
            magnifiedBy:30)
                 inspect
                                                                        [exEnd]

      depth8 image with 3/3/2 rgb interpretation:
                                                                        [exBegin]
        ((Depth8Image
             width:16 
             height:16
             fromArray:(ByteArray withAll:(0 to:16rFF)))
            photometric:#rgb;
            samplesPerPixel:3;
            bitsPerSample:#(3 3 2);
            magnifiedBy:10)
                 inspect
                                                                        [exEnd]

      depth8 image with 2/2/2 rgb interpretation:
                                                                        [exBegin]
        ((Depth8Image
             width:8 
             height:8
             fromArray:(ByteArray withAll:(0 to:16r3F)))
            photometric:#rgb;
            samplesPerPixel:3;
            bitsPerSample:#(2 2 2);
            magnifiedBy:10)
                 inspect
                                                                        [exEnd]

    storing - only a few formats (TIFF, XBM, XPM) currently support storing:
                                                                        [exBegin]
        |img|

        img := Image fromFile:'bitmaps/okSmily_up.bmp'.
        img saveOn:'myImage.tiff'.
        (Image fromFile:'myImage.tiff') inspect
                                                                        [exEnd]

    magnifying (any factor):
                                                                        [exBegin]
        ((Image fromFile:'bitmaps/gifImages/claus.gif') 
            magnifiedTo:(48@48))
                inspect
                                                                        [exEnd]
                                                                        [exBegin]
        ((Image fromFile:'bitmaps/gifImages/claus.gif') 
            magnifiedBy:0.7)
                inspect
                                                                        [exEnd]

    rotating:
                                                                        [exBegin]
        ((Image fromFile:'bitmaps/gifImages/claus.gif') 
            rotated:90)
                inspect
                                                                        [exEnd]
                                                                        [exBegin]
        (((Image fromFile:'bitmaps/gifImages/claus.gif') 
            magnifiedBy:0.3@0.7) rotated:270)
                inspect
                                                                        [exEnd]
                                                                        [exBegin]
        (((Image fromFile:'bitmaps/gifImages/claus.gif') 
            ) rotated:30)
                inspect
                                                                        [exEnd]
    negative:
                                                                        [exBegin]
        ((Image fromFile:'bitmaps/gifImages/claus.gif') 
            negative)
                inspect
                                                                        [exEnd]
"
! !

!Image class methodsFor:'initialization'!

addReader:aReaderClass suffix:aSuffix
    "register an additional image reader.
     This is provided for subclasses, to regster themself when
     loaded (or from the startup scripts)"

    self obsoleteMethodWarning.
    self addReader:aReaderClass suffix:aSuffix mimeType:nil
!

addReader:aReaderClass suffix:aSuffix mimeType:mimeType
    "register an additional image reader.
     This interface is kept for backward compatibility.
     The knowledge has been concentrated in MIMETypes"

    self obsoleteMethodWarning.
    MIMETypes
        defineImageType:mimeType suffix:aSuffix reader:aReaderClass

    "
     Image addReader:GIFReader suffix:'gif'
     Image addReader:GIFReader suffix:'gif' mimeType:'image/gif'
    "

    "Modified: 30.6.1997 / 21:59:59 / cg"
!

fileFormats
    "return the collection of supported file formats.
     The returned dictionary maps file-extensions to image reader classes."

    ^ MIMETypes fileSuffixToImageReaderMapping

    "
     Image fileFormats
    "

    "Modified: 30.6.1997 / 22:05:58 / cg"
!

flushDeviceImages
    "simply release all deviceForms"

    Lobby do:[:anImage |
        anImage restored
    ]

    "Modified: 15.6.1996 / 15:45:02 / cg"
!

initialize
    "initialize class constants"

    "setup tracker of known pictures"
    Lobby isNil ifTrue:[
        Lobby := Registry new.
        ObjectMemory addDependent:self.

        "/ define the algorithm to use for dithering - 
        "/ supported values are:
        "/      #threshold
        "/      #ordered
        "/      #floydSteinberg
        "/      #burkes

        DitherAlgorithm := #floydSteinberg.   

        (Display notNil and:[Display hasGrayscales]) ifFalse:[
            NumberOfDitherColors := 64
        ] ifTrue:[
            "as far as I remember, this is about the number of grey values, the eye can distinguish"
            NumberOfDitherColors := 100
        ].

        "define reader classes"
        self initializeFileFormatTable.
        self initializeMIMETable.

        CollectGarbageWhenRunningOutOfColors := false.

        ImageNotFoundQuerySignal := QuerySignal new.
        ImageNotFoundQuerySignal nameClass:self message:#imageNotFoundQuerySignal.

        ImageSaveErrorSignal := ErrorSignal newSignalMayProceed:true.
        ImageSaveErrorSignal nameClass:self message:#imageSaveErrorSignal.

        FileCreationErrorSignal := ImageSaveErrorSignal newSignalMayProceed:true.
        FileCreationErrorSignal nameClass:self message:#fileCreationErrorSignal.

        CannotRepresentImageSignal := ImageSaveErrorSignal newSignalMayProceed:true.
        CannotRepresentImageSignal nameClass:self message:#cannotRepresentImageSignal.

        InformationLostQuerySignal := QuerySignal new.
        InformationLostQuerySignal nameClass:self message:#informationLostQuerySignal.

        BadImageFormatQuerySignal := QuerySignal new.
        BadImageFormatQuerySignal nameClass:self message:#badImageFormatQuerySignal.
    ]

    "Modified: 27.6.1997 / 16:44:43 / cg"
!

initializeFileFormatTable
    "initialize a default table to map from file extension to reader class.
     The mapping here is a default needed for proper operation of ST/X;
     see the 'smalltalk.rc'/'display.rc' startup files for a real (full) map."

    MIMETypes notNil ifTrue:[
        MIMETypes autoload.
        MIMETypes imageReaderForSuffix:'xbm'  put:XBMReader.
        MIMETypes imageReaderForSuffix:'tiff' put:TIFFReader.
        MIMETypes imageReaderForSuffix:'gif'  put:GIFReader.
    ].

    "
     Image initializeFileFormatTable
    "

    "Modified: 30.6.1997 / 22:07:28 / cg"
!

initializeMIMETable
    "initialize a default table to map from mime type to reader class.
     The mapping here is a default needed for proper operation of ST/X;
     see the 'smalltalk.rc'/'display.rc' startup files for a real (full) map."

    MIMETypes notNil ifTrue:[
        MIMETypes autoload.
        MIMETypes mimeTypeForSuffix:'gif'       put:'image/gif'.
        MIMETypes mimeTypeForSuffix:'tiff'      put:'image/tiff'.
        MIMETypes mimeTypeForSuffix:'tif'       put:'image/tiff'.
        MIMETypes mimeTypeForSuffix:'xbm'       put:'image/x-xbitmap'.
        MIMETypes mimeTypeForSuffix:'xpm'       put:'image/x-xpixmap'.
    ].

    "
     Image initializeMIMETable
    "

    "Created: 27.6.1997 / 16:43:48 / cg"
    "Modified: 21.7.1997 / 17:33:59 / cg"
!

update:something with:aParameter from:changedObject
    "flush all device specific stuff when restarted from a snapshot"

    (something == #restarted) ifTrue:[
        self flushDeviceImages
    ]

    "Created: 21.6.1996 / 19:47:43 / cg"
! !

!Image class methodsFor:'instance creation'!

extent:ext depth:d bits:bits
    "ST-80 compatibility; assume 32-bit padding"

    ^ self extent:ext depth:d bits:bits pad:32
!

extent:ext depth:d bits:bits pad:padding
    "ST-80 compatibility"

    ^ self width:ext x height:ext y depth:d fromArray:bits pad:padding
!

extent:ext depth:d bitsPerPixel:bpp palette:aColormap usingBits:bits
    "ST-80 compatibility"

    |img|

    img := self width:ext x height:ext y depth:bpp fromArray:bits pad:32.
    img colorMap:aColormap.
    ^ img

    "Created: 25.1.1997 / 03:50:22 / cg"
    "Modified: 25.1.1997 / 12:27:35 / cg"
!

extent:ext depth:d palette:aColormap
    "create & return a blank image of the given size.
     ST-80 compatibility"

    |newImage emptyBits|

    newImage := (self implementorForDepth:d) new.
    newImage width:ext x height:ext y depth:d palette:aColormap.
    emptyBits := ByteArray new:(newImage bytesPerRow * ext y).
    newImage bits:emptyBits.
    ^ newImage

    "
     Image extent:16@16 depth:8 palette:nil
     Image extent:16@16 depth:4 palette:nil
     Image extent:16@16 depth:2 palette:nil
    "

    "Created: 6.3.1997 / 15:24:01 / cg"
!

extent:ext depth:d palette:aColormap bits:bits
    "ST-80 compatibility"

    ^ self extent:ext depth:d palette:aColormap bits:bits pad:16

    "Modified: 7.10.1996 / 11:32:16 / cg"
!

extent:ext depth:d palette:aColormap bits:bits pad:padding
    "ST-80 compatibility"

    |img|

    img := self width:ext x height:ext y depth:d fromArray:bits pad:padding.
    img colorMap:aColormap.
    ^ img

    "Modified: 7.10.1996 / 11:32:00 / cg"
!

extent:ext fromArray:bits offset:offset
    "ST-80 compatibility"

    ^ self width:ext x height:ext y fromArray:bits
!

fromForm:aForm
    "create & return an Image given a aForm"

    |depth device vis img|

    depth := aForm depth.
    device := aForm device.

    (device notNil and:[depth == device depth]) ifTrue:[
        "/
        "/ for truecolor displays, return a Depth24Image
        "/ (must do this for depth15 & depth16 displays, since
        "/  Depth16Image has no way to specify r/g/b masks ...)
        "/
        vis := device visualType.
        (vis == #TrueColor or:[vis == #DirectColor]) ifTrue:[
            depth > 8 ifTrue:[
                depth := 24.
            ]
        ].
    ].
    img := (self implementorForDepth:depth) new.
    device isNil ifTrue:[
        ^ img fromForm:aForm.
    ].
    ^ img from:aForm in:(0@0 extent:(aForm extent))

    "
     |f|

     f := Form width:16 height:16.
     f clear.
     f displayLineFromX:0 y:0 toX:15 y:15.
     f inspect.
     (Image fromForm:f) inspect
    "

    "
     |f|

     f := Form width:16 height:16 depth:(Display depth) on:Display.
     f clear.
     f paint:(Color red).
     f displayLineFromX:0 y:0 toX:15 y:15.
     f paint:(Color green).
     f displayLineFromX:15 y:0 toX:0 y:15.
     f inspect.
     (Image fromForm:f) inspect
    "

    "Modified: 11.7.1996 / 11:21:42 / cg"
!

fromImage:anImage
    "create & return an Image given another image. This can be used to
     convert an image to another depth."

    (self == Image or:[anImage class == self]) ifTrue:[^ anImage].
    ^ self new fromImage:anImage.

    "
     |i1 i8|

     i1 := Image fromFile:'bitmaps/SBrowser.xbm'.
     i8 := Depth8Image fromImage:i1.
     i8 inspect
    "
!

fromSubImage:anImage in:aRectangle
    "create & return an Image from a rectangular area in another image. 
     This can also be used to get a subimage in another depth."

    |newImage|

    self == Image ifTrue:[
        newImage := (self implementorForDepth:anImage depth) new.
    ] ifFalse:[
        newImage := self new.
    ].
    ^ newImage fromSubImage:anImage in:aRectangle.

    "
     |i1 i8|

     i1 := Image fromFile:'bitmaps/garfield.gif'.
     i8 := Depth8Image fromSubImage:i1 in:(0@0 corner:20@20).
     i8 inspect
    "
    "
     |i1 i8|

     i1 := Image fromFile:'bitmaps/claus.gif'.
     i8 := Depth8Image fromSubImage:i1 in:(70@50 extent:50@50).
     i8 inspect
    "
    "
     |i1 i8|

     i1 := Image fromFile:'bitmaps/claus.gif'.
     i8 := Image fromSubImage:i1 in:(70@50 extent:50@50).
     i8 inspect
    "

    "Created: 20.9.1995 / 01:05:43 / claus"
    "Modified: 24.4.1997 / 23:13:02 / cg"
!

new
    "create a new image. Redefined to set the photometric to
     greyScale with black being 0 as default."

    ^ super new photometric:(self defaultPhotometric)

    "Modified: 10.6.1996 / 18:08:37 / cg"
!

width:w height:h
    "create a new image, given width, height. Assume a depth of 1."

    |cls|

    cls := self.
    cls == Image ifTrue:[
        cls := self implementorForDepth:1
    ].
    ^ cls new width:w height:h depth:1 
!

width:w height:h depth:d
    "create a new image, given width, height and depth"

    ^ (self implementorForDepth:d) new
        width:w height:h depth:d
!

width:w height:h depth:d fromArray:pixelData
    "create a new image, given width, height, depth and data.
     Data must be a ByteArray containing correctly aligned bits for the specified
     depth (8-bit padded)."

    ^ (self implementorForDepth:d) new 
                width:w height:h depth:d fromArray:pixelData

    "
     Image width:8 
           height:8
           depth:1
           fromArray:#[2r11001100
                       2r00110011
                       2r11001100
                       2r00110011
                       2r11001100
                       2r00110011
                       2r11001100
                       2r00110011].
    "

    "
     Image width:8 
           height:8
           depth:2 
           fromArray:#[4r1100 4r1100
                       4r0011 4r0011
                       4r1100 4r1100
                       4r0011 4r0011
                       4r1100 4r1100
                       4r0011 4r0011
                       4r1100 4r1100
                       4r0011 4r0011].
    "

    "
     Image width:8 
           height:8
           depth:4 
           fromArray:#[16r00 16r01 16rf0 16rf1
                       16r02 16r03 16rf2 16rf3
                       16r04 16r05 16rf4 16rf5
                       16r06 16r07 16rf6 16rf7
                       16r08 16r09 16rf8 16rf9
                       16r0a 16r0b 16rfa 16rfb
                       16r0c 16r0d 16rfc 16rfd
                       16r0e 16r0f 16rfe 16rff].
    "

    "Modified: 10.6.1996 / 18:18:10 / cg"
!

width:w height:h depth:d fromArray:pixelData pad:padding
    "create a new image, given width, height, depth and data.
     Data must be a ByteArray containing correctly aligned bits for the specified
     depth."

    |img newBits srcRowBytes dstRowBytes srcIndex dstIndex|

    img := (self implementorForDepth:d) new width:w height:h depth:d .

    padding ~~ 8 ifTrue:[
        "must repad; ST/X uses byte padding, while ST-80 uses longword
         padding. This is stupid, and may be changed in ST/X with future versions.
        "
        dstRowBytes := img bytesPerRow.
        srcRowBytes := ((w * d + padding - 1) bitShift:-5) bitShift:2.

        newBits := ByteArray new:(dstRowBytes * h).
        srcIndex := 1.
        dstIndex := 1.

        1 to:h do:[:row |
            newBits replaceFrom:dstIndex 
                             to:(dstIndex + dstRowBytes - 1)
                           with:pixelData
                     startingAt:srcIndex.
            srcIndex := srcIndex + srcRowBytes.
            dstIndex := dstIndex + dstRowBytes.
        ].
    ] ifFalse:[
        pixelData class isBytes ifFalse:[
            newBits := ByteArray withAll:pixelData
        ] ifTrue:[
            newBits := pixelData
        ]
    ].
    img bits:newBits.
    ^ img

    "Modified: 8.6.1996 / 10:07:47 / cg"
!

width:w height:h fromArray:anArray
    "create a new image, given width, height. Assume a depth of 1 of the
     receiving class is Image.
     Data must be a ByteArray containing correctly aligned bits for depth 1
     (i.e. 8 bits per byte)."

    |cls d pixels|

    cls := self.
    cls == Image ifTrue:[
        cls := self implementorForDepth:1.
        d := 1.
    ] ifFalse:[
        d := cls imageDepth
    ].
    anArray class isBytes ifFalse:[
        pixels := ByteArray withAll:anArray
    ] ifTrue:[
        pixels := anArray
    ].
    ^ cls new width:w height:h depth:d fromArray:pixels

    "
     Image width:8 
           height:8 
           fromArray:#[2r11001100
                       2r00110011
                       2r11001100
                       2r00110011
                       2r11001100
                       2r00110011
                       2r11001100
                       2r00110011].
    "

    "Modified: 8.6.1996 / 10:07:26 / cg"
! !

!Image class methodsFor:'Signal constants'!

badImageFormatQuerySignal
    "return the (query-) signal, which is raised if some
     bitmap-image could not be loaded due to an unrecognized format.
     If unhandled, the image-load returns nil.
     Otherwise, it returns whatever the handler proceeds with.
     The exception gets either the images fileName or an input stream
     as parameter"

    ^ BadImageFormatQuerySignal

    "Created: 1.2.1997 / 14:40:29 / cg"
!

cannotRepresentImageSignal
    "return the signal, which is raised if some
     bitmap-image could not be saved in the requested format.
     This happens for example, if a true color image is saved in
     xpm format or a color image saved in xbm format.
     Applications should either ask the class before saving
     (via #canRepresent:) or handle the error and use an alternative
     format then."

    ^ CannotRepresentImageSignal

    "Created: 27.2.1997 / 12:24:43 / cg"
!

fileCreationErrorSignal
    "return the signal which is raised if a file could not be
     created in an image-save operation."

    ^ FileCreationErrorSignal

    "Created: 27.2.1997 / 12:32:59 / cg"
!

imageNotFoundQuerySignal
    "return the (query-) signal, which is raised if some
     bitmap-image could not be loaded from a file.
     If unhandled, the image-load returns nil.
     Otherwise, it returns whatever the handler proceeds with.
     The exception gets the images fileName as parameter"

    ^ ImageNotFoundQuerySignal

    "Created: 7.1.1997 / 16:04:49 / cg"
!

imageSaveErrorSignal
    "return the parent signal of all image-save errors."

    ^ ImageSaveErrorSignal

    "Created: 27.2.1997 / 12:32:23 / cg"
!

informationLostQuerySignal
    "return the (query-) signal, which is raised if some
     bitmap-image is saved in a format which cannot represent
     the full image (for example: the images mask).
     If unhandled, the image-save proceeds.
     Otherwise, the handler may show a warn box or whatever and decide
     to proceed or abort the saving."

    ^ InformationLostQuerySignal

    "Created: 27.2.1997 / 12:43:50 / cg"
! !

!Image class methodsFor:'cleanup'!

releaseResourcesOnDevice:aDevice
    "this is sent when a display connection is closed,
     to release all cached Images from that device"

    Lobby 
        unregisterAllForWhich:[:anImage | 
                |ok|
                ok := anImage graphicsDevice == aDevice.
                ok ifTrue:[
                    anImage quickRelease
                ].
                ok
        ]

    "Created: 16.1.1997 / 19:30:44 / cg"
    "Modified: 16.1.1997 / 19:33:49 / cg"
! !

!Image class methodsFor:'file reading'!

fromFile:aFileName
    "read an image from a file - this methods tries to find
     out the file format itself (by the extension and by contents)
     and lets the appropriate reader read the file.
     Also, all bitmap directories along the searchPath are checked
     for the file; thus, you can place your private images into a
     separate directory, thereby overriding system bitmaps easily.
     If the file is unreadable or does not contain an appropriate image,
     the ImageNotFoundQuerySignal is raised, which may be handled to
     proceed with some replacement image. If unhandled, nil is returned."

    |image name pathName fn nm inStream suffix readerClass 
     mustDecompress inPipe|

    "
     before trying each reader, check if the file is readable
    "
    name := aFileName asFilename name.

    inStream := Smalltalk systemFileStreamFor:name.
    inStream isNil ifTrue:[
        inStream := Smalltalk bitmapFileStreamFor:name.
        inStream isNil ifTrue:[

            "/ this signal is a query - if noone seems to
            "/ care, return nil.
            "/ However, a handler may provide a replacement.

"/            ('IMAGE: ' , aFileName , ' does not exist or is not readable') infoPrintCR.
            ^ ImageNotFoundQuerySignal
                        raiseRequestWith:aFileName
                        errorString:('IMAGE [warning]: ''' , aFileName , ''' does not exist or is not readable').
        ].
        name := 'bitmaps/' , name.
    ].
    inStream notNil ifTrue:[pathName := inStream pathName].
    inStream close.

    nm := name.
    fn := nm asFilename.
    suffix := fn suffix.

    "/
    "/ handle compressed-suffix
    "/
    (#('Z' 'gz') includes:suffix) ifTrue:[
        fn := fn withoutSuffix.
        nm := fn name.
        suffix := fn suffix.
        mustDecompress := true.
    ].

    suffix isEmpty ifTrue:[
        suffix := nm.
    ].

    "/ get the imageReader class from the files extension
    "/ and ask it first

    readerClass := MIMETypes imageReaderForSuffix:suffix.
    readerClass notNil ifTrue:[
        mustDecompress == true ifTrue:[
            inPipe := PipeStream readingFrom:'gunzip <' , pathName.
            inPipe notNil ifTrue:[
                [
                    image := readerClass fromStream:inPipe.
                ] valueNowOrOnUnwindDo:[
                    inPipe close
                ].
                image notNil ifTrue:[^ image].
            ]
        ] ifFalse:[
            BadImageFormatQuerySignal handle:[:ex |
                ex errorString infoPrintCR.
                image := nil.
                ex return.
            ] do:[
                image := readerClass fromFile:name.
            ].
            image notNil ifTrue:[^ image].
        ]
    ].

    "
     no known extension - ask all readers if they know
     this format ...
     ... these look into the file, and investigate the header.
     therefore, it takes a bit longer.
    "
    MIMETypes imageReaderClasses do:[:readerClass |
        readerClass notNil ifTrue:[
            (readerClass isValidImageFile:name) ifTrue:[
                ^ readerClass fromFile:name 
            ]
        ]
    ].

    "/ nope - unknown format
    "/
    "/ this signal is a query - if noone seems to
    "/ care, return nil.
    "/ However, a handler may provide a replacement.

"/    'IMAGE: unknown image file format: ' infoPrint. aFileName infoPrintNL.
    ^ ImageNotFoundQuerySignal
                raiseRequestWith:aFileName
                errorString:('IMAGE [warning]: unknown image file format: ''' , aFileName asFilename pathName , '''').

    "
     Image fromFile:'bitmaps/gifImages/claus.gif'
     Image fromFile:'bitmaps/gifImages/garfield.gif'

     Image fromFile:'bitmaps/winBitmaps/a11.ico'
     Image fromFile:'czech.xpm.gz'
     Image fromFile:'bitmaps/hello_world.icon'
    "

    "
     Image fromFile:'fooBar'
    "

    "giving a message for non-existing images:

     Image imageNotFoundQuerySignal 
     handle:[:ex |
        Transcript showCR:ex errorString.
        ex proceedWith:nil
     ] do:[
         Image fromFile:'fooBar'
     ]
    "

    "giving a replacement for non-existing images:

     Image imageNotFoundQuerySignal 
     answer:(Image fromFile:'bitmaps/SmalltalkX.xbm')
     do:[
         Image fromFile:'fooBar'
     ]
    "

    "Modified: / 3.2.1998 / 17:03:51 / cg"
!

fromFile:aFileName on:aDevice
    "read an image from a file and prepare a device representation.
     Return nil (or whatever a handler returned), 
     if the file is unreadable or does not contain an appropriate image."

    |img|

    img := self fromFile:aFileName.
    img notNil ifTrue:[
        ^ img on:aDevice
    ].
    ^ nil

    "Modified: 1.2.1997 / 14:48:07 / cg"
!

fromFile:aFileName resolution:res
    "read an image from a file and (if required) scale the image
     as appropriate (only with very high resolution displays).
     Return nil (or whatever a handler returned), 
     if the file is unreadable or does not contain an appropriate image."

    ^ self fromFile:aFileName resolution:res on:nil

    "Created: 19.12.1996 / 14:02:13 / cg"
    "Modified: 1.2.1997 / 14:48:16 / cg"
!

fromFile:aFileName resolution:dpi on:aDevice
    "read an image from a file and (if required) scale the image
     as appropriate (only with very high resolution displays).
     Prepare a device representation.
     Return nil (or whatever a handler returned), 
     if the file is unreadable or does not contain an appropriate image."

    |img dev dpiH mag|

    img := self fromFile:aFileName.
    img isNil ifTrue:[^ nil].

    "if the devices resolution is within +- 50% of dpi, no magnify is needed"
    dev := aDevice.
    dev isNil ifTrue:[
        "should not happen ..."
        dev := Screen current
    ].

    dpiH := dev horizontalPixelPerInch.
    ((dpi >= (dpiH * 0.75)) and:[dpi <= (dpiH * 1.5)]) ifTrue:[^ img].
    mag := (dpiH / dpi) rounded.
    mag = 0 ifTrue:[^ img].
    mag = 1 ifTrue:[^ img].

    img := img magnifiedBy:(mag @ mag).
    aDevice notNil ifTrue:[
        "should not happen ..."
        ^ img on:aDevice
    ].
    ^ img

    "Modified: 1.2.1997 / 14:48:20 / cg"
!

fromStream:aStream
    "read an image from a stream - this methods tries to find
     out the file format itself (by contents)
     and lets the appropriate reader read the file.
     To do this, the stream must be positionable.
     Return nil (or whatever a handler returned), 
     if the stream does not contain an appropriate image."

    |image|

    "
     ask all readers if they know
     this format ...
    "
    MIMETypes imageReaderClasses do:[:readerClass |
        readerClass notNil ifTrue:[
            image := readerClass fromStream:aStream.
            image notNil ifTrue:[^ image].
        ]
    ].

    "
     nope - unknown format
    "
"/    'Image [info]: unknown image file format in stream: ' infoPrintCR.

    ^ ImageNotFoundQuerySignal
                raiseRequestWith:aStream
                errorString:('IMAGE [warning]: unknown image file format in stream').

    "
     Image fromFile:'bitmaps/dano.tiff'
     Image fromFile:'bitmaps/test.fax'
     Image fromFile:'bitmaps/voice.tiff'
     Image fromFile:'voice.tiff'

     Image fromFile:'../fileIn/bitmaps/claus.gif'
     Image fromFile:'../fileIn/bitmaps/garfield.gif'

     Image fromFile:'../fileIn/bitmaps/founders.im8'
     Image fromFile:'../goodies/faces/next.com/steve.face'

     Image fromFile:'/LocalLibrary/Images/OS2/dos3.ico'
     Image fromFile:'bitmaps/globe1.xbm'
     Image fromFile:'bitmaps/globe1.xbm.Z'
     Image fromFile:'bitmaps/hello_world.icon'
    "

    "Created: 13.9.1996 / 18:06:00 / cg"
    "Modified: 30.6.1997 / 22:03:59 / cg"
!

fromStream:aStream using:aReaderClass
    "read an image from a stream, given an imageReaderClass.
     Use this, if you know the files format, but it has an invalid
     extension (or non-definite header), so #fromStream: could not
     find out the images format.
     Return nil (or whatever a handler returned), 
     if the stream does not contain an appropriate image."

    |image|

    image := aReaderClass fromStream:aStream.
    image notNil ifTrue:[^ image].

    "
     nope - unknown format
    "
"/    'Image [info]: unknown image file format in stream: ' infoPrintCR.

    ^ ImageNotFoundQuerySignal
                raiseRequestWith:aStream
                errorString:('IMAGE: unknown image file format in stream').

    "Created: 1.2.1997 / 14:46:20 / cg"
    "Modified: 1.2.1997 / 14:48:53 / cg"
! !

!Image class methodsFor:'misc'!

ditherAlgorithm
    "return the way we dither - 
        #threshold, or nil        -> no dither
        #pattern, or #ordered     -> orderedDither (ugly, but fast)
        #error or #floydSteinberg -> errorDiffusion; much better
        #burkes                   -> errorDiffusion; even better."

    ^ DitherAlgorithm

    "Created: 17.6.1996 / 18:57:47 / cg"
!

ditherAlgorithm:aSymbol
    "define how to dither - 
        #threshold, or nil        -> no dither
        #pattern, or #ordered     -> orderedDither (ugly, but fast)
        #error or #floydSteinberg -> errorDiffusion; much better
        #burkes                   -> errorDiffusion; even better."

    DitherAlgorithm := aSymbol

    "
     Image dither:#pattern
     Image dither:#floydSteinberg
     Image dither:#burkes
     Image dither:nil
    "

    "Modified: 10.6.1996 / 12:27:09 / cg"
    "Created: 17.6.1996 / 18:57:19 / cg"
!

numberOfDitherColors:n
    "define how many colors (i.e. patterns) to use when
     doing a pattern dither"

    NumberOfDitherColors := n
!

orderedDitherMatrixOfSize:sz
    sz == 2 ifTrue:[
        ^ #[
                0 2 
                3 1
           ].
    ].

    sz == 4 ifTrue:[
        ^ #[
                 0  8  2 10
                12  4 14  6
                 3 11  1  9
                15  7 13  5
           ].
    ].

    sz == 8 ifTrue:[
        ^  #[
                0 32  8 40    2 34 10 42
               48 16 56 24   50 18 58 26
               12 44  4 36   14 46  6 38    
               60 28 52 20   62 30 54 22

                3 35 11 43    1 33  9 41
               51 19 59 27   49 17 57 25
               15 47  7 39   13 45  5 37    
               63 31 55 23   61 29 53 21
            ].
    ].

    ^ nil

    "Created: 7.6.1996 / 14:15:05 / cg"
! !

!Image class methodsFor:'queries'!

bytesPerRowForWidth:width depth:bitsPerPixel padding:padding
    "helper - return the number of bytes in one scanline of an image,
     if scanlines are to be padded to padding-bits and the depth of the image is d."

    |bitsPerRow paddedUnitsPerRow|

    bitsPerRow := width * bitsPerPixel.
    paddedUnitsPerRow := bitsPerRow // padding.
    ((bitsPerRow \\ padding) ~~ 0) ifTrue:[
        paddedUnitsPerRow := paddedUnitsPerRow + 1
    ].
    ^ paddedUnitsPerRow * (padding // 8)


!

defaultPhotometric
    "return the default photometric pixel interpretation"

    ^ #blackIs0

    "Modified: 20.4.1996 / 23:40:41 / cg"
    "Created: 10.6.1996 / 18:08:12 / cg"
!

imageDepth
    "return the depth of images represented by instances of
     this class.
     Must be redefined in concrete subclasses"

    ^ self shouldNotImplement

    "Modified: 20.4.1996 / 23:40:41 / cg"
!

imageFileSuffixes
    "return a collection of suffixes which are recognized as image-file suffix"

    ^ MIMETypes imageFileSuffixes

    "
     Image imageFileSuffixes
    "

    "Modified: 30.6.1997 / 22:04:39 / cg"
!

imageReaderClassForMIME:mimeType
    "return an appropriate imageReader class for a given mime type;
     nil if there is none (or it did not install itself)"

    ^ MIMETypes imageReaderForType:mimeType

    "
     Image imageReaderClassForMIME:'image/tiff'
     Image imageReaderClassForMIME:'image/x-portable-pixmap'
    "

    "Created: 24.6.1997 / 22:32:27 / cg"
    "Modified: 30.6.1997 / 21:54:10 / cg"
!

imageReaderClassForSuffix:aSuffix
    "return an appropriate imageReader class for a given file-suffix;
     nil if there is none (or it did not install itself)"

    ^ MIMETypes imageReaderForSuffix:aSuffix

    "
     Image imageReaderClassForSuffix:'tiff'
     Image imageReaderClassForSuffix:'foo'
    "

    "Modified: 30.6.1997 / 22:05:14 / cg"
!

implementorForDepth:depth
    "return the class, which best implements images of depth"

    depth == 1 ifTrue:[^ Depth1Image].
    depth == 2 ifTrue:[^ Depth2Image].
    depth <= 4 ifTrue:[^ Depth4Image].
    depth <= 8 ifTrue:[^ Depth8Image].
    depth <= 16 ifTrue:[^ Depth16Image].
    depth <= 24 ifTrue:[^ Depth24Image].
    depth <= 32 ifTrue:[^ Depth32Image].
    ^ self

    "Modified: 24.4.1997 / 19:04:52 / cg"
!

isImageFileSuffix:aSuffix
    "return true, if the given suffix is known to be an image files suffix"

     ^ self imageFileSuffixes includes:(aSuffix asLowercase)

    "
     Image isImageFileSuffix:'gif'
     Image isImageFileSuffix:'xbm'
     Image isImageFileSuffix:'foo'
    "

    "Created: 18.4.1997 / 14:55:28 / cg"
!

mimeTypeFromSuffix:suffix
    "search my suffix information for a mime type and
     return it; return nil if unknown"

    ^ MIMETypes mimeTypeForSuffix:suffix

    "
     Image mimeTypeFromSuffix:'gif'      
     Image mimeTypeFromSuffix:'tiff'    
     Image mimeTypeFromSuffix:'foobar'  
    "

    "Modified: 1.7.1997 / 00:17:27 / cg"
! !

!Image class methodsFor:'screen capture'!

fromScreen
    "return an image of the full screen.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    |display|

    display := Screen current.
    ^ self 
        fromScreen:(0@0 corner:(display width @ display height))
        on:display
        grab:true

    "
     Image fromScreen
     Image fromScreen inspect
    "

    "Modified: 26.3.1997 / 10:44:55 / cg"
!

fromScreen:aRectangle
    "return an image of a part of the screen.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    ^ self 
        fromScreen:aRectangle 
        on:Screen current
        grab:true

    "
     Image fromScreen:(0@0 corner:100@100)
     (Image fromScreen:(0@0 corner:100@100)) inspect
    "

    "Modified: 26.3.1997 / 10:45:02 / cg"
!

fromScreen:aRectangle on:aDevice
    "return an image of a part of a screen, which may be on
     another display device.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    ^ self
        fromScreen:aRectangle 
        on:aDevice 
        grab:true

    "
     Image fromScreen:(0@0 corner:100@100)
    "

    "get a snapshot of your friends screen ...

     |dpy2|

     dpy2 := XWorkstation new initializeFor:'idefix:0'.
     (Image fromScreen:(dpy2 bounds) on:dpy2) inspect
    "

    "Modified: 26.3.1997 / 10:45:08 / cg"
!

fromScreen:aRectangle on:aDisplay grab:doGrab
    "return an image of a part of the screen, which may be on
     another Display. If the doGrab argument is true, the display
     is grabbed (i.e. blocked for others) and a camera cursor is
     shown while the readout is done.
     WARNING: with doGrab true, this temporarily grabs the display
              and it may not work from within a buttonMotion
              (use with a false grabArg then)."

    |depth vis img|

    depth := aDisplay depth.

    "/
    "/ for truecolor displays, return a Depth24Image
    "/ (must do this for depth15 & depth16 displays, since
    "/  Depth16Image has no way to specify r/g/b masks ...)
    "/
    vis := aDisplay visualType.
    (vis == #TrueColor or:[vis == #DirectColor]) ifTrue:[
        depth > 8 ifTrue:[
            depth := 24.
        ]
    ].

    img := (self implementorForDepth: depth) new.
    ^ img fromScreen:aRectangle on:aDisplay grab:doGrab

    "Modified: / 26.1.1998 / 22:23:08 / cg"
!

fromScreenArea
    "return an image of a part of the screen; 
     let user specify screen area.
     This is the same as #fromUser - kept for backward compatibility.
     Use #fromUser for ST-80 compatibility.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    |r|

    r := Rectangle fromUser.
    r width == 0 ifTrue:[^ nil].
    r height == 0 ifTrue:[^ nil].
    ^ self fromScreen:r

    "
     Image fromScreenArea
     Image fromScreenArea inspect
    "

    "Modified: / 16.4.1998 / 14:04:58 / cg"
!

fromUser
    "return an image of a part of the screen; let user specify screen area.
     Same as fromScreenArea, for ST-80 compatibility.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    ^ self fromScreenArea

    "
     Image fromUser
     Image fromUser inspect
    "

    "Modified: 26.3.1997 / 10:45:25 / cg"
!

fromView:aView
    "return an image taken from a views contents as currently
     on the screen. The returned image has the same depth and photometric
     as the Display. 
     Notice, that for invisible or partial covered views, 
     the returned Image is NOT correct. 
     You may want to raise the view before using this method.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromView:grab: with a false grabArg then)."

    ^ self
        fromView:aView 
        grab:true

    "
     Image fromView:(Launcher allInstances first window topView)
     Image fromView:(BrowserView allInstances first topView)
    "

    "get a snapshot from whichever view is active:

     |active|

     active := WindowGroup activeGroup topViews first.
     (Image fromView:active) inspect
    "

    "Modified: 9.9.1996 / 22:41:01 / stefan"
    "Modified: 26.3.1997 / 10:45:40 / cg"
!

fromView:aView grab:doGrab
    "return an image taken from a views contents as currently
     on the screen. If the doGrab argument is true, the display
     is grabbed (i.e. blocked for others) and a camera cursor is
     shown while the readout is done.
     The returned image has the same depth and photometric
     as the Display. 
     Notice, that for invisible or partial covered views, 
     the returned Image is NOT correct. 
     You may want to raise the view before using this method.
     WARNING: with doGrab true, this temporarily grabs the display
              and it may not work from within a buttonMotion
              (use with a false grabArg then)."

    |org dev|

    dev := aView graphicsDevice.
    org := dev translatePoint:(0@0)
                         from:(aView id)
                           to:dev rootWindowId.
    ^ self fromScreen:(org extent:aView extent) on:dev grab:doGrab

    "Created: 26.3.1997 / 10:34:20 / cg"
    "Modified: 26.3.1997 / 10:45:50 / cg"
! !

!Image methodsFor:'ST-80 compatibility'!

asCachedImage
    "return the receiver associated to the current screens device.
     For ST-80 compatibility 
     (ST/X uses Image for both device- and nonDevice-images)"

    ^ self on:Screen current

    "Modified: 23.4.1996 / 11:10:32 / cg"
!

asRetainedMedium
    "return the receiver associated to the current screens device.
     For ST-80 compatibility 
     (ST/X uses Image for both device- and nonDevice-images)"

    ^ self on:Screen current

    "Modified: 23.4.1996 / 11:10:32 / cg"
    "Created: 27.1.1997 / 15:49:08 / cg"
!

bounds:newBounds
    ^ self

    "Created: 10.2.1997 / 12:44:46 / cg"
!

containsPoint:aPoint
    "in st-80, images are visualComponents ..."

    ^ self bounds containsPoint:aPoint

    "Created: 6.3.1997 / 15:24:12 / cg"
!

convertToPalette:aColormap renderedBy:anImageRenderer
    "this does not really mimicri the corresponding ST-80 functionality"

    |monoBits convertedImage|

    aColormap size == 2 ifTrue:[
        anImageRenderer class == OrderedDither ifTrue:[
            monoBits := self orderedDitheredMonochromeBits.
        ] ifFalse:[
            monoBits := self floydSteinbergDitheredMonochromeBits.
        ].
        (((aColormap at:1) = Color black)
        and:[(aColormap at:2) = Color white]) ifTrue:[
            "/ ok
        ] ifFalse:[
            (((aColormap at:1) = Color white)
            and:[(aColormap at:2) = Color black]) ifTrue:[
                monoBits invert
            ]
        ].
        convertedImage := Depth1Image width:width height:height fromArray:monoBits.
        convertedImage palette:aColormap.
        ^ convertedImage
    ].

    self error:'unimplemented operation'.
    ^ self

    "Modified: 1.3.1997 / 17:25:50 / cg"
!

paintBasis
    "huh - whats that;
     I guess, I have to return Color for images without a mask,
     and CoverageValue for those with a mask; for now, always return Color"

    ^ ColorValue

    "Created: 6.3.1997 / 15:24:19 / cg"
!

preferredBounds
    ^ self bounds

    "Created: 10.2.1997 / 12:42:36 / cg"
! !

!Image methodsFor:'accessing'!

bitsPerSample
    "return the number of bits per sample.
     The return value is an array of bits-per-plane."

    bitsPerSample notNil ifTrue:[^ bitsPerSample].
    ^ Array with:self depth

    "Modified: 10.6.1996 / 18:04:21 / cg"
!

colorMap
    "return the colormap"

    ^ colorMap
!

colorMap:aColorMap
    "set the colorMap; this also set the photometric to a useful defualt."

    |sz "{ Class: SmallInteger }"
     sameColors|

    (aColorMap isNil and:[colorMap isNil]) ifTrue:[^ self].
    photometric == #palette ifTrue:[
        "/ any change at all ?
        (sz := aColorMap size) == colorMap size ifTrue:[
            sameColors := true.

            1 to:sz do:[:idx |
                (aColorMap at:idx) = (colorMap at:idx) ifFalse:[
                    sameColors := false.
                ]
            ].
            sameColors ifTrue:[^ self].
        ]
    ].

    colorMap := aColorMap.
    colorMap notNil ifTrue:[
        photometric := #palette.
    ] ifFalse:[
        (photometric == #palette) ifTrue:[
            photometric := #blackIs0
        ]
    ].
    deviceForm notNil ifTrue:[
        self release
    ]

    "Modified: 31.8.1995 / 03:05:59 / claus"
    "Modified: 7.3.1997 / 21:24:23 / cg"
!

colorMapFromArray: anArray
    "set the colorMap by reading colors from an array with rgb-byte values."

    |colors|
    colors := OrderedCollection new.

    1 to: anArray size by: 3 do:
    [:i|
        colors add: (Color redByte: (anArray at: i) greenByte: (anArray at: i + 1) blueByte: (anArray at: i + 2))
    ].

    self colorMap: colors
!

container:aVisualContainer
    "ignored here - added to allow images to be used like
     VisualComponents (later, Image should inherit from it)"

    "Created: 28.5.1996 / 23:43:49 / cg"
    "Modified: 29.5.1996 / 10:22:23 / cg"
!

depth
    "return the depth of the image"

    ^ self bitsPerPixel
!

device
    "return the device, the receiver is associated with.
     Return nil, if the image is unassigned."

    ^ device
!

extent
    "return the images extent"

    ^ width@height
!

fileName
    "return the value of the instance variable 'fileName' (automatically generated)"

    ^ fileName

    "Created: / 3.11.1997 / 14:54:46 / cg"
!

fileName:something
    "set the value of the instance variable 'fileName' (automatically generated)"

    fileName := something.

    "Created: / 3.11.1997 / 14:54:46 / cg"
!

fullColorId
    "return the id of the full color image on the device.
     Return nil, if the image is unassigned."

    fullColorDeviceForm isNil ifTrue:[^ nil].
    ^ fullColorDeviceForm id
!

graphicsDevice
    "same as #device - for ST-80 compatibility"

    ^ device
!

height
    "return the height of the image"

    ^ height
!

id
    "return the id of the image on the device.
     Return nil, if the image is unassigned."

    deviceForm isNil ifTrue:[^ nil].
    ^ deviceForm id
!

imageSequence
    "return the frameSequence of which this image is a part of;
     nil if it is not part of a sequence."

    ^ imageSequence

    "Created: / 1.4.1998 / 14:43:00 / cg"
!

imageSequence:aCollection
    "private entry for imageReaders - set the frameSequence of which
     this image is a part of"

    imageSequence := aCollection

    "Created: / 1.4.1998 / 14:42:36 / cg"
!

mask
    ^ mask

    "Created: 21.6.1996 / 12:57:44 / cg"
!

mask:anotherImage
    "set the images mask - currently, this may be nil or a Depth1Image.
     (it is planned to support alpha information in a Depth8 maskImage in
      the near future).
     For depth1 masks: each pixel of the image where a corresponding
                       1-bit is present in the mask will be drawn;
                       0-bit mask pixels lead to transparent pixels.

     For depth8 masks: (future):
                       each pixel specifies the alpha value (0..255),
                       which specifies the transparency of that pixel.
                       0 means completely transparent, 255 means completely
                       opaque. The 1-plane mask is a special case of this,
                       interpreting a 0 as a 0 alpha value and 1's as an
                       alpha value of 255."

    mask := anotherImage.
    maskedPixelsAre0 := false.

    "Created: 27.6.1996 / 17:45:22 / cg"
    "Modified: 12.4.1997 / 12:04:39 / cg"
!

maskedPixelsAre0
    "return true if masked pixels have been cleared to zero"

    maskedPixelsAre0 isNil ifTrue:[^ false].
    ^ maskedPixelsAre0

    "Modified: 14.4.1997 / 17:20:52 / cg"
!

maskedPixelsAre0:aBoolean
    "set/clear the flag which states if masked pixels
     have been set to zero. Knowing this to be true allows
     faster drawing of the image later; (however, not setting
     it will still produce correct output).
     This flag is typically set by image readers."

    maskedPixelsAre0 := aBoolean

    "Modified: 12.4.1997 / 12:08:42 / cg"
!

monochromeId
    "return the id of the monochrome image on the device.
     Return nil, if the image is unassigned."

    monoDeviceForm isNil ifTrue:[^ nil].
    ^ monoDeviceForm id
!

palette 
    "return the colormap; ST-80 compatibility"

    ^ colorMap
!

palette:aColormap 
    "set the colormap; ST-80 compatibility"

    self colorMap:aColormap

    "Created: 1.2.1996 / 15:09:25 / cg"
    "Modified: 8.6.1996 / 09:54:02 / cg"
!

photometric
    "return the photometric, a symbol such as #palette, #rgb etc."

    ^ photometric
!

samplesPerPixel
    "return the number of samples per pixel in the image.
     The return value is an array of bits-per-plane."

    samplesPerPixel notNil ifTrue:[^ samplesPerPixel].
    ^ 1

    "Modified: 10.6.1996 / 18:03:30 / cg"
!

width
    "return the width of the image"

    ^ width
! !

!Image methodsFor:'accessing - pixels'!

at:aPoint
    "WARNING: for now, this returns a pixels color 
     (backward compatibility with ST/X)
     In the future, this will return a pixel value (ST-80 compatibility)
     Use #colorAt: - for future compatibility.

     retrieve the pixel at aPoint; return a color.
     Pixels start at 0@0 for upper left pixel, end at
     (width-1)@(height-1) for lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    self obsoleteMethodWarning:'Image [warning]: the Image>>at: will change semantics soon; use #colorAt:'.
    ^ self colorAtX:(aPoint x) y:(aPoint y)

    "Modified: / 21.6.1997 / 13:10:17 / cg"
    "Modified: / 9.1.1998 / 20:33:52 / stefan"
!

at:aPoint put:aColor
    "WARNING: for now, this expects a pixels color 
     (backward compatibility with ST/X)
     In the future, this will expect a pixel value (ST-80 compatibility)
     Use #colorAt:put: - for future compatibility.

     set the pixel at aPoint to aColor.
     Pixels start at 0@0 for the upper left pixel, end at
     (width-1)@(height-1) for lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    self obsoleteMethodWarning:'Image [warning]: the Image>>at:put: will change semantics soon; use #colorAt:put:'.
    ^ self colorAtX:aPoint x y:aPoint y put:aColor

    "Modified: / 21.6.1997 / 13:16:02 / cg"
    "Modified: / 9.1.1998 / 20:34:15 / stefan"
!

atImageAndMask:aPoint put:aColor
    "set the pixel at x/y to aColor.
     If aColor is a mask color (i.e. Color noColor) mask pixel will be set to black (opaque),
     otherwise to white, so you are able to see the color of the image pixel.
     (used by the bitmap editor)"

    aColor ~= Color noColor
    ifTrue:
    [   
        mask notNil ifTrue: [mask atImageAndMask: aPoint put: Color white].  
        self colorAtX: aPoint x y: aPoint y put:aColor
    ] 
    ifFalse:
    [
        self colorAtX: aPoint x y: aPoint y put: ((colorMap includes: Color black) ifTrue: [Color black] ifFalse: [colorMap first]).
        mask notNil ifTrue: [mask atImageAndMask: aPoint put: Color black].  
    ].
!

atPoint:aPoint
    "ST-80 compatibility: return the pixelValue at:aPoint."

    ^ self pixelAtX:aPoint x y:aPoint y

    "Modified: 24.4.1997 / 16:18:44 / cg"
!

atPoint:aPoint put:aPixelValue
    "ST-80 compatibility: set the pixelValue at:aPoint."

    ^ self pixelAtX:aPoint x y:aPoint y put:aPixelValue

    "Modified: 24.4.1997 / 17:17:59 / cg"
!

atX:x y:y
    "WARNING: for now, this returns a pixels color 
     (backward compatibility with ST/X)
     In the future, this will return a pixel value (ST-80 compatibility)
     Use #colorAt: - for future compatibility.

     Retrieve a pixel at x/y; return a color.
     Pixels start at 0@0 for upper left pixel, end at
     (width-1)@(height-1) for lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    'Image [warning]: the Image>>atX:y: will change semantics soon; use #colorAtX:y:' infoPrintCR.
    ^ self colorAtX:x y:y

    "Modified: 21.6.1997 / 13:10:32 / cg"
!

atX:x y:y put:aColor
    "WARNING: for now, this expects a pixels color 
     (backward compatibility with ST/X)
     In the future, this will expect a pixel value (ST-80 compatibility)
     Use #colorAt:put: - for future compatibility.

     set the pixel at x/y to aColor.
     Pixels start at 0@0 for the upper left pixel, end at
     (width-1)@(height-1) for the lower right pixel.
     This method checks if the color can be stored in the image.
     (i.e. if the receiver is a palette image, the color must be present in there).
     You should not use this method for image-processing, it is very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    'Image [warning]: the Image>>atX:y:put: will change semantics soon; use #colorAtX:y:put:' infoPrintCR.
    ^ self colorAtX:x y:y put:aColor

    "Modified: 21.6.1997 / 13:10:44 / cg"
!

atX:x y:y putValue:aPixelValue
    "set the pixel at x/y to aPixelValue.
     The interpretation of the pixelValue depends on the photometric
     and the colormap. (see also: Image>>atX:y:put:)
     Pixels start at 0@0 for the upper left pixel, end at
     (width-1) @ (height-1) for the lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    self obsoleteMethodWarning:'use #pixelAtX:y:put:'.
    ^ self pixelAtX:x y:y put:aPixelValue.

    "Modified: 24.4.1997 / 17:15:45 / cg"
!

bits
    "return the raw image data; depending on the photometric,
     this has to be interpreted as monochrome, greyscale,
     palette or rgb data. It is also packed to be dense, so
     a 4 bitPerSample palette image will store 2 pixels per byte."

    ^ bytes
!

colorAt:aPoint
    "retrieve a pixel at x/y; return a color.
     Pixels start at 0@0 for upper left pixel, end at
     (width-1)@(height-1) for lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    ^ self colorAtX:(aPoint x) y:(aPoint y)

    "Created: 24.4.1997 / 17:02:31 / cg"
!

colorAt: aPoint put:aColor

    self colorAtX: aPoint x y: aPoint y put:aColor
!

colorAtX:x y:y
    "retrieve a pixel at x/y; return a color.
     Pixels start at 0@0 for upper left pixel, end at
     (width-1)@(height-1) for lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    |pixel|

    pixel := self pixelAtX:x y:y.
    ^ self colorFromValue:pixel

    "Modified: 24.4.1997 / 16:18:53 / cg"
    "Created: 24.4.1997 / 17:00:52 / cg"
!

colorAtX:x y:y put:aColor
    "set the pixel at x/y to aColor.
     Pixels start at 0@0 for the upper left pixel, end at
     (width-1)@(height-1) for the lower right pixel.
     This method checks if the color can be stored in the image.
     (i.e. if the receiver is a palette image, the color must be present in there).
     You should not use this method for image-processing, it is very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    |pixel|

    pixel := self valueFromColor:aColor.
    self pixelAtX:x y:y put:pixel.

    "Modified: 24.4.1997 / 17:36:20 / cg"
!

data
    "for backward compatibility - will vanish"

    self obsoleteMethodWarning:'use #bits'.
    ^ bytes

    "Modified: 24.4.1997 / 17:37:57 / cg"
!

data:aByteArray
    "for backward compatibility - will vanish"

    self obsoleteMethodWarning.
    bytes := aByteArray

    "Modified: 24.4.1997 / 17:38:18 / cg"
!

pixelAtX:x y:y
    "retrieve the pixelValue at aPoint; return an integer number.
     Pixels start at 0/0 for upper left pixel, and end at
     width-1@height-1 for lower right pixel.
     The returned numbers interpretation depends on the photometric
     and the colormap. (see also Image>>at: and Image>>atX:y:)
     You should not use this method for image-processing of
     big images, its very slow ... 
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    ^ self subclassResponsibility

    "Created: 24.4.1997 / 16:06:56 / cg"
!

pixelAtX:x y:y put:aPixelValue
    "set the pixel at x/y to aPixelValue.
     The interpretation of the pixelValue depends on the photometric
     and the colormap. (see also: Image>>atX:y:put:)
     Pixels start at 0@0 for the upper left pixel, end at
     (width-1) @ (height-1) for the lower right pixel.
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    ^ self subclassResponsibility

    "Created: 24.4.1997 / 17:05:11 / cg"
!

rowAt:rowIndex
    "retrieve an array filled with pixel values from
     a single row.
     Notice: row indexing starts at 0.
     This is a slow fallBack method, which works with any depth;
     concrete image subclasses should redefine this for more performance."

    |pixelArray|

    pixelArray := self pixelArraySpecies new:width.
    self rowAt:rowIndex into:pixelArray startingAt:1.
    ^ pixelArray

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     (i rowAt:0) inspect
    "
    "
     |i|

     i := Image fromFile:'bitmaps/SBrowser.xbm'.
     (i rowAt:0) inspect
    "

    "Modified: 24.4.1997 / 15:51:24 / cg"
!

rowAt:rowIndex into:aPixelBuffer
    "fill aBuffer with pixel values retrieved from a single row.
     Notice: row indexing starts at 0."

    ^ self rowAt:rowIndex into:aPixelBuffer startingAt:1

    "Created: 24.4.1997 / 15:44:46 / cg"
    "Modified: 24.4.1997 / 15:51:35 / cg"
!

rowAt:rowIndex into:aPixelBuffer startingAt:startIndex
    "fill aBuffer with pixel values retrieved from a single row.
     Notice: row indexing starts at 0.
     This is a slow fallBack method, which works with any depth;
     concrete image subclasses should redefine this for more performance."

    |w "{ Class: SmallInteger }"|

    w := width-1.
    0 to:w do:[:col |
        aPixelBuffer at:(col + startIndex) put:(self pixelAtX:col y:rowIndex)
    ].

    "Created: 24.4.1997 / 15:05:21 / cg"
    "Modified: 24.4.1997 / 16:52:43 / cg"
!

rowAt:rowIndex putAll:pixelArray
    "store a single rows pixels from bits in the argument;
     Notice: row indexing starts at 0.
     This is a slow fallBack method, which works with any depth;
     concrete image subclasses should redefine this for more performance."

    ^ self rowAt:rowIndex putAll:pixelArray startingAt:1

    "Modified: 24.4.1997 / 15:51:58 / cg"
!

rowAt:rowIndex putAll:pixelArray startingAt:startIndex
    "store a single rows pixels from bits in the pixelArray argument;
     return the pixelArray.
     Notice: row indexing starts at 0.
     This is a slow fallBack method, which works with any depth;
     concrete image subclasses should redefine this for more performance."

    |w "{ Class: SmallInteger }"|

    w := width-1.
    0 to:w do:[:col |
        self pixelAtX:col y:rowIndex put:(pixelArray at:(col + startIndex))
    ].
    ^ pixelArray

    "Modified: 24.4.1997 / 17:05:57 / cg"
!

valueAt:aPoint
    "WARNING: for now, this returns a pixels value 
     (backward compatibility with ST/X)
     In the future, this will return a color (ST-80 compatibility)
     Use #pixelAt: - for future compatibility.

     Retrieve the pixelValue at aPoint; return an integer number.
     Pixels start at 0@0 for upper left pixel, end at
     width-1@height-1 for lower right pixel.
     The returned numbers interpretation depends on the photometric
     and the colormap. (see also Image>>at: and Image>>atX:y:)
     You should not use this method for image-processing, its
     very slow ...
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    'Image [warning]: the Image>>valueAt: will change semantics soon; use #pixelAt:' infoPrintCR.
    ^ self pixelAtX:aPoint x y:aPoint y

    "Modified: 21.6.1997 / 13:11:19 / cg"
!

valueAtX:x y:y
    "WARNING: for now, this returns a pixels value 
     (backward compatibility with ST/X)
     In the future, this will return a color (ST-80 compatibility)
     Use #pixelAt: - for future compatibility.

     Retrieve the pixelValue at aPoint; return an integer number.
     Pixels start at 0/0 for upper left pixel, and end at
     width-1@height-1 for lower right pixel.
     The returned numbers interpretation depends on the photometric
     and the colormap. (see also Image>>at: and Image>>atX:y:)
     You should not use this method for image-processing of
     big images, its very slow ... 
     (it is meant to access individual pixels - for example, in a bitmap editor)"

    'Image [warning]: the Image>>valueAtX:y: will change semantics soon; use #pixelAtX:y' infoPrintCR.
    ^ self pixelAtX:x y:y

    "Modified: 21.6.1997 / 13:11:29 / cg"
! !

!Image methodsFor:'accessing - private'!

bits:aByteArray
    "set the raw data.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    bytes := aByteArray

    "Modified: 23.4.1996 / 11:08:28 / cg"
!

bitsPerSample:aNumber
    "set the number of bits per sample. 

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    bitsPerSample := aNumber

    "Modified: 23.4.1996 / 11:08:31 / cg"
!

depth:d
    "set the depth of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    depth := self depth.
    bitsPerPixel := d.
    d == 24 ifTrue:[
        samplesPerPixel := 3.
        bitsPerSample := #(8 8 8)
    ] ifFalse:[
        samplesPerPixel := 1.
        bitsPerSample := Array with:d 
    ]

    "Modified: 20.2.1997 / 14:39:10 / cg"
!

extent:anExtent
    "set the images extent.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := anExtent x.
    height := anExtent y

    "Modified: 23.4.1996 / 11:08:38 / cg"
!

height:aNumber
    "set the height of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    height := aNumber

    "Modified: 23.4.1996 / 11:08:40 / cg"
!

photometric:aSymbol
    "set the photometric interpretation of the pixel values.
     The argument, aSymbol is one of:
        #blackIs0, #whiteIs0, #palette, #rgb
     See TIFF documentation, from which the photometric concept is borrowed.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    |b|

    photometric := aSymbol.
    bitsPerSample isNil ifTrue:[
        photometric == #rgb ifTrue:[
            b := self class imageDepth // 3.
            bitsPerSample := Array with:b with:b with:b
        ] ifFalse:[
            bitsPerSample := Array with:(self class imageDepth)
        ].
    ].
    samplesPerPixel isNil ifTrue:[
        photometric == #rgb ifTrue:[
            samplesPerPixel := 3
        ] ifFalse:[
            samplesPerPixel := 1
        ]
    ].

    "Modified: 10.6.1996 / 18:21:29 / cg"
!

samplesPerPixel:aNumber
    "set the array of samples per pixel.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    samplesPerPixel := aNumber

    "Modified: 23.4.1996 / 11:08:45 / cg"
!

width:aNumber
    "set the width of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := aNumber

    "Modified: 23.4.1996 / 11:08:48 / cg"
!

width:w height:h 
    "set the width and height of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := w.
    height := h

    "Modified: 23.4.1996 / 11:08:53 / cg"
!

width:w height:h depth:d
    "set the width, height and depth of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := w.
    height := h.
    self depth:d.

    "Modified: 23.4.1996 / 11:08:56 / cg"
!

width:w height:h depth:d fromArray:bits
    "set the width, height, depth and pixels of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := w.
    height := h.
    self depth:d.
    bytes := bits

    "Modified: 23.4.1996 / 11:08:59 / cg"
!

width:w height:h depth:d palette:aColormap
    "set the width, height and depth of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := w.
    height := h.
    self depth:d.
    colorMap := aColormap.
    aColormap notNil ifTrue:[
        photometric := #palette
    ]

    "Modified: 23.4.1996 / 11:08:56 / cg"
    "Created: 6.3.1997 / 15:23:57 / cg"
!

width:w height:h photometric:p samplesPerPixel:spp bitsPerSample:bps colorMap:cm bits:pixels
    "set all relevant internal state of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    ^ self
        width:w 
        height:h 
        photometric:p 
        samplesPerPixel:spp 
        bitsPerSample:bps 
        colorMap:cm 
        bits:pixels 
        mask:nil

    "Modified: 20.6.1996 / 17:10:24 / cg"
!

width:w height:h photometric:p samplesPerPixel:spp bitsPerSample:bps colorMap:cm bits:pixels mask:m
    "set all relevant internal state of the image.

     This interface is only to be used when initializing
     instances or by image readers. Calling for a change of an
     existing image may confuse later pixel interpretation."

    width := w.
    height := h.
    photometric := p.
    samplesPerPixel := spp.
    bitsPerSample := bps.
    colorMap := cm.
    bytes := pixels.
    mask := m.

    "Modified: 23.4.1996 / 11:09:02 / cg"
    "Created: 20.6.1996 / 17:09:53 / cg"
! !

!Image methodsFor:'binary storage'!

readBinaryContentsFrom: stream manager: manager
    "read a binary representation of an image from stream.
     Redefined to flush any device data."

    super readBinaryContentsFrom: stream manager: manager.
    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.
!

storeBinaryDefinitionOn: stream manager: manager
    "store a binary representation of the receiver on stream.
     This is an internal interface for binary storage mechanism.
     Redefined to not store the device form (which is recreated at
     load time anyway)"

    |tDevice tDeviceForm tMonoDeviceForm tFullColorDeviceForm|

    tDevice := device.
    tDeviceForm := deviceForm.
    tMonoDeviceForm := monoDeviceForm.
    tFullColorDeviceForm := fullColorDeviceForm.

    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.

    super storeBinaryDefinitionOn: stream manager: manager.

    device := tDevice.
    deviceForm := tDeviceForm.
    monoDeviceForm := tMonoDeviceForm.
    fullColorDeviceForm := tFullColorDeviceForm.

    "Modified: 23.4.1996 / 09:30:50 / cg"
! !

!Image methodsFor:'converting'!

asBurkesDitheredMonochromeImage
    "return a burkes dithered monochrome image from the receiver image.
     Depending on the images contents, this may or may not look better than
     a floyd-steinberg dithered image.
     Notice, that floyd-steinberg dithering is faster; both because less
     error diffusion is done and due to being specially tuned."

    |monoBits|

    monoBits := self burkesDitheredMonochromeBits.
    ^ Depth1Image width:width height:height fromArray:monoBits

    "
     |i|

     i := Image fromFile:'garfield.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Image fromFile:'claus.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ].
     i := i magnifiedBy:30.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "Created: 10.6.1996 / 12:34:44 / cg"
    "Modified: 12.6.1996 / 13:58:16 / cg"
!

asErrorDitheredMonochromeImage
    "return an error-diffusion dithered monochrome image from the receiver image."

    DitherAlgorithm == #burkes ifTrue:[
        ^ self asBurkesDitheredMonochromeImage
    ].
    DitherAlgorithm == #stevensonArce ifTrue:[
        ^ self asStevensonArgceDitheredMonochromeImage
    ].
    ^ self asFloydSteinbergDitheredMonochromeImage

    "
     |i|

     i := Image fromFile:'garfield.gif'.
     i inspect.
     i asErrorDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Image fromFile:'claus.gif'.
     i inspect.
     i asErrorDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ].
     i := i magnifiedBy:30.
     i inspect.
     i asErrorDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "Modified: 10.6.1996 / 14:22:30 / cg"
!

asFloydSteinbergDitheredDepth8FormOn:aDevice colors:fixColors 
    "return a floyd-steinberg dithered pseudoForm from the picture. 
     Use the colors in the fixColors array.
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create dithered Depth4Images from Depth8Images."

    |pseudoBits f has8BitImage deviceDepth
     map |

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
        has8BitImage := true.
    ] ifFalse:[
        has8BitImage := false.
        aDevice supportedImageFormats do:[:fmt |
            (fmt at:#bitsPerPixel) == 8 ifTrue:[
                has8BitImage := true.
            ]
        ]
    ].
    has8BitImage ifFalse:[^ nil].

    pseudoBits := self floydSteinbergDitheredDepth8BitsColors:fixColors.
    pseudoBits isNil ifTrue:[^ nil].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].

    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256.
    fixColors do:[:clr |
        map at:clr colorId + 1 put:clr
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice 
	drawBits:pseudoBits 
	bitsPerPixel:8 
	depth:deviceDepth  
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width height:height 
	with:(f gcId).
    ^ f

    "
     example: 
        color reduction from Depth8 to Depth4 (dithering) can be done by:

     |img8 reducedImg8 img4 map form|

     map := #( 
                  (0     0   0)
                  (0     0 100)
                  (0    50   0)
                  (0    50 100)
                  (0   100   0)
                  (0   100 100)
                  (100   0   0)
                  (100   0 100)
                  (100  50   0)
                  (100  50 100)
                  (100 100   0)
                  (100 100 100)) collect:[:rgb | (Color red:(rgb at:1)
                                                      green:(rgb at:2)
                                                       blue:(rgb at:3)) on:Display].

     img8 := Image fromFile:'bitmaps/bf.im8'.
     form := img8 paletteImageAsDitheredPseudoFormOn:Display 
                      colors:map 
                        nRed:2
                      nGreen:3
                       nBlue:2.
     img8 := Depth8Image fromForm:form.    'dithered version of original image'.
     img4 := Depth4Image fromImage:img8.
    "

    "Modified: 14.6.1996 / 16:52:34 / cg"
    "Created: 23.6.1997 / 15:25:37 / cg"
!

asFloydSteinbergDitheredDepth8FormOn:aDevice colors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue 
    "return a floyd-steinberg dithered pseudoForm from the palette picture. 
     Use the colors in the fixColors array, which must be fixR x fixG x fixB
     colors assigned to aDevice, such as the preallocated colors of the
     Color class. 
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create dithered Depth4Images from Depth8Images."

    |pseudoBits f has8BitImage deviceDepth
     map |

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
        has8BitImage := true.
    ] ifFalse:[
        has8BitImage := false.
        aDevice supportedImageFormats do:[:fmt |
            (fmt at:#bitsPerPixel) == 8 ifTrue:[
                has8BitImage := true.
            ]
        ]
    ].
    has8BitImage ifFalse:[^ nil].

    pseudoBits := self floydSteinbergDitheredDepth8BitsColors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue.
    pseudoBits isNil ifTrue:[^ nil].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].

    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256.
    fixColors do:[:clr |
        map at:clr colorId + 1 put:clr
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice 
	drawBits:pseudoBits 
	bitsPerPixel:8 
	depth:deviceDepth  
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width height:height 
	with:(f gcId).
    ^ f

    "
     example: 
        color reduction from Depth8 to Depth4 (dithering) can be done by:

     |img8 reducedImg8 img4 map form|

     map := #( 
                  (0     0   0)
                  (0     0 100)
                  (0    50   0)
                  (0    50 100)
                  (0   100   0)
                  (0   100 100)
                  (100   0   0)
                  (100   0 100)
                  (100  50   0)
                  (100  50 100)
                  (100 100   0)
                  (100 100 100)) collect:[:rgb | (Color red:(rgb at:1)
                                                      green:(rgb at:2)
                                                       blue:(rgb at:3)) on:Display].

     img8 := Image fromFile:'bitmaps/bf.im8'.
     form := img8 paletteImageAsDitheredPseudoFormOn:Display 
                      colors:map 
                        nRed:2
                      nGreen:3
                       nBlue:2.
     img8 := Depth8Image fromForm:form.    'dithered version of original image'.
     img4 := Depth4Image fromImage:img8.
    "

    "Modified: 14.6.1996 / 16:52:34 / cg"
!

asFloydSteinbergDitheredGrayFormOn:aDevice
    "return a dithered depth-x grey form from the receiver image."

    |depth bits|

    depth := aDevice depth.
    (depth == 1
    or:[aDevice hasGrayscales not]) ifTrue:[
        "/ for monochrome, there is specialized
        "/ monochrome dither code available

        bits := self floydSteinbergDitheredMonochromeBits.
    ] ifFalse:[
        bits := self floydSteinbergDitheredGrayBitsDepth:depth.
    ].

    ^ self makeDeviceGrayPixmapOn:aDevice depth:depth fromArray:bits

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     (i asFloydSteinbergDitheredGrayFormOn:Display) inspect
    "

    "
     |i|

     i := Image fromFile:'bitmaps/granite.tiff'.
     (i asFloydSteinbergDitheredGrayFormOn:Display) inspect
    "

    "Created: 10.6.1996 / 14:11:39 / cg"
    "Modified: 17.4.1997 / 01:11:54 / cg"
!

asFloydSteinbergDitheredGrayImageDepth:depth
    "return a floyd-steinberg dithered image from the receiver image."

    |ditheredBits|

    (depth == 1) ifTrue:[
        ^ self asFloydSteinbergDitheredMonochromeImage
    ].

    ditheredBits := self floydSteinbergDitheredGrayBitsDepth:depth.
    ^ (self class implementorForDepth:depth)
        width:width height:height fromArray:ditheredBits

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     i inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "

    "
     |i|

     i := Depth24Image width:4 height:1
          fromArray:#[ 
            16rFF 16r00 16r00   16rFF 16r00 16r00  16rFF 16r00 16r00  16rFF 16r00 16r00
            16rFF 16r00 16r00   16rFF 16r00 16r00  16rFF 16r00 16r00  16rFF 16r00 16r00
            16rFF 16r00 16r00   16rFF 16r00 16r00  16rFF 16r00 16r00  16rFF 16r00 16r00
            16rFF 16r00 16r00   16rFF 16r00 16r00  16rFF 16r00 16r00  16rFF 16r00 16r00].
     i := i magnifiedBy:4@1.
     i inspect.

     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "

    "
     |i|

     i := Depth24Image width:4 height:6
          fromArray:#[ 
            16r00 16r00 16r00   16r00 16r00 16r80  16r00 16r00 16rff  16r00 16r80 16r00
            16r00 16r80 16r80   16r00 16rFF 16r00  16r00 16rFF 16r80  16r00 16rFF 16rFF
            16r80 16r00 16r00   16r80 16r00 16r80  16r80 16r00 16rff  16r80 16r80 16r00
            16r80 16r80 16r80   16r80 16rFF 16r00  16r80 16rFF 16r80  16r80 16rFF 16rFF
            16rFF 16r00 16r00   16rFF 16r00 16r80  16rFF 16r00 16rff  16rFF 16r80 16r00
            16rFF 16r80 16r80   16rFF 16rFF 16r00  16rFF 16rFF 16r80  16rFF 16rFF 16rFF].
     i := i magnifiedBy:30.
     i inspect.

     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "
    "
     |i|

     i := Image fromFile:'granite.tiff'.
     i inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "

    "
     |i|

     i := Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ].
     i := i magnifiedBy:30.
     i inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
    "

    "Created: 10.6.1996 / 12:33:47 / cg"
    "Modified: 19.10.1997 / 04:09:04 / cg"
!

asFloydSteinbergDitheredMonochromeFormOn:aDevice
    "return a dithered moncohrome form from the receiver image."

    |monoBits|

    monoBits := self floydSteinbergDitheredMonochromeBits.
    ^ self makeDeviceMonochromeBitmapOn:aDevice fromArray:monoBits

    "
     |i f|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     (i asFloydSteinbergDitheredMonochromeFormOn:Display) inspect.
    "

    "
     |i f|

     i := Depth2Image width:8 height:8
          fromArray:#[
                        4r0000 4r0000
                        4r0000 4r0000
                        4r1111 4r1111
                        4r1111 4r1111
                        4r2222 4r2222
                        4r2222 4r2222
                        4r3333 4r3333
                        4r3333 4r3333
                     ].
     (i asFloydSteinbergDitheredMonochromeFormOn:Display) inspect.
    "

    "Created: 10.6.1996 / 14:11:39 / cg"
    "Modified: 17.4.1997 / 01:14:02 / cg"
!

asFloydSteinbergDitheredMonochromeImage
    "return a floyd-steinberg dithered monochrome image from the receiver image."

    |monoBits|

    monoBits := self floydSteinbergDitheredMonochromeBits.
    ^ Depth1Image width:width height:height fromArray:monoBits

    "
     |i|

     i := Image fromFile:'garfield.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ].
     i := i magnifiedBy:30.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "Created: 10.6.1996 / 12:33:47 / cg"
    "Modified: 17.4.1997 / 01:15:28 / cg"
!

asFloydSteinbergDitheredPseudoFormUsing:colors on:aDevice
    "return a floyd-steinberg dithered pseudoForm from the picture,
     using colors in colors for dithering."

    |pseudoBits pseudoBits8 f has8BitImage deviceDepth
     map d |

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
        has8BitImage := true.
    ] ifFalse:[
        has8BitImage := false.
        aDevice supportedImageFormats do:[:fmt |
            (fmt at:#bitsPerPixel) == 8 ifTrue:[
                has8BitImage := true.
            ]
        ]
    ].
    has8BitImage ifFalse:[
        deviceDepth == 4 ifFalse:[^ nil].

        pseudoBits8 := self nfloydSteinbergDitheredDepth8BitsColors:colors.
        pseudoBits8 isNil ifTrue:[^ nil].
        "/ convert to devices depth

        pseudoBits := ByteArray new:(width*4+7//8 * height).
        pseudoBits8 compressPixels:4 width:width height:height into:pseudoBits mapping:nil.
        d := 4.
    ] ifTrue:[
        pseudoBits := self nfloydSteinbergDitheredDepth8BitsColors:colors.
        pseudoBits isNil ifTrue:[^ nil].
        d := 8.
    ].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].

    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256 withAll:0.
    colors do:[:clr |
        clr notNil ifTrue:[
            map at:clr colorId + 1 put:clr
        ]
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice 
	drawBits:pseudoBits 
	bitsPerPixel:d 
	depth:deviceDepth  
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width height:height 
	with:(f gcId).
    ^ f

    "
     |i|

     i := Image fromFile:'bitmaps/garfield.gif'.
     (i asFloydSteinbergDitheredPseudoFormUsing:(Smalltalk at:#'Color:DitherColors') on:Display) inspect


     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asFloydSteinbergDitheredPseudoFormUsing:(Smalltalk at:#'Color:DitherColors') on:Display) inspect
    "

    "Created: 17.6.1996 / 12:13:35 / cg"
    "Modified: 5.9.1996 / 19:42:57 / cg"
!

asFormOn:aDevice
    "get a device form, with best possible approximation.
     remember it in case someone asks again."

    |form visual|

    ((aDevice == device) and:[deviceForm notNil]) ifTrue:[^ deviceForm].

    mask notNil ifTrue:[
        mask := mask on:aDevice
    ].

    (aDevice depth == 1
    or:[aDevice hasGrayscales not]) ifTrue:[
        ^ self asMonochromeFormOn:aDevice
    ].

    ((visual := aDevice visualType) == #StaticGray) ifTrue:[
        ^ self asGrayFormOn:aDevice
    ].

    (visual == #PseudoColor
     or:[visual == #StaticColor]) ifTrue:[
        form := self asPseudoFormQuickOn:aDevice.
        form notNil ifTrue:[
            ^ form
        ].
    ].

    (photometric == #palette) ifTrue:[
        form := self paletteImageAsFormOn:aDevice
    ] ifFalse:[
        (photometric == #rgb) ifTrue:[
            form := self rgbImageAsFormOn:aDevice
        ] ifFalse:[
            form := self greyImageAsFormOn:aDevice
        ]
    ].

    (device isNil or:[aDevice == device]) ifTrue:[
        "remember this form in the receiver ..."

        form notNil ifTrue:[
            deviceForm := form.
            device isNil ifTrue:[
                device := aDevice.
                Lobby register:self
            ] ifFalse:[
                Lobby registerChange:self
            ].
            "
             can save space, by not keeping the images data-bits
             twice (here and in the device form)
            "
            form forgetBits
        ]
    ].

    ^ form

    "Modified: 10.7.1996 / 21:55:44 / cg"
!

asGrayFormOn:aDevice
    "get a gray device form"

    ^ self asGrayFormOn:aDevice dither:DitherAlgorithm.

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asGrayFormOn:Display) inspect.
    "

    "Modified: 10.6.1996 / 17:39:30 / cg"
    "Created: 10.6.1996 / 18:44:42 / cg"
!

asGrayFormOn:aDevice dither:aDitherAlgorithm
    "get a greyscale device form, using aDitherAlgorithm."

    |depth|

    depth := aDevice depth.
    (depth == 1
    or:[aDevice hasGrayscales not]) ifTrue:[
        ^ self asMonochromeFormOn:aDevice
    ].

    (aDitherAlgorithm isNil
    or:[aDitherAlgorithm == #threshold]) ifTrue:[
        ^ self asThresholdGrayFormOn:aDevice
    ].

    (aDitherAlgorithm == #pattern
    or:[aDitherAlgorithm == #ordered]) ifTrue:[
        ^ self asOrderedDitheredGrayFormOn:aDevice
    ].

    ^ self asFloydSteinbergDitheredGrayFormOn:aDevice.

    "Created: 10.6.1996 / 18:42:01 / cg"
    "Modified: 14.6.1996 / 15:17:28 / cg"
!

asGrayImageDepth:depth
    "get a gray image from the receiver"

    ^ self asGrayImageDepth:depth dither:DitherAlgorithm.

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asGrayFormOn:Display) inspect.
    "

    "Modified: 10.6.1996 / 17:39:30 / cg"
    "Created: 10.6.1996 / 19:07:08 / cg"
!

asGrayImageDepth:depth dither:aDitherAlgorithm
    "get a greyscale image, using aDitherAlgorithm."

    (aDitherAlgorithm isNil
    or:[aDitherAlgorithm == #threshold]) ifTrue:[
        ^ self asThresholdGrayImageDepth:depth
    ].

    (aDitherAlgorithm == #pattern
    or:[aDitherAlgorithm == #ordered]) ifTrue:[
        ^ self asOrderedDitheredGrayImageDepth:depth
    ].

    ^ self asFloydSteinbergDitheredGrayImageDepth:depth

    "Created: 10.6.1996 / 19:08:21 / cg"
!

asMonochromeFormOn:aDevice
    "get a monochrome device form"

    |form|

    ((aDevice == device) and:[monoDeviceForm notNil]) ifTrue:[^ monoDeviceForm].
    depth == 1 ifTrue:[
        ((aDevice == device) and:[deviceForm notNil]) ifTrue:[^ deviceForm].
    ].

    form := self asMonochromeFormOn:aDevice dither:DitherAlgorithm.

    (device isNil or:[aDevice == device]) ifTrue:[
        "remember this form in the receiver ..."

        form notNil ifTrue:[
            monoDeviceForm := form.
            device isNil ifTrue:[
                device := aDevice.
                Lobby register:self
            ] ifFalse:[
                Lobby registerChange:self
            ].
            "
             can save space, by not keeping the images data-bits
             twice (here and in the device form)
            "
            form forgetBits
        ]
    ].

    ^ form

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asMonochromeFormOn:Display) inspect.
    "

    "Modified: 23.10.1997 / 00:44:59 / cg"
!

asMonochromeFormOn:aDevice dither:aDitherAlgorithm
    "get a monochrome device form, using aDitherAlgorithm."

    |monoBits|

    (aDitherAlgorithm isNil
    or:[aDitherAlgorithm == #threshold]) ifTrue:[
        ^ self asThresholdMonochromeFormOn:aDevice
    ].

    aDitherAlgorithm == #burkes ifTrue:[
        monoBits := self burkesDitheredMonochromeBits.
    ] ifFalse:[
        aDitherAlgorithm == #stevensonArce ifTrue:[
            monoBits := self stevensonArceDitheredMonochromeBits.
        ] ifFalse:[
            (aDitherAlgorithm == #pattern
            or:[aDitherAlgorithm == #ordered]) ifTrue:[
                ^ self asOrderedDitheredGrayFormOn:aDevice.
            ] ifFalse:[
                ^ self asFloydSteinbergDitheredMonochromeFormOn:aDevice.
            ]
        ]
    ].

    "/
    "/ make its pixel interpretation correct for the device
    "/
    ^ self makeDeviceMonochromeBitmapOn:aDevice fromArray:monoBits

    "Modified: 10.6.1996 / 20:18:05 / cg"
!

asNearestPaintDepth8FormOn:aDevice colors:fixColors 
    "return a nearest paint pseudoForm from the palette picture. 
     Use the colors in the fixColors array. 
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create dithered Depth4Images from Depth8Images."

    |pseudoBits f has8BitImage deviceDepth
     map |

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
        has8BitImage := true.
    ] ifFalse:[
        has8BitImage := false.
        aDevice supportedImageFormats do:[:fmt |
            (fmt at:#bitsPerPixel) == 8 ifTrue:[
                has8BitImage := true.
            ]
        ]
    ].
    has8BitImage ifFalse:[^ nil].

    pseudoBits := self nearestPaintDepth8BitsColors:fixColors.
    pseudoBits isNil ifTrue:[^ nil].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].

    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256.
    fixColors do:[:clr |
        map at:clr colorId + 1 put:clr
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice 
	drawBits:pseudoBits 
	bitsPerPixel:8 
	depth:deviceDepth  
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width height:height 
	with:(f gcId).
    ^ f

    "
     example: 
        color reduction from Depth8 to Depth4 (dithering) can be done by:

     |img8 reducedImg8 img4 map form|

     map := #( 
                  (0     0   0)
                  (0     0 100)
                  (0    50   0)
                  (0    50 100)
                  (0   100   0)
                  (0   100 100)
                  (100   0   0)
                  (100   0 100)
                  (100  50   0)
                  (100  50 100)
                  (100 100   0)
                  (100 100 100)) collect:[:rgb | (Color red:(rgb at:1)
                                                      green:(rgb at:2)
                                                       blue:(rgb at:3)) on:Display].

     img8 := Image fromFile:'bitmaps/claus.gif'.
     form := img8 asNearestPaintDepth8FormOn:Display 
                      colors:map 
                        nRed:2
                      nGreen:3
                       nBlue:2.
     img8 := Depth8Image fromForm:form.    'dithered version of original image'.
     img4 := Depth4Image fromImage:img8.
    "

    "Modified: 17.6.1996 / 18:52:47 / cg"
    "Created: 23.6.1997 / 15:26:09 / cg"
!

asNearestPaintDepth8FormOn:aDevice colors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue 
    "return a nearest paint pseudoForm from the palette picture. 
     Use the colors in the fixColors array, which must be fixR x fixG x fixB
     colors assigned to aDevice, such as the preallocated colors of the
     Color class. 
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create dithered Depth4Images from Depth8Images."

    |pseudoBits f has8BitImage deviceDepth
     map |

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
        has8BitImage := true.
    ] ifFalse:[
        has8BitImage := false.
        aDevice supportedImageFormats do:[:fmt |
            (fmt at:#bitsPerPixel) == 8 ifTrue:[
                has8BitImage := true.
            ]
        ]
    ].
    has8BitImage ifFalse:[^ nil].

    pseudoBits := self nearestPaintDepth8BitsColors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue.
    pseudoBits isNil ifTrue:[^ nil].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].

    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256.
    fixColors do:[:clr |
        map at:clr colorId + 1 put:clr
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice 
	drawBits:pseudoBits 
	bitsPerPixel:8 
	depth:deviceDepth  
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width height:height 
	with:(f gcId).
    ^ f

    "
     example: 
        color reduction from Depth8 to Depth4 (dithering) can be done by:

     |img8 reducedImg8 img4 map form|

     map := #( 
                  (0     0   0)
                  (0     0 100)
                  (0    50   0)
                  (0    50 100)
                  (0   100   0)
                  (0   100 100)
                  (100   0   0)
                  (100   0 100)
                  (100  50   0)
                  (100  50 100)
                  (100 100   0)
                  (100 100 100)) collect:[:rgb | (Color red:(rgb at:1)
                                                      green:(rgb at:2)
                                                       blue:(rgb at:3)) on:Display].

     img8 := Image fromFile:'bitmaps/claus.gif'.
     form := img8 asNearestPaintDepth8FormOn:Display 
                      colors:map 
                        nRed:2
                      nGreen:3
                       nBlue:2.
     img8 := Depth8Image fromForm:form.    'dithered version of original image'.
     img4 := Depth4Image fromImage:img8.
    "

    "Created: 17.6.1996 / 18:47:46 / cg"
    "Modified: 17.6.1996 / 18:52:47 / cg"
!

asOrderedDitheredGrayFormOn:aDevice
    "return a dithered depth-x grey form from the receiver image.
     Uses an 8x8 dithermatrix."

    |depth bits|

    depth := aDevice depth.
    (depth == 1 
    or:[aDevice hasGrayscales not]) ifTrue:[
        "/ for monochrome, there is highly specialized
        "/ monochrome dither code available

        ^ self asOrderedDitheredMonochromeFormOn:aDevice
    ].

    bits := self
                orderedDitheredGrayBitsWithDitherMatrix:(self class orderedDitherMatrixOfSize:8)
                ditherWidth:8
                depth:depth.

    ^ self makeDeviceGrayPixmapOn:aDevice depth:depth fromArray:bits

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asOrderedDitheredGrayFormOn:Display) inspect
    "

    "Modified: 24.6.1997 / 22:19:30 / cg"
!

asOrderedDitheredGrayImageDepth:depth
    "return a dithered depth-x grey image from the receiver image.
     Uses an 8x8 dithermatrix."

    |dither|

    dither := self class orderedDitherMatrixOfSize:8.

    (depth == 1) ifTrue:[
        "/ for monochrome, there is highly specialized
        "/ monochrome dither code available

        ^ Depth1Image
            width:width
            height:height
            fromArray:(
                self
                    orderedDitheredMonochromeBitsWithDitherMatrix:dither
                    ditherWidth:8)
    ].

    ^ (self class implementorForDepth:depth)
        width:width
        height:height
        fromArray:(
            self
                orderedDitheredGrayBitsWithDitherMatrix:dither
                ditherWidth:8
                depth:depth)

    "
     |i i1 i2 i4 i8|

     i := Image fromFile:'bitmaps/claus.gif'.
     i1 := i asOrderedDitheredGrayImageDepth:1.
     i1 inspect.

     i2 := i asOrderedDitheredGrayImageDepth:2.
     i2 inspect.

     i4 := i asOrderedDitheredGrayImageDepth:4.
     i4 inspect.

     i8 := i asOrderedDitheredGrayImageDepth:8.
     i8 inspect.
    "

    "
     |i i1 i2 i4 i8|
     i := Image fromFile:'bitmaps/garfield.gif'.
     i1 := i asOrderedDitheredGrayImageDepth:1.
     i1 inspect.

     i2 := i asOrderedDitheredGrayImageDepth:2.
     i2 inspect.

     i4 := i asOrderedDitheredGrayImageDepth:4.
     i4 inspect.

     i8 := i asOrderedDitheredGrayImageDepth:8.
     i8 inspect.

     i2 := i8 asOrderedDitheredGrayImageDepth:2.
     i2 inspect.
    "

    "Created: 7.6.1996 / 18:03:54 / cg"
    "Modified: 24.6.1997 / 22:19:36 / cg"
!

asOrderedDitheredMonochromeFormOn:aDevice
    "return a dithered monochrome form from the grey image.
     Uses a 4x4 dithermatrix."

    "/ 4x4 seems a good comprimize between:
    "/    number of grey levels (8x8 is better)
    "/    artifacts             (4x4 is better)
    "/
    "/ could look at the image and decide upon the number of
    "/ distinct colors present. Use 8x8 for high-number,
    "/ 4x4 for small number of colors ...

    ^ self 
        asOrderedDitheredMonochromeFormOn:aDevice 
        ditherMatrix:(self class orderedDitherMatrixOfSize:4)
        ditherWidth:4

"/    ^ self 
"/        asOrderedDitheredMonochromeFormOn:aDevice 
"/        ditherMatrix:(self class orderedDitherMatrixOfSize:8)
"/        ditherWidth:8

    "
     |i f|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     f := i asOrderedDitheredMonochromeFormOn:Display.


     |i f|

     i := (Image fromFile:'bitmaps/winBitmaps/a11.ico') magnifiedBy:10.
     f := i asOrderedDitheredMonochromeFormOn:Display.


     |i f|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     f := i asOrderedDitheredMonochromeFormOn:Display.


     |i f|

     i := (Image fromFile:'bitmaps/PasteButton.tiff') magnifiedBy:10.
     f := i asOrderedDitheredMonochromeFormOn:Display.


     |i f|

     i := (Image fromFile:'bitmaps/blue-ball.gif') magnifiedBy:1.
     f := i asOrderedDitheredMonochromeFormOn:Display.
    "

    "Created: 7.6.1996 / 14:52:32 / cg"
    "Modified: 17.4.1997 / 01:10:10 / cg"
!

asOrderedDitheredMonochromeFormOn:aDevice ditherMatrix:ditherMatrix ditherWidth:dW
    "return a dithered monochrome form from the image.
     Uses the passed ditherMatrix and ditherWidth."

    |monoBits|

    monoBits := self orderedDitheredMonochromeBitsWithDitherMatrix:ditherMatrix ditherWidth:dW.
    ^ self makeDeviceMonochromeBitmapOn:aDevice fromArray:monoBits


    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     (i asOrderedDitheredMonochromeFormOn:Display) inspect


     |i|

     i := Image fromFile:'bitmaps/granite.tiff'.
     (i asOrderedDitheredMonochromeFormOn:Display) inspect


     |i|

     i := (Image fromFile:'bitmaps/PasteButton.tiff') magnifiedBy:10.
     (i asOrderedDitheredMonochromeFormOn:Display) inspect
    "

    "Created: 7.6.1996 / 14:51:42 / cg"
    "Modified: 17.4.1997 / 01:08:24 / cg"
!

asOrderedDitheredMonochromeImage
    "return a dithered monochrome image from the receiver image.
     Uses a 4x4 dithermatrix."

    "/ 4x4 seems a good comprimize between:
    "/    number of grey levels (8x8 is better)
    "/    artifacts             (4x4 is better)
    "/
    "/ could look at the image and decide upon the number of
    "/ distinct colors present. Use 8x8 for high-number,
    "/ 4x4 for small number of colors ...

    ^ self 
        asOrderedDitheredMonochromeImageWithDitherMatrix:(self class orderedDitherMatrixOfSize:4)
        ditherWidth:4

"/    ^ self 
"/        asOrderedDitheredMonochromeImageWithDitherMatrix:(self class orderedDitherMatrixOfSize:8)
"/        ditherWidth:8

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     i asOrderedDitheredMonochromeImage inspect


     |i|

     i := (Image fromFile:'/cdrom/icons/a/a11.ico') magnifiedBy:10.
     i asOrderedDitheredMonochromeImage inspect


     |i|

     i := Image fromFile:'bitmaps/garfield.gif'.
     i asOrderedDitheredMonochromeImage inspect


     |i|

     i := (Image fromFile:'bitmaps/PasteButton.tiff') magnifiedBy:10.
     i asOrderedDitheredMonochromeImage inspect


     |i|

     i := (Image fromFile:'bitmaps/blue-ball.gif') magnifiedBy:1.
     i asOrderedDitheredMonochromeImage inspect


     |i|

     i := Image fromFile:'bitmaps/granite.tiff'.
     i asOrderedDitheredMonochromeImage inspect
    "

    "Created: 7.6.1996 / 15:02:07 / cg"
    "Modified: 10.6.1996 / 11:15:09 / cg"
!

asOrderedDitheredMonochromeImageWithDitherMatrix:ditherMatrix ditherWidth:dW
    "return a dithered monochrome image from the receiver image.
     Uses the passed ditherMatrix and ditherWidth."

    |monoBits|

    monoBits := self orderedDitheredMonochromeBitsWithDitherMatrix:ditherMatrix ditherWidth:dW.
    ^ (Depth1Image width:width height:height fromArray:monoBits) photometric:#blackIs0

    "order-4 dither:

     |i|

     i := Image fromFile:'garfield.gif'.
     i
        asOrderedDitheredMonochromeImageWithDitherMatrix:(Image orderedDitherMatrixOfSize:4)
        ditherWidth:4
    "

    "order-6 dither:

     |i|

     i := Image fromFile:'garfield.gif'.
     i
        asOrderedDitheredMonochromeImageWithDitherMatrix:(Image orderedDitherMatrixOfSize:8)
        ditherWidth:8
    "


    "thresholding at:0.5 (all above 0.5 becomes white):

     |i|

     i := Image fromFile:'garfield.gif'.
     i
        asOrderedDitheredMonochromeImageWithDitherMatrix:(ByteArray new:16 withAll:8)
        ditherWidth:4
    "

    "thresholding at: 0.25 (all above 0.25 becomes white):

     |i|

     i := Image fromFile:'garfield.gif'.
     i
        asOrderedDitheredMonochromeImageWithDitherMatrix:(ByteArray new:16 withAll:3)
        ditherWidth:4
    "

    "thresholding at: 0.75 (all above 0.75 becomes white):

     |i|

     i := Image fromFile:'garfield.gif'.
     i
        asOrderedDitheredMonochromeImageWithDitherMatrix:(ByteArray new:16 withAll:11)
        ditherWidth:4
    "

    "Modified: 7.6.1996 / 17:23:47 / cg"
!

asPseudoFormQuickOn:aDevice
    "return a pseudo-deviceForm from the image.
     Fail if any color is not available (i.e. do not dither)."

    |f d dDev temp temp8 bits clr cMap idMap 
     nClr "{ Class: SmallInteger }"
     usedColors |

    d := self depth.
    (d == 1
    or:[d == 2
    or:[d == 4
    or:[d == 8]]]) ifFalse:[^ nil].

    "/
    "/ see if all of the images colors are representable
    "/ on the device

    bits := self bitsPerPixel.
    nClr := (1 bitShift:bits).
    cMap := Array new:nClr.
    idMap := ByteArray new:nClr.

    photometric == #palette ifTrue:[
        nClr := nClr min:(colorMap size)
    ].

    d == 8 ifTrue:[
        usedColors := bytes usedValues.    "gets us an array filled with used values"
        1 to:nClr do:[:pixel |
            (usedColors includes:(pixel - 1)) ifFalse:[
                clr := Color black
            ] ifTrue:[
                clr := self colorFromValue:pixel-1.
                clr := clr exactOn:aDevice.
                clr isNil ifTrue:[^ nil].
            ].
            cMap at:(pixel) put:clr.
            idMap at:(pixel) put:(clr colorId).
        ].
    ] ifFalse:[
        1 to:nClr do:[:pixel |
            clr := self colorFromValue:pixel-1.
            clr := clr exactOn:aDevice.
            clr isNil ifTrue:[^ nil].

            cMap at:(pixel) put:clr.
            idMap at:(pixel) put:(clr colorId).
        ].
    ].

    "/ got all colors; good - simply change depth & translate pixels

    dDev := aDevice depth.

    (d == 8 and:[dDev == 8]) ifTrue:[
        "/ only translate
        temp := ByteArray uninitializedNew:(width * height).
        bytes expandPixels:8         "xlate only"
                    width:width 
                   height:height
                     into:temp
                  mapping:idMap.
    ] ifFalse:[

        "/ stupid: expandPixels can only handle any-to-8
        "/ compressPixels can only handle 8-to-any

        d ~~ 8 ifTrue:[
            temp8 := ByteArray uninitializedNew:(width * height).

            bytes expandPixels:d      
                         width:width 
                        height:height
                          into:temp8
                       mapping:idMap.
            idMap := nil.
        ] ifFalse:[
            temp8 := bytes.
        ].

        dDev ~~ 8 ifTrue:[
            temp := ByteArray uninitializedNew:((width * dDev + 7) // 8 * height).

            temp8 compressPixels:dDev      
                         width:width 
                        height:height
                          into:temp
                       mapping:idMap.
        ] ifFalse:[
            temp := temp8
        ].
    ].

    f := Form width:width height:height depth:dDev on:aDevice.
    f isNil ifTrue:[^ nil].

    f colorMap:cMap. 
    f initGC.

    aDevice 
        drawBits:temp 
        depth:dDev 
	padding:8
        width:width 
        height:height
        x:0 y:0
        into:(f id) x:0 y:0 
        width:width 
        height:height 
        with:(f gcId).

    ^ f

    "
     (
        (((Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ]))
               magnifiedBy:30
         )
          asPseudoFormQuickOn:Display
      ) inspect
     "

    "Modified: 17.6.1996 / 17:31:49 / cg"
!

asStevensonArceDitheredMonochromeImage
    "return a stevenson-arce dithered monochrome image from the receiver image.
     Depending on the images contents, this may or may not look better than
     a floyd-steinberg dithered image.
     Notice, that floyd-steinberg dithering is faster; both because less
     error diffusion is done and due to being specially tuned."

    |monoBits|

    monoBits := self stevensonArceDitheredMonochromeBits.
    ^ (Depth1Image width:width height:height fromArray:monoBits) photometric:#blackIs0

    "
     |i|

     i := Image fromFile:'garfield.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asStevensonArceDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Image fromFile:'claus.gif'.
     i inspect.
     i asFloydSteinbergDitheredMonochromeImage inspect.
     i asBurkesDitheredMonochromeImage inspect.
     i asStevensonArceDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "
     |i|

     i := Depth4Image
             width:4 
             height:4
             fromArray:#[ 
                            16r01 16r23
                            16r45 16r67
                            16r89 16rab
                            16rcd 16ref 
                        ].
     i := i magnifiedBy:30.
     i inspect.
     i asStevensonArceDitheredMonochromeImage inspect.
     i asOrderedDitheredMonochromeImage inspect.
    "

    "Created: 10.6.1996 / 12:38:29 / cg"
    "Modified: 12.6.1996 / 13:58:24 / cg"
!

asThresholdGrayFormOn:aDevice
    "return a thresholded grey form from the receiver image."

    |depth|

    (depth := aDevice depth) == 1 ifTrue:[
        ^ self asThresholdMonochromeFormOn:aDevice
    ].

    ^ self 
        makeDeviceGrayPixmapOn:aDevice 
        depth:depth 
        fromArray:(self
                        orderedDitheredGrayBitsWithDitherMatrix:#[8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8]
                        ditherWidth:4
                        depth:depth)

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asThresholdGrayFormOn:Display) inspect.
    "

    "Created: 10.6.1996 / 18:38:31 / cg"
    "Modified: 24.6.1997 / 22:19:46 / cg"
!

asThresholdGrayImageDepth:depth
    "return a thresholded depth-x grey image from the receiver image."

    (depth == 1) ifTrue:[
        "/ for monochrome, there is highly specialized
        "/ monochrome dither code available

        ^ Depth1Image
            width:width
            height:height
            fromArray:(
                self
                    orderedDitheredMonochromeBitsWithDitherMatrix:#[8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8]
                    ditherWidth:4)
    ].

    ^ (self class implementorForDepth:depth)
        width:width
        height:height
        fromArray:(
            self
                orderedDitheredGrayBitsWithDitherMatrix:#[8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8]
                ditherWidth:4
                depth:depth)

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     (i asThresholdGrayImageDepth:1) inspect.
     (i asThresholdGrayImageDepth:2) inspect.
     (i asThresholdGrayImageDepth:4) inspect.
     (i asThresholdGrayImageDepth:8) inspect.

     (i asOrderedDitheredGrayImageDepth:1) inspect.
     (i asOrderedDitheredGrayImageDepth:2) inspect.
     (i asOrderedDitheredGrayImageDepth:4) inspect.
     (i asOrderedDitheredGrayImageDepth:8) inspect.

     (i asFloydSteinbergDitheredGrayImageDepth:1) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:2) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:4) inspect.
     (i asFloydSteinbergDitheredGrayImageDepth:8) inspect.
    "

    "Created: 7.6.1996 / 18:13:33 / cg"
    "Modified: 24.6.1997 / 22:19:52 / cg"
!

asThresholdMonochromeFormOn:aDevice
    "return a threshold monochrome form from the image."

    ^ self 
        asOrderedDitheredMonochromeFormOn:aDevice 
        ditherMatrix:#[8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8]
        ditherWidth:4

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     i inspect.
     (i asThresholdMonochromeFormOn:Display) inspect


     |i|

     i := (Image fromFile:'bitmaps/a11.ico') magnifiedBy:10.
     i inspect.
     (i asThresholdMonochromeFormOn:Display) inspect


     |i|

     i := Image fromFile:'bitmaps/granite.tiff'.
     i inspect.
     (i asThresholdMonochromeFormOn:Display) inspect


     |i|

     i := (Image fromFile:'bitmaps/PasteButton.tiff') magnifiedBy:10.
     i inspect.
     (i asThresholdMonochromeFormOn:Display) inspect
    "

    "Created: 7.6.1996 / 16:15:17 / cg"
    "Modified: 10.6.1996 / 17:44:29 / cg"
!

asThresholdMonochromeImage
    "return a threshold monochrome image from the image.
     Threshold means: brightness < 0.5 -> black / otherwise white"

    ^ self
        asOrderedDitheredMonochromeImageWithDitherMatrix:#[8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8]
        ditherWidth:4

    "
     |i i2|

     i := Image fromFile:'bitmaps/claus.gif'.
     i2 := i asThresholdMonochromeImage


     |i i2|

     i := (Image fromFile:'/cdrom/icons/a/a11.ico') magnifiedBy:10.
     i2 := i asThresholdMonochromeImage


     |i i2|

     i := Image fromFile:'bitmaps/granite.tiff'.
     i2 := i asThresholdMonochromeImage


     |i i2|

     i := (Image fromFile:'bitmaps/PasteButton.tiff') magnifiedBy:10.
     i2 := i asThresholdMonochromeImage
    "

    "Created: 7.6.1996 / 16:15:17 / cg"
    "Modified: 7.6.1996 / 17:16:10 / cg"
!

asThresholdMonochromeImage:thresholdBrighness
    "return a threshold monochrome image from the image.
     The argument (0..1) gives the threshold value;
     Threshold means: brightness < threshold -> black / otherwise white"

    |value ditherMatrix|

    value := thresholdBrighness * 255.
    ditherMatrix := ByteArray new:256 withAll:(value truncated).

    ^ self
        asOrderedDitheredMonochromeImageWithDitherMatrix:ditherMatrix
        ditherWidth:16

    "
     |i|

     i := Image width:4 height:4 depth:4
                fromArray:#[ 16r01 16r23
                             16r45 16r67
                             16r89 16rab
                             16rcd 16ref ].
     i := i magnifiedBy:30.
     i inspect.
     (i asThresholdMonochromeImage:0.125) inspect.
     (i asThresholdMonochromeImage:0.25) inspect.
     (i asThresholdMonochromeImage:0.5) inspect.
     (i asThresholdMonochromeImage:0.75) inspect.
     (i asThresholdMonochromeImage:1) inspect.
    "

    "
     |i|

     i := Image fromFile:'bitmaps/claus.gif'.
     i inspect.
     (i asThresholdMonochromeImage:0.125) inspect.
     (i asThresholdMonochromeImage:0.25) inspect.
     (i asThresholdMonochromeImage:0.5) inspect.
     (i asThresholdMonochromeImage:0.625) inspect.
     (i asThresholdMonochromeImage:0.75) inspect.
    "

    "Created: 7.6.1996 / 16:15:17 / cg"
    "Modified: 8.6.1996 / 14:51:57 / cg"
!

clearMaskedPixels
    "assuming that I already have a device representation
     in deviceForm, clear any masked pixels.
     This will allow faster drawing in the future."

    maskedPixelsAre0 == true ifTrue:[^ self].   "/ already cleared
    mask isNil ifTrue:[^ self].         "/ no mask
    deviceForm isNil ifTrue:[^ self].   "/ no device rep.
    mask depth ~~ 1 ifTrue:[^ self].    "/ not done with alpha masks

    deviceForm foreground:Color allColor background:Color noColor.
    deviceForm function:#and.
    deviceForm copyPlaneFrom:(mask asFormOn:device) x:0 y:0 toX:0 y:0 width:width height:height.
    maskedPixelsAre0 := true.

    "Created: 12.4.1997 / 12:18:05 / cg"
    "Modified: 12.4.1997 / 12:20:19 / cg"
!

fromForm:aForm
    "setup the receiver from a form"

    |map c0 c1|

    width := aForm width.
    height := aForm height.
    bitsPerSample := self bitsPerSample.
    samplesPerPixel := self samplesPerPixel.

    aForm hasBits ifFalse:[
        "/ must read the data from the device
        self from:aForm in:(0@0 extent:aForm extent).
        ^ self
    ].

    "/ the form has all data available

    bytes := aForm bits.
    map := aForm colorMap.

    aForm depth == 1 ifTrue:[
        map isNil ifTrue:[
            photometric := #whiteIs0
        ] ifFalse:[
            c0 := map at:1.
            c1 := map at:2.
            ((c0 = Color white)
            and:[c1 = Color black]) ifTrue:[
                photometric := #whiteIs0
            ] ifFalse:[
                ((c0 = Color black)
                and:[c1 = Color white]) ifTrue:[
                    photometric := #blackIs0
                ] ifFalse:[
                    photometric := #palette.
                    colorMap := Array with:c0 with:c1.
                ]
            ]
        ]
    ] ifFalse:[
        map notNil ifTrue:[
            photometric := #palette.
            colorMap := map copy.
        ] ifFalse:[
            "
             photometric stays at default
             (which is rgb for d24, greyscale for others)
            "
        ]
    ].

    "Modified: 5.7.1996 / 16:24:31 / cg"
!

fromImage:anImage
    "setup the receiver from another image.
     Color precision may be lost, if conversion is from a higher depth
     image. 
     WARNING:
     This implementation is a slow fallback (the loop over the
     source pixels is very slow). If this method is used heavily, you
     may want to redefine it in concrete subclasses for common source images."

    |map clr mappedRowPixels 
     h "{ Class: SmallInteger }"
     w "{ Class: SmallInteger }"|

    width := anImage width.
    height := anImage height.
    bitsPerSample := self bitsPerSample.
    samplesPerPixel := self samplesPerPixel.
    self colormapFromImage:anImage.

    self mask:anImage mask.

    self depth = anImage depth ifTrue:[
        bytes := anImage bits copy.
        ^ self.
    ].

    bytes := ByteArray uninitializedNew:(self bytesPerRow * height).

    self depth >= anImage depth ifTrue:[
        anImage depth <= 12 ifTrue:[

            "/ if my depth is greater, all colors can be represented,
            "/ and the loop can be done over pixel values ...

            colorMap isNil ifTrue:[
                map := Array new:(1 bitShift:anImage depth).
                1 to:map size do:[:i |
                    clr := anImage colorFromValue:(i - 1).
                    map at:i put:(self valueFromColor:clr).
                ].
            ].
            mappedRowPixels := self pixelArraySpecies new:width.
            h := height-1.
            w := width.
            0 to:h do:[:row |
                anImage rowAt:row into:mappedRowPixels startingAt:1.
                map notNil ifTrue:[
                    1 to:w do:[:i |
                        mappedRowPixels at:i put:(map at:(mappedRowPixels at:i)+1)
                    ].
                ].
                self rowAt:row putAll:mappedRowPixels
            ].
            ^ self
        ].
    ].

    anImage colorsFromX:0 y:0 toX:(width-1) y:(height-1) do:[:x :y :clr |
        self colorAtX:x y:y put:clr
    ].

    "
     |i i2 i4 i8 i16 i24|

     i := Image fromFile:'bitmaps/SBrowser.xbm'.
     i inspect.
     i2 := Depth2Image fromImage:i.
     i2 inspect.
     i4 := Depth4Image fromImage:i.
     i4 inspect.
     i8 := Depth8Image fromImage:i.
     i8 inspect.
     i24 := Depth24Image fromImage:i.
     i24 inspect.
    "

    "
     |i i24|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     Transcript showCR:(
        Time millisecondsToRun:[
            i24 := Depth24Image fromImage:i.
        ]
     ).
     i24 inspect.
    "
    "
     |i i24|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     MessageTally spyOn:[
        i24 := Depth24Image fromImage:i.
     ]
    "

    "Modified: 23.6.1997 / 19:38:59 / cg"
!

fromSubImage:anImage in:aRectangle
    "setup the receiver from another image, extracting a rectangular
     area. Color precision may be lost, if conversion is from a higher depth
     image. For palette & rgb images, this may fail if a color cannot be
     represented.

     WARNING:
       This implementation is a slow fallback (the loop over the
       source pixels is very slow). If this method is used heavily, you
       may want to redefine it in concrete subclasses for the common case of
       of creating a subImage with the same depth & palette."

    |x0 y0|

    width := aRectangle width.
    height := aRectangle height.
    bytes := ByteArray uninitializedNew:(self bytesPerRow * height).
    bitsPerSample := self bitsPerSample.
    samplesPerPixel := self samplesPerPixel.
    self colormapFromImage:anImage.
    x0 := aRectangle left.
    y0 := aRectangle top.

    ((photometric == anImage photometric)
    and:[self bitsPerPixel = anImage bitsPerPixel
    and:[colorMap = anImage colorMap]]) ifTrue:[
        "/ can do it by value
        anImage valuesFromX:x0  y:y0 
                        toX:(x0 + aRectangle width - 1) y:(y0 + aRectangle height - 1)
                         do:[:x :y :pixelValue |
            self pixelAtX:x-x0 y:y-y0 put:pixelValue.
        ]
    ] ifFalse:[
        "/ must do it by colors
        anImage colorsFromX:x0  y:y0 
                        toX:(x0 + aRectangle width - 1) y:(y0 + aRectangle height - 1)
                         do:[:x :y :clr |
            self colorAtX:x-x0 y:y-y0 put:clr.
        ]
    ].

    anImage mask notNil
    ifTrue:
    [
        mask := anImage mask class new fromSubImage:anImage mask in:aRectangle
    ]

    "
     |i i2 i4 i8 i16 i24|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     i inspect.
     i4 := Depth4Image fromSubImage:i in:(300@160 corner:340@200).
     i4 inspect.
     i8 := Depth8Image fromSubImage:i in:(300@160 corner:340@200).
     i8 inspect.
     i24 := Depth24Image fromSubImage:i in:(300@160 corner:340@200).
     i24 inspect.
    "

    "Created: 20.9.1995 / 01:06:02 / claus"
    "Modified: 20.9.1995 / 10:15:37 / claus"
    "Modified: 24.4.1997 / 17:25:48 / cg"
!

monochromeOn:aDevice
    "return a monochrome device image of the receiver for aDevice.
     (monochrome, even if device supports colors)"

    ((aDevice == device) and:[monoDeviceForm notNil]) ifTrue:[^ self].
    (device notNil and:[aDevice ~~ device]) ifTrue:[
        "oops, I am already accociated to another device
         - need a copy ...
        "
        ^ self copy monochromeOn:aDevice
    ].
    monoDeviceForm := self asMonochromeFormOn:aDevice.
!

on:aDevice
    "return an image with the same pixels as the receiver, but
     associated to aDevice. If the receiver is not yet bound to
     a device, this will be the receiver. Otherwise, a new image
     is returned."

    ((aDevice == device) and:[deviceForm notNil]) ifTrue:[^ self].

    (device notNil and:[aDevice ~~ device]) ifTrue:[
        "oops, I am already accociated to another device
         - need a copy ...
        "
        ^ self copy on:aDevice
    ].
    deviceForm := self asFormOn:aDevice.
    device := aDevice.
    Lobby register:self

    "Modified: 16.9.1996 / 21:02:33 / cg"
!

onDevice:aDevice
    "return an image with the same pixels as the receiver, but
     associated to aDevice. If the receiver is not yet bound to
     a device, this will be the receiver. Otherwise, a new image
     is returned."

    ^ self on:aDevice

    "Created: 28.3.1997 / 16:05:53 / cg"
! !

!Image methodsFor:'converting greyscale images'!

greyImageAsFormOn:aDevice
    "return a grey-deviceForm from the grey image."

    |pictureDepth nPlanes|


    nPlanes := samplesPerPixel.
    (nPlanes == 2) ifTrue:[
        'Image [info]: alpha plane ignored' infoPrintCR.
        nPlanes := 1
    ].

    pictureDepth := bitsPerSample at:1.

    "monochrome is very easy ..."

    (pictureDepth == 1) ifTrue:[
        ^ Form width:width height:height fromArray:bytes on:aDevice
    ].

    (aDevice visualType == #TrueColor) ifTrue:[
        ^ self greyImageAsTrueColorFormOn:aDevice
    ].

    "/ PseudoColor conversion also works for StaticColor,
    "/ GrayScale and DirectColor; although possibly with suboptimal results

    ^ self greyImageAsPseudoFormOn:aDevice

    "Modified: 19.10.1997 / 05:17:25 / cg"
!

greyImageAsPseudoFormOn:aDevice
    "return an 8-bit pseudo Form from the grey image"

    |wideBits pictureDepth f map  
     colorMap usedColors nUsed aColor 
     nColors "{ Class: SmallInteger }"
     range id
     cube nR nG nB grayColors
     fit|

    pictureDepth := bitsPerSample at:1.

    (#(2 4 8) includes:pictureDepth) ifFalse:[
        self error:'currently only depth-2, 4 or 8 supported'.
        ^ nil
    ].

    wideBits := ByteArray uninitializedNew:(width * height).

    (pictureDepth == 8) ifTrue:[
        "for 8bits, we scan for used colors first;
         to avoid allocating too many colors"

        (grayColors := aDevice fixGrayColors) notNil ifTrue:[
            DitherAlgorithm == #floydSteinberg ifTrue:[
                f := self
                       asFloydSteinbergDitheredDepth8FormOn:aDevice 
                       colors:grayColors 
            ].
            f notNil ifTrue:[^ f].
        ].

        (cube := aDevice fixColors) notNil ifTrue:[
            nR := aDevice numFixRed.
            nG := aDevice numFixGreen.
            nB := aDevice numFixBlue.

            DitherAlgorithm == #floydSteinberg ifTrue:[
                f := self
                       asFloydSteinbergDitheredDepth8FormOn:aDevice 
                       colors:cube 
                       nRed:nR
                       nGreen:nG
                       nBlue:nB.
            ] ifFalse:[
                f := self
                       asNearestPaintDepth8FormOn:aDevice 
                       colors:cube 
                       nRed:nR
                       nGreen:nG
                       nBlue:nB.
            ].
            f notNil ifTrue:[^ f].
        ].

        usedColors := bytes usedValues.
        nUsed := usedColors max + 1.

        colorMap := Array new:nUsed.
        range := 100 / 255.0.
        usedColors do:[:grey |
            colorMap at:(grey + 1) put:(Color gray:(range * grey))
        ].
    ] ifFalse:[
        nColors := (1 bitShift:pictureDepth).
        colorMap := Array new:nColors.
        range := 100 / (nColors - 1) asFloat.
        1 to:nColors do:[:i |
            colorMap at:i put:(Color gray:(i - 1) * range).
        ].
    ].
    photometric ~~ #blackIs0 ifTrue:[
        colorMap reverse
    ].

    "allocate those colors & setup the translation map"

    fit := true.
    map := ByteArray uninitializedNew:256.
    nColors := colorMap size.
    1 to:nColors do:[:i |
        aColor := colorMap at:i.
        aColor notNil ifTrue:[
            aColor := aColor on:aDevice.
            colorMap at:i put:aColor.
            id := aColor colorId.
            id isNil ifTrue:[
                id := 0.
                fit := false.
            ].
            map at:i put:id
        ]
    ].

    fit ifFalse:[
        "/ here comes the hard part - some grey value
        "/ could not be allocated.
        "/ Must dither.

    ].

    "expand & translate"
    bytes expandPixels:pictureDepth
                width:width 
               height:height
                 into:wideBits
              mapping:map.

    f := Form width:width height:height depth:8 on:aDevice.
    f isNil ifTrue:[^ nil].
    f colorMap:colorMap. 
    f initGC.
    aDevice 
	drawBits:wideBits 
	depth:8 
	padding:8
	width:width height:height
        x:0 y:0
        into:(f id) 
	x:0 y:0 
	width:width height:height 
	with:(f gcId).
    ^ f

    "Modified: 19.10.1997 / 05:19:44 / cg"
!

greyImageAsTrueColorFormOn:aDevice
    "return a true-color device-form for the grey-image receiver.
     TODO: the pixel loops ought to be implemented as inline primitive code ..."

    |depth  
     myDepth    "{ Class: SmallInteger }"
     nColors    "{ Class: SmallInteger }"
     colorValues
     scaleDown 
     scaleRed   "{ Class: SmallInteger }"
     scaleGreen "{ Class: SmallInteger }"
     scaleBlue  "{ Class: SmallInteger }"
     redShift   "{ Class: SmallInteger }"
     blueShift  "{ Class: SmallInteger }"
     greenShift "{ Class: SmallInteger }"
     form bestFormat usedDeviceDepth usedDeviceBitsPerPixel imageBits 
     greyValue  "{ Class: SmallInteger }"
     h          "{ Class: SmallInteger }"
     w          "{ Class: SmallInteger }"
     pixelArray newPixelArray i|

    "/ this is a slow fallback method; this ought to be
    "/ redefined in DepthxImage for more performance.

    depth := aDevice depth.
    myDepth := self depth.
    myDepth > 12 ifTrue:[
        self error:'unsupported trueColor depth in greyImageAsTrueColorFormOn:'.
        ^ nil
    ].

    "/ compute scale to map from my pixels into devices range

    scaleDown := 1 bitShift:myDepth.
    scaleRed := (1 bitShift:aDevice bitsRed).
    scaleGreen := (1 bitShift:aDevice bitsGreen).
    scaleBlue := (1 bitShift:aDevice bitsBlue).
    redShift := aDevice shiftRed.
    greenShift := aDevice shiftGreen.
    blueShift := aDevice shiftBlue.

    "/ prepare the map

    nColors := (1 bitShift:myDepth).
    colorValues := Array new:nColors.
    1 to:nColors do:[:i |
        |v gv bv rv nv|

        "/ scale down to 0..1
        v := (i-1) / scaleDown.
        rv := (v * scaleRed) truncated.
        gv := (v * scaleGreen) truncated.
        bv := (v * scaleBlue) truncated.
        nv := rv bitShift:redShift.
        nv := nv bitOr:(gv bitShift:greenShift).
        nv := nv bitOr:(bv bitShift:blueShift).
        colorValues at:i put:nv
    ].
    photometric == #whiteIs0 ifTrue:[
        "/ reverse the order; 0 is brightest
        colorValues reverse
    ].

    bestFormat := self bestSupportedImageFormatFor:aDevice.
    usedDeviceDepth := bestFormat at:#depth.
    usedDeviceBitsPerPixel := bestFormat at:#bitsPerPixel.

    "/ the temporary helper image is only needed to allow
    "/ the rowAt:putAll: calls below.

    i := (Image implementorForDepth:usedDeviceBitsPerPixel) new.
    i width:width height:height.
    imageBits := ByteArray new:(i bytesPerRow * height).
    i bits:imageBits.

    "/ now, walk over the image and replace
    "/ colorMap indices by color values in the bits array

    h := height - 1.
    w := width - 1.
    pixelArray := self pixelArraySpecies new:width.
    newPixelArray := i pixelArraySpecies new:width.

    0 to:h do:[:y |
        self rowAt:y into:pixelArray.
        0 to:w do:[:x |
            greyValue := pixelArray at:(x+1).
            newPixelArray at:(x+1) put:(colorValues at:greyValue + 1).
        ].
        i rowAt:y putAll:newPixelArray.
    ].

    form := Form width:width height:height depth:usedDeviceDepth on:aDevice.
    form isNil ifTrue:[^ nil].
    form initGC.

    form 
        copyBitsFrom:imageBits 
	bitsPerPixel:usedDeviceBitsPerPixel 
	depth:usedDeviceDepth 
	padding:8
        width:width height:height 
        x:0 y:0 
	toX:0 y:0. 

    ^ form

    "Created: 20.10.1995 / 22:05:10 / cg"
    "Modified: 24.4.1997 / 23:11:16 / cg"
! !

!Image methodsFor:'converting palette images'!

paletteImageAsFormOn:aDevice
    "return a device-form for the palette-image receiver"

    |type ddepth|

    ddepth := aDevice depth.

    aDevice hasGrayscales not ifTrue:[
        ^ self asMonochromeFormOn:aDevice
    ].

    ((type := aDevice visualType) == #StaticGray) ifTrue:[
        ddepth == 8 ifTrue:[
            ^ self paletteImageAsGray8FormOn:aDevice
        ].
        ^ self asGrayFormOn:aDevice
    ].

    (type == #TrueColor) ifTrue:[
        DitherAlgorithm == #floydSteinberg ifTrue:[
            ddepth == 8 ifTrue:[
                self depth == 8 ifTrue:[
                    "/ use fixColor dither algorithm
                    ^ self asDitheredTrueColor8FormOn:aDevice
                ]
            ]
        ].

        ^ self paletteImageAsTrueColorFormOn:aDevice
    ].

    "/ the PseudoColor conversion also works for
    "/ StaticColor, GrayScale & DirectColor; although possibly with suboptimal results

    ^ self paletteImageAsPseudoFormOn:aDevice

    "Modified: 14.6.1996 / 19:31:01 / cg"
!

paletteImageAsPseudoFormOn:aDevice
    "return a pseudo-deviceForm from the palette image."

    |tempImage d temp8|

    d := self depth.
    (#(1 2 4 8) includes:d) ifTrue:[ 
        "
         fallback code for some depth's:
         create a temporary Depth8Image and use its conversion method
        "
        temp8 := ByteArray uninitializedNew:(width * height).

        bytes expandPixels:d      
                     width:width 
                   height:height
                     into:temp8
                  mapping:nil.

        tempImage := Image width:width height:height depth:8 fromArray:temp8.
        tempImage colorMap:colorMap.
        ^ tempImage paletteImageAsPseudoFormOn:aDevice
    ].
    ^ self subclassResponsibility
!

paletteImageAsTrueColorFormOn:aDevice
    "return a true-color device-form for the palette-image receiver."

    |depth 
     nColors "{ Class: SmallInteger }"
     w       "{ Class: SmallInteger }"
     h       "{ Class: SmallInteger }"
     pixel   "{ Class: SmallInteger }"
     colorValues 
     scaleRed scaleGreen scaleBlue redShift greenShift blueShift
     form imageBits bestFormat usedDeviceDepth usedDeviceBitsPerPixel 
     i pixelArray newPixelArray|

    "/ this is a slow fallback method; this ought to be
    "/ redefined in DepthxImage for more performance.

    depth := aDevice depth.

    "/ gather r/g/b values for all colors in the map ...

    nColors := colorMap size.

    "/ precompute scales to map from 0..100 into devices range
    "/ (this may be different for the individual components)

    scaleRed := ((1 bitShift:aDevice bitsRed) - 1) / 100.
    scaleGreen := ((1 bitShift:aDevice bitsGreen) - 1) / 100.
    scaleBlue := ((1 bitShift:aDevice bitsBlue) - 1) / 100.
    redShift := aDevice shiftRed.
    greenShift := aDevice shiftGreen.
    blueShift := aDevice shiftBlue.

    colorValues := Array uninitializedNew:nColors.

    1 to:nColors do:[:index |
        |clr rv gv bv v "{ Class: SmallInteger }" |

        clr := colorMap at:index.
        clr notNil ifTrue:[
            rv := (clr red * scaleRed) rounded.
            gv := (clr green * scaleGreen) rounded.
            bv := (clr blue * scaleBlue) rounded.

            v := rv bitShift:redShift.
            v := v bitOr:(gv bitShift:greenShift).
            v := v bitOr:(bv bitShift:blueShift).
            colorValues at:index put:v.
"/ clr print. ' ' print.
"/ rv print. ' ' print. gv print. ' ' print. bv print. ' ' print.
"/ ' -> ' print. v printNL.

        ]
    ].

    bestFormat := self bestSupportedImageFormatFor:aDevice.
    usedDeviceDepth := bestFormat at:#depth.
    usedDeviceBitsPerPixel := bestFormat at:#bitsPerPixel.

    "/ the temporary helper image is only needed to allow
    "/ the rowAt:putAll: calls below.

    i := (Image implementorForDepth:usedDeviceBitsPerPixel) new.
    i width:width height:height.
    imageBits := ByteArray new:(i bytesPerRow * height).
    i bits:imageBits.

    "/ now, walk over the image and replace
    "/ colorMap indices by color values in the bits array

    h := height - 1.
    w := width - 1.
    pixelArray := self pixelArraySpecies new:width.
    newPixelArray := i pixelArraySpecies new:width.

    0 to:h do:[:y |
        self rowAt:y into:pixelArray.
        0 to:w do:[:x |
            pixel := pixelArray at:(x+1).
            newPixelArray at:(x+1) put:(colorValues at:pixel + 1).
        ].
        i rowAt:y putAll:newPixelArray.
    ].

    form := Form width:width height:height depth:usedDeviceDepth on:aDevice.
    form isNil ifTrue:[^ nil].
    form initGC.

    form 
        copyBitsFrom:imageBits 
	bitsPerPixel:usedDeviceBitsPerPixel 
	depth:usedDeviceDepth 
	padding:8
        width:width height:height 
        x:0 y:0 
	toX:0 y:0. 

    ^ form

    "Modified: 24.4.1997 / 20:43:34 / cg"
! !

!Image methodsFor:'converting rgb images'!

asDitheredTrueColor8FormOn:aDevice
    "convert an rgb image to a dithered depth8-form on aDevice.
     Return the device-form. 
     This method is only valid for trueColor displays."

    |fixColors pixel
     dstIdx     "{ Class: SmallInteger }"
     shiftRed   "{ Class: SmallInteger }"
     shiftGreen "{ Class: SmallInteger }"
     shiftBlue  "{ Class: SmallInteger }"
     nRed    "{ Class: SmallInteger }"
     nGreen  "{ Class: SmallInteger }"
     nBlue   "{ Class: SmallInteger }"|

    shiftRed := aDevice shiftRed.
    shiftGreen := aDevice shiftGreen.
    shiftBlue := aDevice shiftBlue.

    nRed := (1 bitShift:aDevice bitsRed).
    nGreen := (1 bitShift:aDevice bitsGreen).
    nBlue := (1 bitShift:aDevice bitsBlue).

    fixColors := Array new:(nRed * nGreen * nBlue).

    dstIdx := 1.
    0 to:nRed - 1 do:[:sR |
        0 to:nGreen - 1 do:[:sG |
            0 to:nBlue - 1 do:[:sB |
                pixel := (sR bitShift:shiftRed)
                         + (sG bitShift:shiftGreen)
                         + (sB bitShift:shiftBlue).
                fixColors at:dstIdx put:(Color colorId:(pixel)).
                dstIdx := dstIdx + 1
            ]
        ]
    ].

    ^ self
        asFloydSteinbergDitheredDepth8FormOn:aDevice 
        colors:fixColors 
        nRed:nRed nGreen:nGreen nBlue:nBlue

    "Created: 14.6.1996 / 17:23:52 / cg"
    "Modified: 23.6.1997 / 15:22:36 / cg"
!

rgbImageAsFormOn:aDevice
    "convert am rgb image to a device-form on aDevice.
     Return the device-form."

    |visual|

    visual := aDevice visualType.
    (visual == #StaticGray) ifTrue:[
        ^ self asGrayFormOn:aDevice
    ].
    (visual == #TrueColor) ifTrue:[
        ^ self rgbImageAsTrueColorFormOn:aDevice
    ].

    "/ PseudoColor conversion also works for StaticColor, GrayScale
    "/ and DirectColor displays; although possibly with suboptimal results

    ^ self rgbImageAsPseudoFormOn:aDevice

    "Modified: 14.6.1996 / 19:30:06 / cg"
!

rgbImageAsPseudoFormOn:aDevice
    "return a pseudocolor form from the rgb-picture"

    |n     "{ Class: SmallInteger }"
     depth "{ Class: SmallInteger }"
     palette f|

    (depth := self depth) <= 8 ifTrue:[
        "/ simulate it via a temporary palette image

        palette := Array new:(1 bitShift:depth).

        n := (1 bitShift:depth)-1.
        0 to:n do:[:pixelValue |
            palette at:(pixelValue+1) put:(self colorFromValue:pixelValue)
        ].

        colorMap := palette.
        photometric := #palette.
        f := self paletteImageAsPseudoFormOn:aDevice.
        colorMap := nil.
        photometric := #rgb.
        ^ f
    ].

    ^ self subclassResponsibility

    "Modified: 8.6.1996 / 10:58:25 / cg"
!

rgbImageAsTrueColorFormOn:aDevice
    "return a truecolor form from the rgb-picture."

    |bestFormat usedDeviceDepth usedDeviceBitsPerPixel
     form|

    bestFormat := self bestSupportedImageFormatFor:aDevice.
    usedDeviceDepth := bestFormat at:#depth.
    usedDeviceBitsPerPixel := bestFormat at:#bitsPerPixel.

    "/ kludge for some 15bit XFree servers (should return 16 here - does not)
    usedDeviceBitsPerPixel == 15 ifTrue:[
        usedDeviceBitsPerPixel := 16
    ].

    "/
    "/ only the trivial case, where the depths match
    "/ is handled here; conversions are supposed to
    "/ be done in concrete subclasses (see D24Image)
    "/
    self bitsPerPixel == usedDeviceBitsPerPixel ifFalse:[
        'Image [warning]: unimplemented trueColor depth in rgbImageAsTrueColorFormOn: ' errorPrint. self bitsPerPixel errorPrintCR.
        ^ self asMonochromeFormOn:aDevice
    ].

    form := Form width:width height:height depth:usedDeviceDepth on:aDevice.
    form isNil ifTrue:[
        'Image [warning]: display bitmap creation failed' errorPrintCR.
        ^ nil
    ].
    form initGC.

    form 
        copyBitsFrom:bytes 
	bitsPerPixel:usedDeviceBitsPerPixel 
	depth:usedDeviceDepth 
	padding:8
        width:width height:height 
        x:0 y:0 
	toX:0 y:0. 

    ^ form

    "Modified: 10.1.1997 / 17:53:23 / cg"
! !

!Image methodsFor:'copying'!

postCopy
    "redefined to also copy the pixels and the colorMap
     and clear out any device handles in the copy."

    bytes := bytes copy.
    colorMap := colorMap copy.
    device := deviceForm := monoDeviceForm := fullColorDeviceForm := nil.
    mask := mask copy.
! !

!Image methodsFor:'displaying'!

asImage
    "ST-80 compatibility
    "
    ^ self
!

displayFilledOn:aGC
    "display the receiver as an opaque image.
     This allows Images to be wrapped by a FillingWrapper"

    aGC displayOpaqueForm:self x:0 y:0.

    "Created: 29.5.1996 / 10:34:18 / cg"
    "Modified: 29.5.1996 / 10:52:36 / cg"
!

displayOn:aGC at:aPoint
    "draw the receiver in the graphicsContext, aGC.
     Smalltalk-80 compatibility"

    self displayOn:aGC x:aPoint x y:aPoint y.

    "Modified: 12.5.1996 / 20:16:38 / cg"
!

displayOn:aGC x:x y:y
    "draw the receiver in the graphicsContext, aGC.
     Smalltalk-80 compatibility"

    aGC displayForm:self x:x y:y.

    "Modified: 23.4.1996 / 11:12:31 / cg"
    "Created: 12.5.1996 / 20:14:31 / cg"
!

displayOpaqueOn:aGC x:x y:y
    "draw the receiver in the graphicsContext, aGC.
     Smalltalk-80 compatibility"

    aGC displayOpaqueForm:self x:x y:y.

    "Modified: 23.4.1996 / 11:12:31 / cg"
    "Created: 22.10.1996 / 16:35:49 / cg"
!

displayStrokedOn:aGC
    "display the receiver as an non opaque image.
     This allows Images to be wrapped by a StrokingWrapper"

    aGC displayForm:self x:0 y:0.

    "Created: 29.5.1996 / 10:34:18 / cg"
    "Modified: 29.5.1996 / 10:52:30 / cg"
! !

!Image methodsFor:'dither helpers'!

burkesDitheredMonochromeBits
    "return the bitmap for a dithered monochrome bitmap from the image.
     Works for any source depths / photometric"

    |dstIndex        "{Class: SmallInteger }"
     nextDst         "{Class: SmallInteger }"
     bytesPerMonoRow "{Class: SmallInteger }"
     monoBits greyValues
     errorArray
     errorArray1
     e t
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" 
     grey|

    self depth > 12 ifTrue:[
        ^ self floydSteinbergDitheredMonochromeBits
    ].

    w := width.
    h := height.

    bytesPerMonoRow := (w + 7) // 8.
    monoBits := ByteArray uninitializedNew:(bytesPerMonoRow * h).
    (monoBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    errorArray := Array new:(w+4).
    errorArray1 := Array new:(w+4) withAll:0.

    dstIndex := 1.

    "/ fetch scaled brightness values outside of loop into a table;
    "/ use table-value in loop

    greyValues := self greyMapForRange:(255*1024).
    
    0 to:(h-1) do:[:y |
        nextDst := dstIndex + bytesPerMonoRow.
        byte := 0.
        bitCnt := 8.

        t := errorArray.
        errorArray := errorArray1.
        errorArray1 := t.

        errorArray1 atAllPut:0.

        self valuesAtY:y from:0 to:(w-1) do:[:x :pixel |
            |eP "{Class: SmallInteger }" 
             eD 
             eI "{Class: SmallInteger }" 
             xE "{Class: SmallInteger }"
             xN "{Class: SmallInteger }" |

            "/ get the colors grey value [0 .. 1]
            grey := greyValues at:(pixel + 1).

            "/ adjust error
            xE := x + 2 + 1.
            grey := (grey + (errorArray at:xE)).

            byte := byte bitShift:1.
            grey > (127*1024) ifTrue:[
                byte := byte bitOr:1.      "/ white
                e := grey - (255*1024)
            ] ifFalse:[
                e := grey                  "/ black
            ].

            e ~= 0 ifTrue:[
                "/ distribute the error:
                "/                  XX  8  4
                "/             2  4  8  4  2

                eD := e.
                eI := e // 32.

                eP := eI * 8. eD := eD - eP.

                xN := xE + 1.
                errorArray at:xN put:(errorArray at:xN) + eP.

                eD := eD - eP.
                errorArray1 at:xE put:(errorArray1 at:xE) + eP.

                eP := eI * 4. eD := eD - eP.
                xN := xE + 2.
                errorArray at:xN put:(errorArray at:xN) + eP.

                eD := eD - eP.
                xN := xE - 1.
                errorArray1 at:xN put:(errorArray1 at:xN) + eP.

                eD := eD - eP.
                xN := xE + 1.
                errorArray1 at:xN put:(errorArray1 at:xN) + eP.

                eP := eI * 2. eD := eD - eP.
                xN := xE - 2.
                errorArray1 at:xN put:(errorArray1 at:xN) + eP.

                eD := eD.
                xN := xE + 2.
                errorArray1 at:xN put:(errorArray1 at:xN) + eP.
            ].

            bitCnt := bitCnt - 1.
            bitCnt == 0 ifTrue:[
                monoBits at:dstIndex put:byte.
                dstIndex := dstIndex + 1.
                byte := 0.
                bitCnt := 8.
            ].

        ].
        bitCnt ~~ 8 ifTrue:[
            byte := byte bitShift:bitCnt.
            monoBits at:dstIndex put:byte.
        ].

        dstIndex := nextDst.
    ].

    ^ monoBits

    "Created: 10.6.1996 / 12:18:20 / cg"
    "Modified: 10.6.1996 / 13:16:33 / cg"
!

floydSteinbergDitheredDepth8BitsColors:colors
    "return a floyd-steinberg dithered bitmap from the receiver picture,
     which must be a depth-8 image. 
     This method expects an array of colors to be used for dithering
     (which need not be a colorCubes colors)."

    |pseudoBits  
     rgbBytes 
     ditherRGBBytes ditherColors
     w       "{Class: SmallInteger }"
     h       "{Class: SmallInteger }"
     index   "{Class: SmallInteger }"
     fixR    "{Class: SmallInteger }"
     fixG    "{Class: SmallInteger }"
     fixB    "{Class: SmallInteger }"
     ditherIds failed map lastColor colorsByDistance qScramble
     clrLookup lookupPos
     error clr|

    self depth ~~ 8 ifTrue:[^ nil].

    "/
    "/ collect color components as integer values (for integer arithmetic)
    "/
    rgbBytes := ByteArray uninitializedNew:256 * 3.

    photometric == #palette ifTrue:[
        lastColor := colorMap size - 1
    ] ifFalse:[
        lastColor := 255.
    ].
    index := 1.
    0 to:lastColor do:[:pix |
        clr := self colorFromValue:pix.
        rgbBytes at:index put:(clr redByte).
        rgbBytes at:index+1 put:(clr greenByte).
        rgbBytes at:index+2 put:(clr blueByte).

        index := index + 3.
    ].

    "/ collect valid ditherColors ...

    ditherColors := colors select:[:clr | clr notNil].

    "/ ... and sort by manhatten distance from black

    qScramble := #(
                "/  2rX00X00X00X00

                    2r000000000000    "/ 0
                    2r000000000100    "/ 1
                    2r000000100000    "/ 2
                    2r000000100100    "/ 3
                    2r000100000000    "/ 4
                    2r000100000100    "/ 5
                    2r000100100000    "/ 6
                    2r000100100100    "/ 7
                    2r100000000000    "/ 8
                    2r100000000100    "/ 9
                    2r100000100000    "/ a
                    2r100000100100    "/ b
                    2r100100000000    "/ c
                    2r100100000100    "/ d
                    2r100100100000    "/ e
                    2r100100100100    "/ f
                  ).

    ditherColors := ditherColors sort:[:a :b |
                                |cr "{Class: SmallInteger }"
                                 cg "{Class: SmallInteger }"
                                 cb "{Class: SmallInteger }"
                                 i1 "{Class: SmallInteger }"
                                 i2 "{Class: SmallInteger }"|

                                cr := a redByte.
                                cg := a greenByte.
                                cb := a blueByte.
                                i1 := qScramble at:((cr bitShift:-4) bitAnd:16r0F) + 1.
                                i1 := i1 + ((qScramble at:((cg bitShift:-4) bitAnd:16r0F) + 1) bitShift:-1).
                                i1 := i1 + ((qScramble at:((cb bitShift:-4) bitAnd:16r0F) + 1) bitShift:-2).

                                cr := b redByte.
                                cg := b greenByte.
                                cb := b blueByte.
                                i2 := qScramble at:((cr bitShift:-4) bitAnd:16r0F) + 1.
                                i2 := i2 + ((qScramble at:((cg bitShift:-4) bitAnd:16r0F) + 1) bitShift:-1).
                                i2 := i2 + ((qScramble at:((cb bitShift:-4) bitAnd:16r0F) + 1) bitShift:-2).

                                i1 < i2
                    ].
    ditherIds := (ditherColors asArray collect:[:clr | clr colorId]) asByteArray.

    "/ build an index table, for fast lookup from manhatten-r-g-b distance
    "/ to the position in the colorList

    clrLookup := ByteArray new:(4096).
    index := 0.
    ditherColors keysAndValuesDo:[:clrPosition :clr |
        |r g b i|

        r := clr redByte.
        g := clr greenByte.
        b := clr blueByte.
        i := qScramble at:((r bitShift:-4) bitAnd:16r0F) + 1.
        i := i + ((qScramble at:((g bitShift:-4) bitAnd:16r0F) + 1) bitShift:-1).
        i := i + ((qScramble at:((b bitShift:-4) bitAnd:16r0F) + 1) bitShift:-2).
        lookupPos := i.

        [index < lookupPos] whileTrue:[        
            clrLookup at:(index+1) put:(clrPosition-1-1).
            index := index + 1
        ]
    ].
    clrLookup from:index+1 to:4096 put:(ditherColors size - 1).

"/    [index <= (4095)] whileTrue:[
"/        clrLookup at:(index+1) put:(ditherColors size - 1).
"/        index := index + 1.
"/    ].

    "/ collect ditherColor components

    lastColor := ditherColors size.
    ditherIds := ByteArray uninitializedNew:lastColor.
    ditherRGBBytes := ByteArray uninitializedNew:(lastColor * 3).
    index := 1.
    1 to:lastColor do:[:pix |
        clr := ditherColors at:pix.
        ditherRGBBytes at:index put:(clr redByte).
        ditherRGBBytes at:index+1 put:(clr greenByte).
        ditherRGBBytes at:index+2 put:(clr blueByte).
        ditherIds at:pix put:clr colorId.

        index := index + 3.
    ].

    pseudoBits := ByteArray uninitializedNew:(width * height).

    w := width + 2.
    error := ByteArray uninitializedNew:w*(3*2).

    w := width.
    h := height.

    failed := true.

%{
    int __x, __y;
    int __eR, __eG, __eB;
    unsigned char *srcP, *dstP;
    unsigned char *rgbP;
    unsigned char *idP;
    unsigned char *dp;
    unsigned char *__clrLookup;
    short *errP, *eP;
    int __fR, __fG, __fB;
    int iR, iG, iB;
    int idx;
    int __w = __intVal(w);
    int __h = __intVal(h);
    int __nColors = __intVal(lastColor);
    int __wR = -1, __wG, __wB;
    static int __qScramble[16] = {
                    0x000 /* 2r000000000000    0 */,
                    0x004 /* 2r000000000100    1 */,
                    0x020 /* 2r000000100000    2 */,
                    0x024 /* 2r000000100100    3 */,
                    0x100 /* 2r000100000000    4 */,
                    0x104 /* 2r000100000100    5 */,
                    0x120 /* 2r000100100000    6 */,
                    0x124 /* 2r000100100100    7 */,
                    0x800 /* 2r100000000000    8 */,
                    0x804 /* 2r100000000100    9 */,
                    0x820 /* 2r100000100000    a */,
                    0x824 /* 2r100000100100    b */,
                    0x900 /* 2r100100000000    c */,
                    0x904 /* 2r100100000100    d */,
                    0x920 /* 2r100100100000    e */,
                    0x924 /* 2r100100100100    f */,
                  };

    if (__isByteArray(__INST(bytes))
     && __isByteArray(pseudoBits)
     && __isByteArray(rgbBytes)
     && __isByteArray(ditherRGBBytes)
     && __isByteArray(ditherIds)
     && __isByteArray(clrLookup)
     && __isByteArray(error)) {
        failed = false;

        srcP = __ByteArrayInstPtr(_INST(bytes))->ba_element;
        dstP = __ByteArrayInstPtr(pseudoBits)->ba_element;
        rgbP = __ByteArrayInstPtr(rgbBytes)->ba_element;
        idP = __ByteArrayInstPtr(ditherIds)->ba_element;
        __clrLookup = __ByteArrayInstPtr(clrLookup)->ba_element;
        errP = (short *) __ByteArrayInstPtr(error)->ba_element;

        /*
         * clear error accumulator
         */
        eP = errP;
        bzero(eP, (__w+2) * 2 * 3);

        for (__y=__h; __y>0; __y--) {
            __eR = __eG = __eB = 0;

            eP = &(errP[3]);
            __eR = eP[0];
            __eG = eP[1];
            __eB = eP[2];

            for (__x=__w; __x>0; __x--) {
                int __want;
                int pix;
                int __wantR, __wantG, __wantB;
                int idx;
                int tR, tG, tB;
                int nR, nG, nB;
                int dR, dG, dB;
                int minDelta, bestIdx;
                int cnt;

                pix = *srcP++;

                /*
                 * wR, wG and wB is the wanted r/g/b value;
                 */
                idx = pix+pix+pix;  /* pix * 3 */

                __wantR = rgbP[idx] + __eR;
                __wantG = rgbP[idx+1] + __eG;
                __wantB = rgbP[idx+2] + __eB;

#define RED_SCALE 30
#define GREEN_SCALE 59
#define BLUE_SCALE 11
#define GOOD_DELTA 30

#define xRED_SCALE 1
#define xGREEN_SCALE 1
#define xBLUE_SCALE 1
#define xGOOD_DELTA 3

#define FAST_LOOKUP
/* #define ONE_SHOT */
#define NPROBE 8

#ifndef FAST_LOOKUP
                if ((__wantR == __wR)
                 && (__wantG == __wG)
                 && (__wantB == __wB)) {
                    /*
                     * same color again - reuse last bestMatch
                     */
                } else 
#endif
                {
                    __wR = __wantR;
                    __wG = __wantG;
                    __wB = __wantB;

#ifdef FAST_LOOKUP
                    if(__wR > 255) __wR = 255;
                    else if (__wR < 0) __wR = 0;
                    if(__wG > 255) __wG = 255;
                    else if (__wG < 0) __wG = 0;
                    if(__wB > 255) __wB = 255;
                    else if (__wB < 0) __wB = 0;

                    {
                        int lookupIndex;
                        int idx, idx0;
                        int d, delta;
                        unsigned char *dp0;

                        dp = __ByteArrayInstPtr(ditherRGBBytes)->ba_element;
                        lookupIndex =    __qScramble[((__wR & 0xF0)>>4)];
                        lookupIndex |=   __qScramble[((__wG & 0xF0)>>4)] >> 1;
                        lookupIndex |=   __qScramble[((__wB & 0xF0)>>4)] >> 2;
                        idx = bestIdx =__clrLookup[lookupIndex];
                        dp += (idx+idx+idx);

                        /* try color at lookupIndex */

                        d = dp[0];
                        delta = (__wR - d) * RED_SCALE;
                        if (delta < 0) delta = -delta;

                        d = dp[1];
                        if (__wG > d) 
                            delta += (__wG - d) * GREEN_SCALE;
                        else 
                            delta += (d - __wG) * GREEN_SCALE;
                        d = dp[2];
                        if (__wB > d) 
                            delta += (__wB - d) * BLUE_SCALE;
                        else 
                            delta += (d - __wB) * BLUE_SCALE;

                        if (delta <= GOOD_DELTA) {
                            goto foundBest;
                        }
                        minDelta = delta;
# ifndef ONE_SHOT
                        idx0 = idx; dp0 = dp;
                        cnt = 0;
                        while ((++cnt <= NPROBE) && (idx > 0)) {
                            /* try previous color(s) */

                            idx--; dp -= 3;
                            d = dp[0];
                            delta = (__wR - d) * RED_SCALE;
                            if (delta < 0) delta = -delta;
                            d = dp[1];
                            if (__wG > d) 
                                delta += (__wG - d) * GREEN_SCALE;
                            else 
                                delta += (d - __wG) * GREEN_SCALE;
                            d = dp[2];
                            if (__wB > d) 
                                delta += (__wB - d) * BLUE_SCALE;
                            else 
                                delta += (d - __wB) * BLUE_SCALE;

                            if (delta < minDelta) {
                                bestIdx = idx;
                                if (delta <= GOOD_DELTA) {
                                    goto foundBest;
                                }
                                minDelta = delta;
                            }
                        }

                        idx = idx0; dp = dp0;
                        cnt = 0;
                        while ((++cnt <= NPROBE) && (++idx < __nColors)) {
                            /* try next color */

                            dp += 3;
                            d = dp[0];
                            delta = (__wR - d) * RED_SCALE;
                            if (delta < 0) delta = -delta;
                            d = dp[1];
                            if (__wG > d) 
                                delta += (__wG - d) * GREEN_SCALE;
                            else 
                                delta += (d - __wG) * GREEN_SCALE;
                            d = dp[2];
                            if (__wB > d) 
                                delta += (__wB - d) * BLUE_SCALE;
                            else 
                                delta += (d - __wB) * BLUE_SCALE;

                            if (delta < minDelta) {
                                bestIdx = idx;
                                if (delta <= GOOD_DELTA) {
                                    goto foundBest;
                                }
                                minDelta = delta;
                            }
                        }
# endif
                    }
        foundBest: ;
#else
/*
                    if(__wR > 255) __wR = 255;
                    else if (__wR < 0) __wR = 0;
                    if(__wG > 255) __wG = 255;
                    else if (__wG < 0) __wG = 0;
                    if(__wB > 255) __wB = 255;
                    else if (__wB < 0) __wB = 0;
*/

                    /* find the best matching color */

                    minDelta = 99999;
                    bestIdx = -1;
                    dp = __ByteArrayInstPtr(ditherRGBBytes)->ba_element;
                    for (idx = 0; idx<__nColors; idx++) {
                        int d, delta;

                        d = dp[0];
                        delta = (__wR - d) * RED_SCALE;
                        if (delta < 0) delta = -delta;
                        if (delta < minDelta) {
                            d = dp[1];
                            if (__wG > d) 
                                delta += (__wG - d) * GREEN_SCALE;
                            else 
                                delta += (d - __wG) * GREEN_SCALE;
                            if (delta < minDelta) {
                                d = dp[2];
                                if (__wB > d) 
                                    delta += (__wB - d) * BLUE_SCALE;
                                else 
                                    delta += (d - __wB) * BLUE_SCALE;

                                if (delta < minDelta) {
                                    bestIdx = idx;
                                    if (delta <= GOOD_DELTA) {
                                        break;
                                    }
                                    minDelta = delta;
                                }
                            }
                        }
                        dp += 3;
                    }
#endif
                }
                dp = __ByteArrayInstPtr(ditherRGBBytes)->ba_element;
                dp += bestIdx * 3;
                dR = dp[0];
                dG = dp[1];
                dB = dp[2];

/*
printf("want: %d/%d/%d (%d/%d/%d) got: %d/%d/%d\n",
                __wantR, __wantG, __wantB,
                __wR, __wG, __wB,
                dR, dG, dB);
*/
                /*
                 * store the corresponding dither colors colorId
                 */
                *dstP++ = idP[bestIdx];

                /*
                 * the new error & distribute the error
                 */
                __eR = __wantR - dR; 
                if (__eR) {
                    tR = __eR >> 4;  /* 16th of error */
                    nR = eP[3] + (tR * 7);/* from accu: error for (x+1 / y) */
                    eP[0] = tR*5;         /* 5/16th for (x / y+1) */
                    eP[-3] = tR*3;        /* 3/16th for (x-1 / y+1) */
                    eP[3] = __eR - (tR*15);  /* 1/16th for (x+1 / y+1) */
                    __eR = nR;
                } else {
                    __eR = eP[3];
                    eP[0] = eP[-3] = eP[3] = 0;
                }

                __eG = __wantG - dG;
                if (__eG) {
                    tG = __eG >> 4;
                    nG = eP[4] + (tG * 7);/* plus 7/16'th of this error */
                    eP[1] = tG*5;
                    eP[-2] = tG*3;
                    eP[4] = __eG - (tG*15);
                    __eG = nG;
                } else {
                    __eG = eP[4];
                    eP[1] = eP[-2] = eP[4] = 0;
                }

                __eB = __wantB - dB; 
                if (__eB) {
                    tB = __eB >> 4;
                    nB = eP[5] + (tB * 7);
                    eP[2] = tB*5;
                    eP[-1] = tB*3;
                    eP[5] = __eB - (tB*15);
                    __eB = nB;
                } else {
                    __eB = eP[5];
                    eP[2] = eP[-1] = eP[5] = 0;
                }

                eP += 3;
            }
        }
    }
%}.
    failed ifTrue:[
        self primitiveFailed.
        ^ nil
    ].

    ^ pseudoBits
!

floydSteinbergDitheredDepth8BitsColors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue 
    "return a floyd-steinberg dithered bitmap from the receiver picture,
     which must be a depth-8 image. 
     This is a special-cased dither method for 8-bit palette images being displayed on
     an 8-bit pseudoColor display, AND fixColor dithering is used.
     Use the colors in the fixColors array, which must be fixR x fixG x fixB
     colors assigned to aDevice, such as the preallocated colors of the
     Color class. 
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create dithered Depth4Images from Depth8Images."

    |pseudoBits  
     rgbBytes
     w       "{Class: SmallInteger }"
     h       "{Class: SmallInteger }"
     index   "{Class: SmallInteger }"
     fixR    "{Class: SmallInteger }"
     fixG    "{Class: SmallInteger }"
     fixB    "{Class: SmallInteger }"
     fixGfixB
     fixIds failed map lastColor 
     rgbIDX  "{Class: SmallInteger }"
     idxAndErrRBytes idxAndErrGBytes idxAndErrBBytes
     error clr|

    self depth ~~ 8 ifTrue:[^ nil].

    fixR := nRed.
    fixR == 0 ifTrue:[ ^ nil].
    fixG := nGreen.
    fixG == 0 ifTrue:[ ^ nil].
    fixB := nBlue.
    fixB == 0 ifTrue:[ ^ nil].

    "/ simple check
    (fixR * fixG * fixB) ~~ fixColors size ifTrue:[
        self error:'invalid color array passed'.
        ^ nil
    ].
    fixIds := (fixColors asArray collect:[:clr | clr colorId]) asByteArray.

    "/
    "/ collect color components as integer values (for integer arithmetic)
    "/
    rgbBytes := ByteArray uninitializedNew:256 * 3.

    index := 1.

    photometric == #palette ifTrue:[
        lastColor := colorMap size - 1
    ] ifFalse:[
        lastColor := 255.
    ].
    0 to:lastColor do:[:pix |
        clr := self colorFromValue:pix.
        rgbBytes at:index put:(clr redByte).
        rgbBytes at:index+1 put:(clr greenByte).
        rgbBytes at:index+2 put:(clr blueByte).

        index := index + 3.
    ].

    pseudoBits := ByteArray uninitializedNew:(width * height).

    w := width + 2.
    error := ByteArray new:w*(3*2).

    w := width.
    h := height.

    idxAndErrRBytes := ByteArray uninitializedNew:256*2.
    idxAndErrGBytes := ByteArray uninitializedNew:256*2.
    idxAndErrBBytes := ByteArray uninitializedNew:256*2.

    fixGfixB := fixG * fixB.
    index := 1.
    0 to:255 do:[:i |
        rgbIDX := (i * (fixR-1) + 128) // 255. "red index rounded"
        idxAndErrRBytes at:index put:(rgbIDX * fixGfixB).
        idxAndErrRBytes at:index+1 put:i - (rgbIDX * 255 // (fixR-1)) + 128.

        rgbIDX := (i * (fixG-1) + 128) // 255. "green index rounded"
        idxAndErrGBytes at:index put:(rgbIDX * fixB). 
        idxAndErrGBytes at:index+1 put:i - (rgbIDX * 255 // (fixG-1)) + 128.

        rgbIDX := (i * (fixB-1) + 128) // 255. "blue index rounded"
        idxAndErrBBytes at:index put:(rgbIDX ).
        idxAndErrBBytes at:index+1 put:i - (rgbIDX * 255 // (fixB-1)) + 128.
        index := index + 2.
    ].

    failed := true.

%{
    int __x, __y;
    int __eR, __eG, __eB;
    unsigned char *srcP, *dstP;
    unsigned char *rgbP;
    unsigned char *idP;
    unsigned char *__idxAndErrRBytes, *__idxAndErrGBytes, *__idxAndErrBBytes;
    short *errP, *eP;
    int idx;
    int __w = __intVal(w);

    if (__isByteArray(__INST(bytes))
     && __isByteArray(pseudoBits)
     && __isByteArray(rgbBytes)
     && __isByteArray(fixIds)
     && __isByteArray(error)
     && __bothSmallInteger(fixR, fixG)
     && __isSmallInteger(fixB)) {
        failed = false;

        srcP = __ByteArrayInstPtr(_INST(bytes))->ba_element;
        dstP = __ByteArrayInstPtr(pseudoBits)->ba_element;
        rgbP = __ByteArrayInstPtr(rgbBytes)->ba_element;
        idP = __ByteArrayInstPtr(fixIds)->ba_element;
        __idxAndErrRBytes = __ByteArrayInstPtr(idxAndErrRBytes)->ba_element;
        __idxAndErrGBytes = __ByteArrayInstPtr(idxAndErrGBytes)->ba_element;
        __idxAndErrBBytes = __ByteArrayInstPtr(idxAndErrBBytes)->ba_element;
        errP = (short *) __ByteArrayInstPtr(error)->ba_element;

        eP = errP;

        for (__y=__intVal(h); __y>0; __y--) {
            __eR = __eG = __eB = 0;

            eP = &(errP[3]);
            __eR = eP[0];
            __eG = eP[1];
            __eB = eP[2];

            for (__x=__w; __x>0; __x--) {
                int __want;
                int pix;
                int idx;
                int tR, tG, tB;
                int nR, nG, nB;
                int iRGB;

                pix = *srcP++;

                /*
                 * wR, wG and wB is the wanted r/g/b value;
                 */
                pix = pix+pix+pix;  /* pix * 3 */

                /*
                 * compute indexR/G/B and the new error:
                 */
                __want = rgbP[pix]   + __eR;
                if (__want > 255) __want = 255;
                else if (__want < 0) __want = 0;
                __want += __want;
                idx = __idxAndErrRBytes[__want];   
                __eR = __idxAndErrRBytes[__want+1]; 
                __eR -= 128;

                __want = rgbP[pix+1] + __eG;
                if (__want > 255) __want = 255;
                else if (__want < 0) __want = 0;
                __want += __want;
                idx += __idxAndErrGBytes[__want];   
                __eG = __idxAndErrGBytes[__want+1]; 
                __eG -= 128;

                __want = rgbP[pix+2] + __eB;
                if (__want > 255) __want = 255;
                else if (__want < 0) __want = 0;
                __want += __want;
                idx += __idxAndErrBBytes[__want];   
                __eB = __idxAndErrBBytes[__want+1]; 
                __eB -= 128;

                /*
                 * store the corresponding dither colors colorId
                 */
                *dstP++ = idP[idx];

                /*
                 * distribute the error
                 */
                if (__eR) {
                    tR = __eR >> 4;  /* 16th of error */
                    nR = eP[3] + (tR * 7);/* from accu: error for (x+1 / y) */
                    eP[0] = tR*5;         /* 5/16th for (x / y+1) */
                    eP[-3] = tR*3;        /* 3/16th for (x-1 / y+1) */
                    eP[3] = __eR - (tR*15);  /* 1/16th for (x+1 / y+1) */
                    __eR = nR;
                } else {
                    __eR = eP[3];
                    eP[0] = eP[-3] = eP[3] = 0;
                }

                if (__eG) {
                    tG = __eG >> 4;  /* 16th of error */
                    nG = eP[4] + (tG * 7);/* plus 7/16'th of this error */
                    eP[1] = tG*5;
                    eP[-2] = tG*3;
                    eP[4] = __eG - (tG*15);
                    __eG = nG;
                } else {
                    __eG = eP[4];
                    eP[1] = eP[-2] = eP[4] = 0;
                }

                if (__eB) {
                    tB = __eB >> 4;  /* 16th of error */
                    nB = eP[5] + (tB * 7);
                    eP[2] = tB*5;
                    eP[-1] = tB*3;
                    eP[5] = __eB - (tB*15);
                    __eB = nB;
                } else {
                    __eB = eP[5];
                    eP[2] = eP[-1] = eP[5] = 0;
                }

                eP += 3;
            }
        }
    }
%}.
    failed ifTrue:[
        self primitiveFailed.
        ^ nil
    ].

    ^ pseudoBits


!

floydSteinbergDitheredGrayBitsDepth:depth
    "return the bits for dithering a gray image from the image.
     Works for any source depths / photometric, 
     but possibly slow since each pixel is processed individually.
     Redefined by some subclasses for more performance (D8Image/D24Image)"

    |dstIndex        "{Class: SmallInteger }"
     nextDst         "{Class: SmallInteger }"
     bytesPerOutRow  "{Class: SmallInteger }"
     outBits greyValues greyErrors greyPixels greyLevels
     errorArray
     nextErrorArray
     t
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" 
     grey
     eR eRB eB eLB |

    depth > 8 ifTrue:[
        self error:'unimplemented conversion'.
        ^ nil
    ].

    w := width.
    h := height.

    bytesPerOutRow := ((w * depth) + 7) // 8.
    outBits := ByteArray uninitializedNew:(bytesPerOutRow * h).
    (outBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    greyLevels := (1 bitShift:depth) - 1.

    errorArray := Array new:w+2.
    nextErrorArray := Array new:w+2.

    nextErrorArray atAllPut:0.

    dstIndex := 1.
    bitCnt := 8.

    self depth <= 12 ifTrue:[
        "/ fetch scaled brightness values outside of loop into a table;
        "/ use table-value in loop

        greyValues := self greyMapForRange:(greyLevels).

        greyPixels := greyValues collect:[:v | v isNil ifTrue:[
                                                   0
                                               ] ifFalse:[
                                                   v truncated]].

        greyPixels := ByteArray withAll:greyPixels.

        greyErrors := greyValues collect:[:v | v isNil ifTrue:[
                                                   0
                                               ] ifFalse:[
                                                   ((v - v truncated) * 1024) truncated
                                               ]].

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerOutRow.
            byte := 0.

            t := errorArray.
            errorArray := nextErrorArray.
            nextErrorArray := t.

            nextErrorArray atAllPut:0.

            self valuesAtY:y from:0 to:(w-1) do:[:x :value |
                |e     "{ Class: SmallInteger }"
                 pixel "{ Class: SmallInteger }"
                 error "{ Class: SmallInteger }"
                 e16   "{ Class: SmallInteger }"
                 xE    "{ Class: SmallInteger }" 
                 xN    "{ Class: SmallInteger }" |

                pixel := greyPixels at:(value + 1).

                "/ adjust error
                xE := x + 2.
                error := (greyErrors at:(value + 1)) + (errorArray at:xE).

                byte := byte bitShift:depth.
                error > 512 "0.5" ifTrue:[
                    pixel := pixel + 1.      
                    e := error - 1024 "1.0"
                ] ifFalse:[
                    e := error              
                ].
                byte := byte bitOr:pixel.

                e ~= 0 ifTrue:[
                    e16 := e // 16.

                    eR  := e16 * 7.              "/ 7/16 to right
                    eRB := e16 * 1.              "/ 1/16 to right below
                    eB  := e16 * 5.              "/ 5/16 to below
                    eLB := e - eR - eRB - eB.  "/ 3/16 to left below

                    xN := xE + 1.
                    eR ~= 0 ifTrue:[
                        errorArray     at:xN put:(errorArray at:xN) + eR.
                    ].
                    eRB ~= 0 ifTrue:[
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eRB.
                    ].
                    eB ~= 0 ifTrue:[
                        nextErrorArray at:xE put:(nextErrorArray at:xE) + eB.
                    ].
                    eLB ~= 0 ifTrue:[
                        xN := xE - 1.
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eLB.
                    ].
                ].

                bitCnt := bitCnt - depth.
                bitCnt == 0 ifTrue:[
                    outBits at:dstIndex put:byte.
                    dstIndex := dstIndex + 1.
                    byte := 0.
                    bitCnt := 8.
                ].

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                outBits at:dstIndex put:byte.
                bitCnt := 8.
            ].

            dstIndex := nextDst.
        ].
    ] ifFalse:[
        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerOutRow.
            byte := 0.

            t := errorArray.
            errorArray := nextErrorArray.
            nextErrorArray := t.

            nextErrorArray atAllPut:0.

            self colorsAtY:y from:0 to:(w-1) do:[:x :clr |
                |e     "{ Class: SmallInteger }"
                 pixel "{ Class: SmallInteger }"
                 error "{ Class: SmallInteger }"
                 e16   "{ Class: SmallInteger }"
                 xE    "{ Class: SmallInteger }" 
                 xN    "{ Class: SmallInteger }" |

                grey := (clr brightness * greyLevels).
                pixel := grey truncated.
                error := ((grey - pixel) * 1024) truncated.

                "/ adjust error
                xE := x + 2.
                error := error + (errorArray at:xE).

                byte := byte bitShift:depth.
                error > 512 "0.5" ifTrue:[
                    pixel := pixel + 1.      
                    e := error - 1024 "1.0"
                ] ifFalse:[
                    e := error              
                ].

                byte := byte bitOr:pixel.

                e ~= 0 ifTrue:[
                    e16 := e // 16.

                    eR  := e16 * 7.              "/ 7/16 to right
                    eRB := e16 * 1.              "/ 1/16 to right below
                    eB  := e16 * 5.              "/ 5/16 to below
                    eLB := e - eR - eRB - eB.  "/ 3/16 to left below

                    xN := xE + 1.
                    eR ~= 0 ifTrue:[
                        errorArray     at:xN put:(errorArray at:xN) + eR.
                    ].
                    eRB ~= 0 ifTrue:[
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eRB.
                    ].
                    eB ~= 0 ifTrue:[
                        nextErrorArray at:xE put:(nextErrorArray at:xE) + eB.
                    ].
                    eLB ~= 0 ifTrue:[
                        xN := xE - 1.
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eLB.
                    ].
                ].

                bitCnt := bitCnt - depth.
                bitCnt == 0 ifTrue:[
                    outBits at:dstIndex put:byte.
                    dstIndex := dstIndex + 1.
                    byte := 0.
                    bitCnt := 8.
                ].

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                outBits at:dstIndex put:byte.
                bitCnt := 8.
            ].

            dstIndex := nextDst.
        ].
    ].

    ^ outBits

    "Created: 10.6.1996 / 13:28:22 / cg"
    "Modified: 11.6.1996 / 00:13:38 / cg"
!

floydSteinbergDitheredMonochromeBits
    "return the bitmap for a dithered monochrome bitmap from the image.
     Works for any source depths / photometric, 
     but very very slow since each pixel is processed individually.
     Redefined by some subclasses for more performance (D8Image)"

    |dstIndex        "{Class: SmallInteger }"
     nextDst         "{Class: SmallInteger }"
     bytesPerMonoRow "{Class: SmallInteger }"
     monoBits greyValues
     errorArray
     nextErrorArray
     e eD t
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" 
     grey dT 
     eR eRB eB eLB |

    w := width.
    h := height.

    bytesPerMonoRow := (w + 7) // 8.
    monoBits := ByteArray uninitializedNew:(bytesPerMonoRow * h).
    (monoBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    errorArray := Array new:w+2.
    nextErrorArray := Array new:w+2.

    nextErrorArray atAllPut:0.

    dstIndex := 1.
    bitCnt := 8.
    byte := 0.

    self depth <= 12 ifTrue:[
        "/ fetch scaled brightness values outside of loop into a table;
        "/ use table-value in loop

        greyValues := self greyMapForRange:(255 * 1024).
        greyValues := greyValues collect:[:v | v rounded].

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerMonoRow.

            t := errorArray.
            errorArray := nextErrorArray.
            nextErrorArray := t.

            nextErrorArray atAllPut:0.

            self valuesAtY:y from:0 to:(w-1) do:[:x :pixel |
%{
                int __grey, __e;
                int __byte = __intVal(byte);
                OBJ *__errorArray = __ArrayInstPtr(errorArray)->a_element;
                OBJ *__nextErrorArray = __ArrayInstPtr(nextErrorArray)->a_element;
                int __x = __intVal(x);
                int __eR, __eB, __eRB, __eLB, __eI;
                int __bitCnt = __intVal(bitCnt);

                __grey = __intVal(__ArrayInstPtr(greyValues)->a_element[__intVal(pixel)]);
                __grey += __intVal(__errorArray[__x+1]);
                
                __byte <<= 1;
                if (__grey > 127*1024) {
                    __e = __grey - (255*1024);
                    __byte |= 1;
                } else {
                    __e = __grey;
                }
                if (__e) {
                    __eI = __e >> 4;
                    __eR  = __eI * 7;            
                    __eRB = __eI * 1;            
                    __eB  = __eI * 5;            
                    __eLB = __e - __eR - __eRB - __eB;
                    __errorArray[__x+2] = __MKSMALLINT(__intVal(__errorArray[__x+2]) + __eR);
                    __nextErrorArray[__x+2] = __MKSMALLINT(__intVal(__nextErrorArray[__x+2]) + __eRB);
                    __nextErrorArray[__x+1] = __MKSMALLINT(__intVal(__nextErrorArray[__x+1]) + __eB);
                    __nextErrorArray[__x  ] = __MKSMALLINT(__intVal(__nextErrorArray[__x  ]) + __eLB);

                }
                __bitCnt--;
                if (__bitCnt == 0) {
                    int __dstIndex = __intVal(dstIndex);

                    __ByteArrayInstPtr(monoBits)->ba_element[__dstIndex-1] = __byte;
                    dstIndex = __MKSMALLINT(__dstIndex + 1);
                    __byte = 0;
                    __bitCnt = 8;
                }
                byte = __MKSMALLINT(__byte);
                bitCnt = __MKSMALLINT(__bitCnt);
%}.

"/                |eI "{ Class: SmallInteger }" 
"/                 xE "{ Class: SmallInteger }" 
"/                 xN "{ Class: SmallInteger }" |
"/
"/                "/ get the colors grey value [0 .. 1]
"/                grey := greyValues at:(pixel + 1).
"/
"/                "/ adjust error
"/                xE := x + 2.
"/                grey := (grey + (errorArray at:xE)).
"/
"/                byte := byte bitShift:1.
"/                grey > (127*1024) ifTrue:[
"/                    byte := byte bitOr:1.      "/ white
"/                    e := grey - (255*1024)
"/                ] ifFalse:[
"/                    e := grey                  "/ black
"/                ].
"/                e ~= 0 ifTrue:[
"/                    eD := e.
"/                    eI := e // 16.
"/
"/                    eR  := eI * 7.              "/ 7/16 to right
"/                    eRB := eI * 1.              "/ 1/16 to right below
"/                    eB  := eI * 5.              "/ 5/16 to below
"/                    eLB := eD - eR - eRB - eB.  "/ 3/16 to left below
"/
"/                    xN := xE + 1.
"/                    eR ~= 0 ifTrue:[
"/                        errorArray     at:xN put:(errorArray at:xN) + eR.
"/                    ].
"/                    eRB ~= 0 ifTrue:[
"/                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eRB.
"/                    ].
"/                    eB ~= 0 ifTrue:[
"/                        nextErrorArray at:xE put:(nextErrorArray at:xE) + eB.
"/                    ].
"/                    eLB ~= 0 ifTrue:[
"/                        xN := xE - 1.
"/                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eLB.
"/                    ].
"/                ].
"/
"/                bitCnt := bitCnt - 1.
"/                bitCnt == 0 ifTrue:[
"/                    monoBits at:dstIndex put:byte.
"/                    dstIndex := dstIndex + 1.
"/                    byte := 0.
"/                    bitCnt := 8.
"/                ].
                  0
            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                monoBits at:dstIndex put:byte.
                bitCnt := 8.
                byte := 0.
            ].

            dstIndex := nextDst.
        ].
    ] ifFalse:[
        'Image [info]: slow floydSteinberg dither ..' infoPrintCR.

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerMonoRow.

            t := errorArray.
            errorArray := nextErrorArray.
            nextErrorArray := t.

            nextErrorArray atAllPut:0.

            self colorsAtY:y from:0 to:(w-1) do:[:x :clr |
                |eI "{ Class: SmallInteger }" 
                 xE "{ Class: SmallInteger }" 
                 xN "{ Class: SmallInteger }" |

                "/ get the colors grey value [0 .. 1]
                grey := (clr brightness * 255).

                "/ adjust error
                xE := x + 2.
                grey := (grey + (errorArray at:xE)) rounded.

                byte := byte bitShift:1.
                grey > 127 ifTrue:[
                    byte := byte bitOr:1.      "/ white
                    e := grey - 255
                ] ifFalse:[
                    e := grey                  "/ black
                ].

                e ~= 0 ifTrue:[
                    eD := e.
                    eI := e // 16.
                    eR  := eI * 7.              "/ 7/16 to right
                    eRB := eI * 1.              "/ 1/16 to right below
                    eB  := eI * 5.              "/ 5/16 to below
                    eLB := eD - eR - eRB - eB.  "/ 3/16 to left below

                    xN := xE + 1.
                    eR ~= 0 ifTrue:[
                        errorArray     at:xN put:(errorArray at:xN) + eR.
                    ].
                    eRB ~= 0 ifTrue:[
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eRB.
                    ].
                    eB ~= 0 ifTrue:[
                        nextErrorArray at:xE put:(nextErrorArray at:xE) + eB.
                    ].
                    eLB ~= 0 ifTrue:[
                        xN := xE - 1.
                        nextErrorArray at:xN put:(nextErrorArray at:xN) + eLB.
                    ].
                ].

                bitCnt := bitCnt - 1.
                bitCnt == 0 ifTrue:[
                    monoBits at:dstIndex put:byte.
                    dstIndex := dstIndex + 1.
                    byte := 0.
                    bitCnt := 8.
                ].

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                monoBits at:dstIndex put:byte.
                bitCnt := 8.
                byte := 0.
            ].

            dstIndex := nextDst.
        ].
    ].

    ^ monoBits

    "Created: 8.6.1996 / 16:39:46 / cg"
    "Modified: 10.6.1996 / 15:12:11 / cg"
!

nearestPaintDepth8BitsColors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue 
    "return a nearest paint bitmap from the receiver picture,
     which must be a depth-8 image. 
     This is a special-cased dither method for 8-bit palette images being displayed on
     an 8-bit pseudoColor display, AND fixColor dithering is used.
     Use the colors in the fixColors array, which must be fixR x fixG x fixB
     colors assigned to aDevice, such as the preallocated colors of the
     Color class."

    |pseudoBits  
     fixR    "{Class: SmallInteger }"
     fixG    "{Class: SmallInteger }"
     fixB    "{Class: SmallInteger }"
     fixGfixB 
     r       "{Class: SmallInteger }"
     g       "{Class: SmallInteger }"
     b       "{Class: SmallInteger }"
     idx     "{Class: SmallInteger }"
     idMap lastColor 
     clr|

    self depth ~~ 8 ifTrue:[^ nil].

    fixR := nRed.
    fixR == 0 ifTrue:[ ^ nil].
    fixG := nGreen.
    fixG == 0 ifTrue:[ ^ nil].
    fixB := nBlue.
    fixB == 0 ifTrue:[ ^ nil].

    "/ simple check
    (fixR * fixG * fixB) ~~ fixColors size ifTrue:[
        self error:'invalid color array passed'.
        ^ nil
    ].

    "/
    "/ collect colorIds
    "/
    photometric == #palette ifTrue:[
        lastColor := colorMap size - 1
    ] ifFalse:[
        lastColor := 255.
    ].
    fixGfixB := fixG * fixB.

    idMap := ByteArray uninitializedNew:256.

    0 to:lastColor do:[:pix |
        clr := self colorFromValue:pix.
        r := clr redByte.
        g := clr greenByte.
        b := clr blueByte.
        idx := ((r * (fixR-1) + 128) // 255) * fixGfixB.
        idx := idx + (((g * (fixG-1) + 128) // 255) * fixB).
        idx := idx + ((b * (fixB-1) + 128) // 255).
        idMap at:(pix+1) put:(fixColors at:(idx+1)) colorId.
    ].

    pseudoBits := ByteArray uninitializedNew:(width * height).

    "/ translate

    bytes expandPixels:8         "xlate only"
                width:width 
               height:height
                 into:pseudoBits
              mapping:idMap.

    ^ pseudoBits

    "Modified: 18.6.1996 / 09:18:09 / cg"
!

nfloydSteinbergDitheredDepth8BitsColors:colors
    "return a floyd-steinberg dithered bitmap from the receiver picture,
     which must be a depth-8 image. 
     This method expects an array of colors to be used for dithering
     (which need not be a colorCubes colors)."

    |pseudoBits  
     rgbBytes 
     ditherRGBBytes ditherColors
     w       "{Class: SmallInteger }"
     h       "{Class: SmallInteger }"
     index   "{Class: SmallInteger }"
     numR    "{Class: SmallInteger }"
     numG    "{Class: SmallInteger }"
     numB    "{Class: SmallInteger }"
     bitsR    "{Class: SmallInteger }"
     bitsG    "{Class: SmallInteger }"
     bitsB    "{Class: SmallInteger }"
     maxBits  "{Class: SmallInteger }"
     maskR    "{Class: SmallInteger }"
     maskG    "{Class: SmallInteger }"
     maskB    "{Class: SmallInteger }"
     shR      "{Class: SmallInteger }"
     shG      "{Class: SmallInteger }"
     shB      "{Class: SmallInteger }"
     ditherIds failed map lastColor colorsByDistance qScramble
     clrLookup lookupPos cube nCube
     dR  "{Class: SmallInteger }"
     dG  "{Class: SmallInteger }"
     dB  "{Class: SmallInteger }"
     iR    "{Class: SmallInteger }"
     iRG   "{Class: SmallInteger }"
     iRGB  "{Class: SmallInteger }"
     clr 
     rI  "{Class: SmallInteger }"
     gI  "{Class: SmallInteger }"
     bI  "{Class: SmallInteger }"
     maxIDX  "{Class: SmallInteger }"
     subCubeColorCollection
     error 
     dl "{Class: SmallInteger }"|

    self depth ~~ 8 ifTrue:[^ nil].

    "/
    "/ collect color components as integer values (for integer arithmetic)
    "/
    rgbBytes := ByteArray uninitializedNew:256 * 3.

    photometric == #palette ifTrue:[
        lastColor := colorMap size - 1
    ] ifFalse:[
        lastColor := 255.
    ].
    index := 1.
    0 to:lastColor do:[:pix |
        clr := self colorFromValue:pix.
        rgbBytes at:index put:(clr redByte).
        rgbBytes at:index+1 put:(clr greenByte).
        rgbBytes at:index+2 put:(clr blueByte).

        index := index + 3.
    ].

    "/ collect valid ditherColors ...

    ditherColors := colors select:[:clr | clr notNil].
    ditherColors := ditherColors select:[:clr | clr colorId notNil].

    "/ collect ditherColor components

    lastColor := ditherColors size.
    ditherIds := ByteArray uninitializedNew:lastColor.
    ditherRGBBytes := ByteArray uninitializedNew:(lastColor * 3).
    index := 1.
    1 to:lastColor do:[:pix |
        clr := ditherColors at:pix.
        ditherRGBBytes at:index put:(clr redByte).
        ditherRGBBytes at:index+1 put:(clr greenByte).
        ditherRGBBytes at:index+2 put:(clr blueByte).
        ditherIds at:pix put:clr colorId.

        index := index + 3.
    ].

    "/ place the ditherColor positions into a color cube
    bitsR := 5.
    bitsG := 6.
    bitsB := 4.
    maxBits := 6.

"/    bitsR := 4.
"/    bitsG := 4.
"/    bitsB := 3.

    numR := 1 bitShift:bitsR.
    numG := 1 bitShift:bitsG.
    numB := 1 bitShift:bitsB.

    maskR := (numR-1) bitShift:(bitsG + bitsB).
    maskG := (numG-1) bitShift:bitsB.
    maskB := numB-1.

    shR := -16+bitsR+bitsG+bitsB.
    shG := -16+bitsG+bitsB.
    shB := -16+bitsB.

    maxIDX := numR*numG*numB.
    cube := Array new:maxIDX.
    1 to:lastColor do:[:clrIdx |
        clr := ditherColors at:clrIdx.
        rI := clr scaledRed. rI := (rI bitShift:shR) bitAnd:maskR.
        gI := clr scaledGreen. gI := (gI bitShift:shG) bitAnd:maskG.
        bI := clr scaledBlue. bI := (bI bitShift:shB) bitAnd:maskB.
        index := rI + gI + bI + 1.
        subCubeColorCollection := cube at:index.
        subCubeColorCollection isNil ifTrue:[
            subCubeColorCollection := OrderedCollection new.
            cube at:index put:subCubeColorCollection.
        ].
        subCubeColorCollection add:(clrIdx - 1).
    ].

    shR := 1 bitShift:(bitsG+bitsB).
    shG := 1 bitShift:(bitsB).
    shB := 1.

    1 to:maxIDX do:[:i |
        subCubeColorCollection := cube at:i.
        subCubeColorCollection notNil ifTrue:[
            cube at:i put:(subCubeColorCollection asByteArray)
        ]
    ].

"/    nCube := cube copy.
"/
"/    cube keysAndValuesDo:[:i :indices |
"/        indices notNil ifTrue:[
"/            nCube at:i put:(indices asByteArray)
"/        ] ifFalse:[
"/            "/ find nearest color
"/        
"/            dl := 1.
"/            [dl < maxBits] whileTrue:[
"/                dR := dl negated.
"/                [dR <= dl] whileTrue:[
"/                    iR := i + (dR * shR).
"/                    (iR > 0 and:[iR < maxIDX]) ifTrue:[    
"/                        dG := dl negated.
"/                        [dG < dl] whileTrue:[
"/                            iRG := iR + (dG * shG).
"/                            (iRG > 0 and:[iRG < maxIDX]) ifTrue:[
"/                                dB := dl negated.
"/                                [dB < dl] whileTrue:[
"/                                    iRGB := iRG + dB.
"/                                    (iRG > 0 and:[iRG < maxIDX]) ifTrue:[
"/                                        (cube at:iRGB) notNil ifTrue:[
"/                                            nCube at:i put:(cube at:iRGB).
"/                                            dB := dG := dR := dl := 999.
"/                                        ]
"/                                    ].
"/                                    dB := dB + 1.
"/                                ]
"/                            ].
"/                            dG := dG + 1.
"/                        ]
"/                    ].
"/                    dR := dR + 1.
"/                ].
"/                dl := dl + 1.
"/            ]
"/        ]
"/    ].
"/self halt.

    "/ now, cube contains collections of colors which are
    "/ positioned in a subCube; quickly accessed by a lookup 

    pseudoBits := ByteArray uninitializedNew:(width * height).

    w := width.
    h := height.
    error := ByteArray new:(w+2)*3*2.

%{
#define BITSR   5
#define BITSG   6
#define BITSB   4
#define MAXBITS 6

#define xBITSR   4
#define xBITSG   4
#define xBITSB   3
#define xMAXBITS 4

#define NR      32 /* (1<<BITSR) */
#define NG      64 /* (1<<BITSG) */
#define NB      16 /* (1<<BITSB) */
#define MAXRGB  64 /* (1<<MAXBITS) */

#define SHR     (BITSG+BITSB)
#define SHG     BITSB

#define REMEMBER_SEARCH
#define xNO_FLOYD_STEINBERG

    int __x, __y;
    int __eR, __eG, __eB;
    unsigned char *srcP, *dstP;
    unsigned char *rgbP;
    unsigned char *idP;
    short *errP, *eP;
    int __fR, __fG, __fB;
    int idx;
    int __w = __intVal(w);
    int __h = __intVal(h);
    int __nColors = __intVal(lastColor);
    int __wR = -1, __wG, __wB;
    OBJ *__cube;
    int cubeIndex, cubeIndex2;

    if (__isByteArray(__INST(bytes))
     && __isByteArray(pseudoBits)
     && __isByteArray(rgbBytes)
     && __isByteArray(ditherRGBBytes)
     && __isByteArray(ditherIds)
     && __isByteArray(error)) {
        failed = false;

        srcP = __ByteArrayInstPtr(_INST(bytes))->ba_element;
        dstP = __ByteArrayInstPtr(pseudoBits)->ba_element;
        rgbP = __ByteArrayInstPtr(rgbBytes)->ba_element;
        idP = __ByteArrayInstPtr(ditherIds)->ba_element;
        errP = (short *) __ByteArrayInstPtr(error)->ba_element;
        __cube = __ArrayInstPtr(cube)->a_element;

        eP = errP;

        for (__y=__h; __y>0; __y--) {
            eP = &(errP[3]);
            __eR = eP[0];
            __eG = eP[1];
            __eB = eP[2];

            for (__x=__w; __x>0; __x--) {
                int __want;
                int pix;
                int __wantR, __wantG, __wantB;
                int idx;
                int tR, tG, tB;
                int nR, nG, nB;
                int __dR, __dG, __dB;
                int minDelta, bestIdx;
                int __iR, __iG, __iB;
                int cR, cG, cB;   
                int delta;
                OBJ subCubeColors;

                pix = *srcP++;

                /*
                 * wR, wG and wB is the wanted r/g/b value;
                 */
                idx = pix+pix+pix;  /* pix * 3 */

                __wR = __wantR = rgbP[idx] + __eR;
                __wG = __wantG = rgbP[idx+1] + __eG;
                __wB = __wantB = rgbP[idx+2] + __eB;
                if(__wR > 255) __wR = 255;
                else if (__wR < 0) __wR = 0;
                if(__wG > 255) __wG = 255;
                else if (__wG < 0) __wG = 0;
                if(__wB > 255) __wB = 255;
                else if (__wB < 0) __wB = 0;

                __iR = __wR >> (8-BITSR);
                __iG = __wG >> (8-BITSG);
                __iB = __wB >> (8-BITSB);

                cubeIndex = (__iR<<SHR) + (__iG<<SHG) + __iB;
                subCubeColors = __cube[cubeIndex];

                if (subCubeColors == nil) {
                    /* search around in spirals, for the first match */

                    delta = 1;
                    while (delta < MAXRGB) {
                        /* check plane above */
                        cR = __iR + delta;
                        if ((unsigned)cR < NR) {
                            for (cG=__iG-delta; cG<=__iG+delta; cG++) {
                                if ((unsigned)cG < NG) {
                                    for (cB=__iB-delta; cB<=__iB+delta; cB++) {
                                        if ((unsigned)cB < NB) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /* check plane below */
                        cR = __iR - delta;
                        if ((unsigned)cR < NR) {
                            for (cG=__iG-delta; cG<=__iG+delta; cG++) {
                                if ((unsigned)cG < NG) {
                                    for (cB=__iB-delta; cB<=__iB+delta; cB++) {
                                        if ((unsigned)cB < NB) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /* check plane to the right */
                        cG = __iG + delta;
                        if ((unsigned)cG < NG) {
                            for (cR=__iR-delta+1; cR<=__iR+delta-1; cR++) {
                                if ((unsigned)cR < NR) {
                                    for (cB=__iB-delta; cB<=__iB+delta; cB++) {
                                        if ((unsigned)cB < NB) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /* check plane to the left */
                        cG = __iG - delta;
                        if ((unsigned)cG < NG) {
                            for (cR=__iR-delta+1; cR<=__iR+delta-1; cR++) {
                                if ((unsigned)cR < NR) {
                                    for (cB=__iB-delta; cB<=__iB+delta; cB++) {
                                        if ((unsigned)cB < NB) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /* check plane at back */
                        cB = __iB + delta;
                        if ((unsigned)cB < NB) {
                            for (cR=__iR-delta+1; cR<=(__iR+delta-1); cR++) {
                                if ((unsigned)cR < NR) {
                                    for (cG=__iG-delta+1; cG<=(__iG+delta-1); cG++) {
                                        if ((unsigned)cG < NG) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /* check plane at front */
                        cB = __iB - delta;
                        if ((unsigned)cB < NB) {
                            for (cR=__iR-delta+1; cR<=(__iR+delta-1); cR++) {
                                if ((unsigned)cR < NR) {
                                    for (cG=__iG-delta+1; cG<=(__iG+delta-1); cG++) {
                                        if ((unsigned)cG < NG) {
                                            cubeIndex2 = (cR<<SHR) + (cG<<SHG) + cB;
                                            subCubeColors = __cube[cubeIndex2];
                                            if (__isNonNilObject(subCubeColors)) {
                                                goto found;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        delta = delta + 1;

                    }
                    /* cannot happen - will lead to a segmentation violation ... */
                    subCubeColors = nil;

    found:
                    __iR = cR;
                    __iG = cG;
                    __iB = cB;
                    bestIdx = __ByteArrayInstPtr(subCubeColors)->ba_element[0];
#ifdef REMEMBER_SEARCH
                    __cube[cubeIndex] = __MKSMALLINT(bestIdx);
#endif
                } else {
#ifdef REMEMBER_SEARCH
                    if (__isSmallInteger(subCubeColors)) {
                        bestIdx = __intVal(subCubeColors);
                    } else 
#endif
                    {
                        bestIdx = __ByteArrayInstPtr(subCubeColors)->ba_element[0];
                    }
                }
                
                /*
                 * ok, now, we have found a collection of nearby
                 * colors in subCubeColors.
                 *
                 * since the error is at most 1/16 (i.e. roughly 6%),
                 * dont care for searching the best - simply take the
                 * first color found there.
                 * (statistic reduces the error to even a smaller value).
                 * There is no real problem due to that error, since
                 * it will be diffused anyway ...
                 */

#ifndef NO_FLOYD_STEINBERG
                {
                    unsigned char *dp;

                    /*
                     * fetch that colors r/g/b components
                     */
                    dp = __ByteArrayInstPtr(ditherRGBBytes)->ba_element;
                    dp += bestIdx * 3;
                    __dR = dp[0];
                    __dG = dp[1];
                    __dB = dp[2];
                }
#endif

/*
printf("want: %d/%d/%d (%d/%d/%d) got: %d/%d/%d\n",
                __wantR, __wantG, __wantB,
                __wR, __wG, __wB,
                __dR, __dG, __dB);
*/
                /*
                 * store the corresponding dither colors colorId
                 */
                *dstP++ = idP[bestIdx];

#ifndef NO_FLOYD_STEINBERG
                /*
                 * the new error & distribute the error
                 */
                __eR = __wantR - __dR; 
                if (__eR) {
                    tR = __eR >> 4;  /* 16th of error */
                    nR = eP[3] + (tR * 7);/* from accu: error for (x+1 / y) */
                    eP[0] = tR*5;         /* 5/16th for (x / y+1) */
                    eP[-3] = tR*3;        /* 3/16th for (x-1 / y+1) */
                    eP[3] = __eR - (tR*15);  /* 1/16th for (x+1 / y+1) */
                    __eR = nR;
                } else {
                    __eR = eP[3];
                    eP[0] = eP[-3] = eP[3] = 0;
                }

                __eG = __wantG - __dG;
                if (__eG) {
                    tG = __eG >> 4;
                    nG = eP[4] + (tG * 7);/* plus 7/16'th of this error */
                    eP[1] = tG*5;
                    eP[-2] = tG*3;
                    eP[4] = __eG - (tG*15);
                    __eG = nG;
                } else {
                    __eG = eP[4];
                    eP[1] = eP[-2] = eP[4] = 0;
                }

                __eB = __wantB - __dB; 
                if (__eB) {
                    tB = __eB >> 4;
                    nB = eP[5] + (tB * 7);
                    eP[2] = tB*5;
                    eP[-1] = tB*3;
                    eP[5] = __eB - (tB*15);
                    __eB = nB;
                } else {
                    __eB = eP[5];
                    eP[2] = eP[-1] = eP[5] = 0;
                }

                eP += 3;
#endif
            }

            /*
             * allow for an interrupt after every row.
             * but care to refetch C variables
             */
            if (InterruptPending) {
                int d_srcP = srcP - __ByteArrayInstPtr(_INST(bytes))->ba_element;
                int d_dstP = dstP - __ByteArrayInstPtr(pseudoBits)->ba_element;
                int d_errP = errP - (short *) __ByteArrayInstPtr(error)->ba_element;

                __interrupt__();

                srcP = __ByteArrayInstPtr(_INST(bytes))->ba_element + d_srcP;
                dstP = __ByteArrayInstPtr(pseudoBits)->ba_element + d_dstP;
                rgbP = __ByteArrayInstPtr(rgbBytes)->ba_element;
                idP = __ByteArrayInstPtr(ditherIds)->ba_element;
                errP = (short *) __ByteArrayInstPtr(error)->ba_element + d_errP;
                __cube = __ArrayInstPtr(cube)->a_element;
            }
        }
    }
%}.
    failed ifTrue:[
        self primitiveFailed.
        ^ nil
    ].

    ^ pseudoBits
!

orderedDitheredGrayBitsDepth:depth
    "return the bitmap for a dithered depth-bitmap from the image"

    ^ self
        orderedDitheredGrayBitsWithDitherMatrix:(self class orderedDitherMatrixOfSize:8)
        ditherWidth:8
        depth:depth.

    "Created: 24.6.1997 / 22:20:12 / cg"
!

orderedDitheredGrayBitsWithDitherMatrix:ditherMatrix ditherWidth:dW depth:depth
    "return the bitmap for a dithered depth-bitmap from the image;
     with a constant ditherMatrix, this can be used for thresholding.
     Works for any source depths / photometric, 
     but very very slow since each pixel is processed individually.
     Redefined by some subclasses for more performance (D8Image)"

    |f last dH nDither   
     greyLevels greyValues greyPixels greyErrors
     dstIndex        "{Class: SmallInteger }"
     nextDst         
     bytesPerOutRow  "{Class: SmallInteger }"
     pixelsPerByte   "{Class: SmallInteger }"
     outBits
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" |

    depth > 8 ifTrue:[
        'IMAGE: unimplemented orderedDither conversion' errorPrintCR.
        ^ nil
    ].

    nDither := ditherMatrix size.
    dH := nDither / dW.

    w := width.
    h := height.

    greyLevels := 1 bitShift:depth.
    pixelsPerByte := 8 / depth.

    bytesPerOutRow := (w * depth + 7) // 8.
    outBits := ByteArray uninitializedNew:(bytesPerOutRow * h).
    (outBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    dstIndex := 1.

    self bitsPerPixel <= 12 ifTrue:[
        "/ fetch scaled brightness values outside of loop into a table;
        "/ use table-value in loop

        greyValues := self greyMapForRange:(greyLevels-1).
        greyPixels := greyValues collect:[:v | v isNil ifTrue:[
                                                   0
                                               ] ifFalse:[
                                                   v truncated]].
        greyPixels := ByteArray withAll:greyPixels.

        greyErrors := greyValues collect:[:v | v isNil ifTrue:[
                                                   0
                                               ] ifFalse:[
                                                   ((v - v truncated) * nDither) rounded
                                               ]].
        greyErrors := ByteArray withAll:greyErrors.

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerOutRow.
            byte := 0.
            bitCnt := 8.

            self valuesAtY:y from:0 to:(w-1) do:[:x :value |
%{
                int __dW = __intVal(dW);
                int __byte = __intVal(byte);
                int __value = __intVal(value);
                int __dT;
                int __dstIdx;
                int __pixel, __grey;
                int __bitCnt = __intVal(bitCnt);
                unsigned char *__greyPixels = __ByteArrayInstPtr(greyPixels)->ba_element;
                unsigned char *__greyErrors = __ByteArrayInstPtr(greyErrors)->ba_element;

                __pixel = __greyPixels[__value];
                __grey = __greyErrors[__value];

                __dT = __ByteArrayInstPtr(ditherMatrix)
                            ->ba_element[__intVal(x) % __dW 
                                         + (__intVal(y) % __intVal(dH)) * __dW];

                if (__grey > __dT) {
                    __pixel++;
                }
                __byte = (__byte << __intVal(depth)) | __pixel;

                __bitCnt = __bitCnt - __intVal(depth);
                if (__bitCnt == 0) {
                    __dstIdx = __intVal(dstIndex);
                    __ByteArrayInstPtr(outBits)->ba_element[__dstIdx-1] = __byte;
                    __dstIdx = __dstIdx + 1;
                    dstIndex = __MKSMALLINT(__dstIdx);
                    __byte = 0;
                    __bitCnt = 8;
                }
                byte = __MKSMALLINT(__byte);
                bitCnt = __MKSMALLINT(__bitCnt);
%}.
                0

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                outBits at:dstIndex put:byte.
            ].
            dstIndex := nextDst.
        ].
    ] ifFalse:[
        'Image [info]: slow ordered dither ..' infoPrintCR.

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerOutRow.
            byte := 0.
            bitCnt := 8.

            "/ this is the representaion independent (but slow)
            "/ inner loop - it extracts colors from the receiver
            
            self colorsAtY:y from:0 to:(w-1) do:[:x :clr |
                |dstClr grey dT pixel|

                "/ get the colors grey value [0 .. 1]
                grey := clr brightness.

                "/ remap into [0 .. greyLevels-1]
                grey := grey * (greyLevels-1).
            
                "/ get threshold pixel [0 .. greyLevels-1]

                pixel := grey truncated.  

                "/ compute the error [0..1]
                grey := grey - pixel.

                "/ map into dither space [0 .. nDither]
                grey := (grey * (nDither)) rounded.

%{
                int __dW = __intVal(dW);
                int __byte = __intVal(byte);
                int __dT;
                int __dstIdx;
                int __pixel;
                int __bitCnt = __intVal(bitCnt);

                __dT = __ByteArrayInstPtr(ditherMatrix)
                            ->ba_element[__intVal(x) % __dW 
                                         + (__intVal(y) % __intVal(dH)) * __dW];

                __pixel = __intVal(pixel);

                if (__intVal(grey) > __dT) {
                    __pixel++;
                }
                __byte = (__byte << __intVal(depth)) | __pixel;
            
                __bitCnt = __bitCnt - __intVal(depth);
                if (__bitCnt == 0) {
                    __dstIdx = __intVal(dstIndex);
                    __ByteArrayInstPtr(outBits)->ba_element[__dstIdx-1] = __byte;
                    __dstIdx = __dstIdx + 1;
                    dstIndex = __MKSMALLINT(__dstIdx);
                    __byte = 0;
                    __bitCnt = 8;
                }
                byte = __MKSMALLINT(__byte);
                bitCnt = __MKSMALLINT(__bitCnt);
%}.
                0

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                outBits at:dstIndex put:byte.
            ].
            dstIndex := nextDst.
        ].
    ].
    ^ outBits
!

orderedDitheredMonochromeBits
    "return the bitmap for a dithered monochrome bitmap from the image;
     using a default ditherMatrix."

    ^ self
        orderedDitheredMonochromeBitsWithDitherMatrix:(self class orderedDitherMatrixOfSize:8)
        ditherWidth:8

    "Created: 11.6.1996 / 16:48:57 / cg"
!

orderedDitheredMonochromeBitsWithDitherMatrix:ditherMatrix ditherWidth:dW
    "return the bitmap for a dithered monochrome bitmap from the image;
     with a constant ditherMatrix, this can be used for thresholding.
     Works for any source depths / photometric, 
     but very very slow since each pixel is processed individually.
     Redefined by some subclasses for more performance (D8Image)"

    |f last dH nDither   
     greyValues
     dstIndex        "{Class: SmallInteger }"
     nextDst         "{Class: SmallInteger }"
     bytesPerMonoRow "{Class: SmallInteger }"
     monoBits
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" |

    nDither := ditherMatrix size.
    dH := nDither / dW.

    w := width.
    h := height.

    bytesPerMonoRow := (w + 7) // 8.
    monoBits := ByteArray uninitializedNew:(bytesPerMonoRow * h).
    (monoBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    dstIndex := 1.

    self bitsPerPixel <= 12 ifTrue:[
        "/ fetch scaled brightness values outside of loop into a table;
        "/ use table-value in loop

        greyValues := self greyByteMapForRange:nDither.

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerMonoRow.
            byte := 0.
            bitCnt := 8.
            self valuesAtY:y from:0 to:(w-1) do:[:x :value |
%{
                int __dW = __intVal(dW);
                int __byte = __intVal(byte);
                int __dT;
                int __dstIdx;
                int __bitCnt = __intVal(bitCnt);
                int __grey;
                unsigned char *__greyValues = __ByteArrayInstPtr(greyValues)->ba_element;

                __grey = __greyValues[__intVal(value)];

                __dT = __ByteArrayInstPtr(ditherMatrix)
                            ->ba_element[__intVal(x) % __dW 
                                         + (__intVal(y) % __intVal(dH)) * __dW];

                __byte = __byte << 1;
                if (__grey > __dT) {
                    __byte = __byte | 1;            /* white */
                }
                __bitCnt = __bitCnt - 1;
                if (__bitCnt == 0) {
                    __dstIdx = __intVal(dstIndex);
                    __ByteArrayInstPtr(monoBits)->ba_element[__dstIdx-1] = __byte;
                    __dstIdx = __dstIdx + 1;
                    dstIndex = __MKSMALLINT(__dstIdx);
                    __byte = 0;
                    __bitCnt = 8;
                }
                byte = __MKSMALLINT(__byte);
                bitCnt = __MKSMALLINT(__bitCnt);
%}.
                0
            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                monoBits at:dstIndex put:byte.
            ].
            dstIndex := nextDst.
        ].
    ] ifFalse:[
        'Image [info]: slow ordered dither ..' infoPrintCR.

        0 to:(h-1) do:[:y |
            nextDst := dstIndex + bytesPerMonoRow.
            byte := 0.
            bitCnt := 8.
            self colorsAtY:y from:0 to:(w-1) do:[:x :clr |
                |dstClr grey dT|

                "/ get the colors grey value [0 .. 1]
                grey := clr brightness.

                "/ map into dither space [0 .. nDither]
                grey := (grey * (nDither)) rounded.

%{
                int __dW = __intVal(dW);
                int __byte = __intVal(byte);
                int __dT;
                int __dstIdx;
                int __bitCnt = __intVal(bitCnt);

                __dT = __ByteArrayInstPtr(ditherMatrix)
                            ->ba_element[__intVal(x) % __dW 
                                         + (__intVal(y) % __intVal(dH)) * __dW];

                __byte = __byte << 1;
                if (__intVal(grey) > __dT) {
                    __byte = __byte | 1;            /* white */
                }
                __bitCnt = __bitCnt - 1;
                if (__bitCnt == 0) {
                    __dstIdx = __intVal(dstIndex);
                    __ByteArrayInstPtr(monoBits)->ba_element[__dstIdx-1] = __byte;
                    __dstIdx = __dstIdx + 1;
                    dstIndex = __MKSMALLINT(__dstIdx);
                    __byte = 0;
                    __bitCnt = 8;
                }
                byte = __MKSMALLINT(__byte);
                bitCnt = __MKSMALLINT(__bitCnt);
%}.
                0

"/                dT := ditherMatrix at:(x \\ dW) + (y \\ dH * dW) + 1.
"/     
"/                byte := byte bitShift:1.
"/                grey < dT ifTrue:[
"/                    byte := byte bitOr:1.
"/                ].
"/                bitCnt := bitCnt - 1.
"/                bitCnt == 0 ifTrue:[
"/                    monoBits at:dstIndex put:byte.
"/                    dstIndex := dstIndex + 1.
"/                    byte := 0.
"/                    bitCnt := 8.
"/                ].

            ].
            bitCnt ~~ 8 ifTrue:[
                byte := byte bitShift:bitCnt.
                monoBits at:dstIndex put:byte.
            ].
            dstIndex := nextDst.
        ].
    ].
    ^ monoBits
!

stevensonArceDitheredMonochromeBits
    "return the bitmap for a dithered monochrome bitmap from the image.
     Works for any source depths / photometric"

    |dstIndex        "{Class: SmallInteger }"
     nextDst         "{Class: SmallInteger }"
     bytesPerMonoRow "{Class: SmallInteger }"
     monoBits greyValues
     errorArray
     errorArray1 errorArray2 errorArray3
     e t
     w               "{Class: SmallInteger }"
     h               "{Class: SmallInteger }"
     bitCnt          "{Class: SmallInteger }"
     byte            "{Class: SmallInteger }" 
     grey
     xE              "{Class: SmallInteger }" |

    self depth > 12 ifTrue:[
        ^ self floydSteinbergDitheredMonochromeBits
    ].

    w := width.
    h := height.

    bytesPerMonoRow := (w + 7) // 8.
    monoBits := ByteArray uninitializedNew:(bytesPerMonoRow * h).
    (monoBits isNil or:[bytes isNil]) ifTrue:[
        ^ nil
    ].

    errorArray := Array new:(w+6).
    errorArray1 := Array new:(w+6) withAll:0.
    errorArray2 := Array new:(w+6) withAll:0.
    errorArray3 := Array new:(w+6) withAll:0.

    dstIndex := 1.

    "/ fetch scaled brightness values outside of loop into a table;
    "/ use table-value in loop

    greyValues := self greyMapForRange:(255 * 1024).
    
    0 to:(h-1) do:[:y |
        nextDst := dstIndex + bytesPerMonoRow.
        byte := 0.
        bitCnt := 8.

        t := errorArray.
        errorArray := errorArray1.
        errorArray1 := errorArray2.
        errorArray2 := errorArray3.
        errorArray3 := t.

        errorArray3 atAllPut:0.

        self valuesAtY:y from:0 to:(w-1) do:[:x :pixel |
            |eP eD|

            "/ get the colors grey value [0 .. 1]
            grey := greyValues at:(pixel + 1).

            "/ adjust error
            xE := x + 3 + 1.
            grey := (grey + (errorArray at:xE)).

            byte := byte bitShift:1.
            grey > (127 * 1024) ifTrue:[
                byte := byte bitOr:1.      "/ white
                e := grey - (255 * 1024)
            ] ifFalse:[
                e := grey                  "/ black
            ].

            e ~= 0 ifTrue:[
                "/ distribute the error:
                "/                  XX    32
                "/         12    26    30    16
                "/            12    26    12
                "/          5    12    12     5

                eD := e.
                e := e // 200.

                eP := e * 32. eD := eD - eP.
                errorArray at:xE+2 put:(errorArray at:xE+2) + eP.

                eP := e * 30. eD := eD - eP.
                errorArray1 at:xE+1 put:(errorArray1 at:xE+1) + eP.

                eP := e * 16. eD := eD - eP.
                errorArray1 at:xE+3 put:(errorArray1 at:xE+3) + eP.

                eP := e * 26. eD := eD - eP.
                errorArray1 at:xE-1 put:(errorArray1 at:xE-1) + eP.

                eD := eD - eP.
                errorArray2 at:xE put:(errorArray2 at:xE) + eP.

                eP := e * 12. eD := eD - eP.
                errorArray1 at:xE-3 put:(errorArray1 at:xE-3) + eP.

                eD := eD - eP.
                errorArray2 at:xE-2 put:(errorArray2 at:xE-2) + eP.

                eD := eD - eP.
                errorArray2 at:xE+2 put:(errorArray2 at:xE+2) + eP.

                eD := eD - eP.
                errorArray3 at:xE-1 put:(errorArray3 at:xE-1) + eP.

                eD := eD - eP.
                errorArray3 at:xE+1 put:(errorArray3 at:xE+1) + eP.

                eP := e * 5. eD := eD - eP.
                errorArray3 at:xE-3 put:(errorArray3 at:xE-3) + eP.

                eP := eD.
                errorArray3 at:xE+3 put:(errorArray3 at:xE+3) + eP.
            ].

            bitCnt := bitCnt - 1.
            bitCnt == 0 ifTrue:[
                monoBits at:dstIndex put:byte.
                dstIndex := dstIndex + 1.
                byte := 0.
                bitCnt := 8.
            ].

        ].
        bitCnt ~~ 8 ifTrue:[
            byte := byte bitShift:bitCnt.
            monoBits at:dstIndex put:byte.
        ].

        dstIndex := nextDst.
    ].

    ^ monoBits

    "Created: 10.6.1996 / 12:38:35 / cg"
    "Modified: 10.6.1996 / 12:52:20 / cg"
! !

!Image methodsFor:'drawing'!

fillRectangle: aRectangle withColor:aColor
    "draw a rectangle with some pixel value.
    By using #atImageAndMask:put: it also works on images with mono masks."

    |xI "{ Class: SmallInteger }"
     yI "{ Class: SmallInteger }"
     wI "{ Class: SmallInteger }" 
     hI "{ Class: SmallInteger }"|

    xI := aRectangle left.
    yI := aRectangle top.
    wI := aRectangle width.
    hI := aRectangle height.

    yI to:yI+hI-1 do:[:yRun |
        xI to:xI+wI-1 do:[:xRun |
            self atImageAndMask:xRun@yRun put:aColor
        ]
    ]


!

fillRectangleX:x y:y width:w height:h with:aColor
    "fill a rectangular area with a aColor"

    self fillRectangleX:x y:y width:w height:h withValue:(self valueFromColor:aColor)
!

fillRectangleX:x y:y width:w height:h withValue:aPixelValue
    "fill a rectangular area with some pixel value.
     May be redefined in concrete subclasses for more performance, if req'd."

    |xI "{ Class: SmallInteger }"
     yI "{ Class: SmallInteger }"
     wI "{ Class: SmallInteger }" 
     hI "{ Class: SmallInteger }"|

    xI := x.
    yI := y.
    wI := w.
    hI := h.

    yI to:yI+hI-1 do:[:yRun |
        xI to:xI+wI-1 do:[:xRun |
            self pixelAtX:xRun y:yRun put:aPixelValue
        ]
    ]

    "Created: 22.4.1997 / 14:02:14 / cg"
    "Modified: 24.4.1997 / 17:24:58 / cg"
!

floodFillAt: aPoint withColor: aColor
    "fill a area with aColor like a flood up to surrounded pixels having different colors.
     By using #atImageAndMask:put: it also works on images with mono masks."

    |getSurroundingPixelsBlock detectedColor detectedPixels allDetectedPixels anotherPixels seemsToBeAll|

    getSurroundingPixelsBlock := 
    [:aPoint|
        |surroundingPixels|
        surroundingPixels := OrderedCollection new.
        (0 <= aPoint y) & (self height - 1 > aPoint y)
            ifTrue: [surroundingPixels add: (aPoint x@(aPoint y + 1))].
        (0 < aPoint y) & (self height - 1 >= aPoint y)
            ifTrue: [surroundingPixels add: (aPoint x@(aPoint y - 1))].
        (0 <= aPoint x) & (self width - 1 > aPoint x)
            ifTrue: [surroundingPixels add: (aPoint x + 1@(aPoint y))].
        (0 < aPoint x) & (self width - 1 >= aPoint x)
            ifTrue: [surroundingPixels add: (aPoint x - 1@(aPoint y))].
        surroundingPixels
    ].

    (mask notNil and: [(mask colorAt: aPoint) = Color black])
    ifTrue:
    [
        ^(mask floodFillAt: aPoint withColor: Color white) do: [:p| self atImageAndMask: p put: aColor].
    ].
    detectedColor := self colorAt: aPoint.
    detectedPixels := Set with: aPoint.
    allDetectedPixels := OrderedCollection with: aPoint.
    seemsToBeAll := false.
    [seemsToBeAll] whileFalse:
    [
        |sizeBefore|
        sizeBefore := allDetectedPixels size.
        anotherPixels := OrderedCollection new.
        detectedPixels do:
        [:p|
            (getSurroundingPixelsBlock value: p) select: [:sp| 
                (detectedColor = (self colorAt: sp) and: [mask isNil or: [(mask colorAt: sp) = Color white]])
                ifTrue: [anotherPixels add: sp]]
        ].
        detectedPixels := (anotherPixels asSet reject: [:p| allDetectedPixels includes: p]).
        allDetectedPixels addAll: detectedPixels.
        seemsToBeAll := sizeBefore = allDetectedPixels size.
    ]. 
    allDetectedPixels do: [:p| self atImageAndMask: p put: aColor].
    ^allDetectedPixels
!

rectangle: aRectangle withColor:aColor
    "draw a rectangle with some pixel value.
    By using #atImageAndMask:put: it also works on images with mono masks."

    |xI "{ Class: SmallInteger }"
     yI "{ Class: SmallInteger }"
     wI "{ Class: SmallInteger }" 
     hI "{ Class: SmallInteger }"|

    xI := aRectangle left.
    yI := aRectangle top.
    wI := aRectangle width.
    hI := aRectangle height.

    xI to:xI+wI-1 do:[:xRun|
        self atImageAndMask: xRun@yI put:aColor.
        self atImageAndMask: xRun@(yI+hI-1) put:aColor
    ].
    yI+1 to:yI+hI-2 do:[:yRun|
        self atImageAndMask: xI@yRun put:aColor.
        self atImageAndMask: xI+wI-1@yRun put:aColor
    ].


! !

!Image methodsFor:'enumerating'!

atY:y from:x1 to:x2 do:aBlock
    "perform aBlock for each pixel from x1 to x2 in row y.
     The block is passed the color at each pixel.
     The code here provides a generic and slow implementation, and
     should be redefined in concrete subclasses, to avoid some processing
     when going from pixel to pixel (i.e. the byte-index and mask computations
     and also the color allocation)."

    self obsoleteMethodWarning:'use #colorsAtY:from:to:do:'.
    self colorsAtY:y from:x1 to:x2 do:aBlock

    "Modified: 7.6.1996 / 19:13:30 / cg"
!

colorsAtY:y from:x1 to:x2 do:aBlock
    "perform aBlock for each pixel from x1 to x2 in row y.
     The block is passed the color at each pixel.
     The code here provides a generic and slow implementation, and
     should be redefined in concrete subclasses, to avoid some processing
     when going from pixel to pixel (i.e. the byte-index and mask computations
     and also the color allocation)."

    |xStart "{Class: SmallInteger }"
     xEnd   "{Class: SmallInteger }"|

    xStart := x1.
    xEnd := x2.
    xStart to:xEnd do:[:xRun |
        aBlock value:xRun value:(self atX:xRun y:y)
    ]

    "Created: 7.6.1996 / 19:12:51 / cg"
    "Modified: 10.6.1996 / 10:27:29 / cg"
!

colorsFromX:xStart y:yStart toX:xEnd y:yEnd do:aBlock
    "perform aBlock for each color in a rectangular area of the image.
     Notice, that x and y coordinates start at 0@0 for the upper left corner.
     The block is passed the x and y coordinates and pixelValue at each pixel.
     The code here provides a generic and slow implementation, and
     should be redefined in concrete subclasses, to avoid some processing
     when going from pixel to pixel (i.e. the byte-index and mask computations,
     and especially, the color allocations)."

    |yS "{Class: SmallInteger }"
     yE "{Class: SmallInteger }"
     yR|

    yS := yStart.
    yE := yEnd.

    yS to:yE do:[:yRun |    
        yR := yRun.
        self colorsAtY:yRun from:xStart to:xEnd do:[:xRun :color |
            aBlock value:xRun value:yR value:color 
        ]
    ]

    "Modified: 11.7.1996 / 19:50:47 / cg"
!

valueAtY:y from:x1 to:x2 do:aBlock
    "perform aBlock for each pixelValue from x1 to x2 in row y.
     Obsolete - remains for backward compatibility."

    self obsoleteMethodWarning:'use #valuesAtY:from:to:do:'.
    self valuesAtY:y from:x1 to:x2 do:aBlock

    "Modified: 7.6.1996 / 19:11:06 / cg"
!

valuesAtY:y from:x1 to:x2 do:aBlock
    "WARNING: for now, this enumerates pixel values 
     (backward compatibility with ST/X)
     In the future, this will enumerate colors 
     Use #pixelAtT:from:to:do: - for future compatibility.

     perform aBlock for each pixelValue from x1 to x2 in row y.
     Notice, that x and y coordinates start at 0@0 for the upper left corner.
     The block is passed the x coordinate and the pixelValue at each pixel.
     (see also Image>>atY:from:to:do:).
     The code here provides a generic and slow implementation, and
     should be redefined in concrete subclasses, to avoid some processing
     when going from pixel to pixel (i.e. the byte-index and mask computations)."

    |xStart "{Class: SmallInteger }"
     xEnd   "{Class: SmallInteger }"|

    xStart := x1.
    xEnd := x2.
    xStart to:xEnd do:[:xRun |
        aBlock value:xRun value:(self pixelAtX:xRun y:y)
    ]

    "Created: 7.6.1996 / 19:09:51 / cg"
    "Modified: 24.4.1997 / 16:55:38 / cg"
!

valuesFromX:xStart y:yStart toX:xEnd y:yEnd do:aBlock
    "perform aBlock for each pixelValue in a rectangular area of the image.
     Notice, that x and y coordinates start at 0@0 for the upper left corner.
     The block is passed the x and y coordinates and pixelValue at each pixel.
     The code here provides a generic and slow implementation, and
     should be redefined in concrete subclasses, to avoid some processing
     when going from pixel to pixel (i.e. the byte-index and mask computations)."

    |xS "{Class: SmallInteger }"
     xE "{Class: SmallInteger }"|

    xS := xStart.
    xE := xEnd.
    yStart to:yEnd do:[:yRun |    
        self valuesAtY:yRun from:xStart to:xEnd do:[:xRun :pixel |
            aBlock value:xRun value:yRun value:pixel
        ]
    ]

    "Modified: 7.6.1996 / 19:09:29 / cg"
! !

!Image methodsFor:'image manipulations'!

applyPixelValuesTo:pixelFunctionBlock into:newImage
    "helper for withPixelFunctionAppliedToValues:
     enumerate pixelValues and evaluate the block for each.
     To be redefined by subclasses for better performance."

    |w  "{Class: SmallInteger }"
     h  "{Class: SmallInteger }"
     newPixel newPixelRow pixelRow|

    w := width.
    h := height - 1.

    newPixelRow := Array new:width.
    pixelRow := self pixelArraySpecies new:width.

    0 to:h do:[:y |
        self rowAt:y into:pixelRow.
        1 to:w do:[:col |
            newPixel := pixelFunctionBlock
                            value:self
                            value:(pixelRow at:col)
                            value:(col-1)
                            value:y.
            newPixelRow at:col put:newPixel.
        ].
        newImage rowAt:y putAll:newPixelRow
    ].

    "Modified: 24.4.1997 / 16:18:31 / cg"
!

colorMapProcessing:aBlock
    "a helper for all kinds of colormap manipulations.
     The argument, aBlock is called for every colormap entry, 
     and the returned value will replace the original entry in the map.
     This will fail for non-palette images.
     See examples in Image>>copyWithColorMapProcessing:"

    |nColors "{ Class: SmallInteger }"|

    colorMap isNil ifTrue:[
        ^ self error:'image has no colormap'
    ].

    nColors := colorMap size.
    1 to:nColors do:[:index |
        |clr|

        clr := colorMap at:index.
        clr notNil ifTrue:[
            colorMap at:index put:(aBlock value:clr)
        ]
    ]

    "Modified: 23.4.1996 / 11:13:55 / cg"
!

copyWithColorMapProcessing:aBlock
    "a helper to create & return new images based on the receiver with
     some colorMap processing. The receiver is copied, and the copied images
     colormap is modified by replacing entries with the result of the processing block,
     which is called with the original color values. The block is supposed to return
     a color.
     CAVEAT: this only works with palette images (i.e. not for rgb or greyScale)"

    |newImage|

    newImage := self copy.
    newImage colorMap isNil ifTrue:[
        self error:'no colormap in image'.
        ^ nil
    ].

    "
     the code below manipulates the colormap.
     For non-palette images, special code is required
    "
    newImage colorMapProcessing:aBlock.
    ^ newImage

    "
     leave red component only:

     (Image fromFile:'bitmaps/gifImages/claus.gif') 
        copyWithColorMapProcessing:[:clr | Color red:(clr red) green:0 blue:0] 
    "

    "
     make it reddish:

     (Image fromFile:'bitmaps/gifImages/claus.gif') 
        copyWithColorMapProcessing:[:clr | Color red:((clr red * 2) min:100) green:clr green blue:clr blue] 
    "

    "
     invert:

     (Image fromFile:'bitmaps/gifImages/claus.gif') 
        copyWithColorMapProcessing:[:clr | Color red:(100 - clr red) green:(100 - clr green) blue:(100 - clr green)] 
    "

    "
     lighter:

     (Image fromFile:'bitmaps/gifImages/claus.gif') 
        copyWithColorMapProcessing:[:clr | |r g b|
                                                r := clr red.  g := clr green.  b := clr blue.
                                                Color red:(r + (100-r//2)) 
                                                      green:(g + (100-g//2)) 
                                                      blue:(b + (100-b//2))]
    "

    "
     darker:

     (Image fromFile:'bitmaps/gifImages/claus.gif') 
        copyWithColorMapProcessing:[:clr | Color red:(clr red//2) green:(clr green // 2) blue:(clr blue // 2)] 
    "

    "Modified: 24.4.1997 / 18:28:05 / cg"
!

darkened
    "return a new image which is slightly darker than the receiver.
     The receiver must be a palette image (currently).
     CAVEAT: this only works with palette images (i.e. not for rgb or greyScale).
     CAVEAT: Need an argument, which specifies by how much it should be darker."

     ^ self 
        copyWithColorMapProcessing:[:clr | clr darkened]

    "
     (Image fromFile:'bitmaps/gifImages/claus.gif') inspect
     (Image fromFile:'bitmaps/gifImages/claus.gif') darkened inspect
    "

    "Modified: 24.4.1997 / 18:28:31 / cg"
!

easyRotateBitsInto:destinationImage angle:degrees
    "helper for rotation - does the actual pixel shuffling.
     by degrees clockwise. Here, only 90, 180 and 270 degrees
     are implemented. Hard angles are done in #hardRotate:.
     The code here is depth-independent (but not too fast);
     can be redefined in subclasses for more performance"

    |w  "{Class: SmallInteger }"
     h  "{Class: SmallInteger }"
     c2 r2 pixelArray|

    w := width - 1.
    h := height - 1.

    pixelArray := self pixelArraySpecies new:width.

    degrees = 90 ifTrue:[
        0 to:h do:[:row |
            self rowAt:row into:pixelArray startingAt:1.
            c2 := h-row.
            0 to:w do:[:col |
                destinationImage pixelAtX:c2 y:col put:(pixelArray at:(col+1)).
            ]
        ].
        ^ self.
    ].

    degrees = 180 ifTrue:[
        0 to:h do:[:row |
            self rowAt:row into:pixelArray startingAt:1.
            r2 := h - row.
            0 to:w do:[:col |
                destinationImage pixelAtX:(w-col) y:r2 put:(pixelArray at:(col+1)).
            ]
        ].
        ^ self
    ].

    "/ degrees = 270
    0 to:h do:[:row |
        self rowAt:row into:pixelArray startingAt:1.
        0 to:w do:[:col |
            destinationImage pixelAtX:row y:(w-col) put:(pixelArray at:(col+1)).
        ]
    ]

    "Created: 23.4.1997 / 14:36:45 / cg"
    "Modified: 24.4.1997 / 17:26:26 / cg"
!

flipHorizontal
    "destructively inplace horizontal flip"

    |h  "{Class: SmallInteger }"
     pixelArray|

    h := height - 1.

    pixelArray := self pixelArraySpecies new:width.

    0 to:h do:[:row |
        self rowAt:row into:pixelArray.
        pixelArray reverse.
        self rowAt:row putAll:pixelArray.
    ].

    mask notNil ifTrue:[
        mask flipHorizontal
    ].

    "/ flush device info
    self restored

    "
     (Image fromFile:'bitmaps/gifImages/garfield.gif') flipHorizontal inspect
    "

    "Modified: 24.4.1997 / 18:29:13 / cg"
!

flipVertical
    "inplace vertical flip"

    |h           "{Class: SmallInteger }"
     bytesPerRow "{Class: SmallInteger }"
     buffer 
     indexLow    "{Class: SmallInteger }"
     indexHi     "{Class: SmallInteger }"|

    bytesPerRow := self bytesPerRow.
    buffer := ByteArray new:bytesPerRow.

    h := height - 1.

    indexLow := 1.
    indexHi := bytesPerRow * h + 1.

    0 to:(h // 2) do:[:row |
        buffer replaceFrom:1 to:bytesPerRow with:bytes startingAt:indexLow.
        bytes replaceFrom:indexLow to:(indexLow + bytesPerRow - 1) with:bytes startingAt:indexHi.
        bytes replaceFrom:indexHi to:(indexHi + bytesPerRow - 1) with:buffer startingAt:1.
        indexLow := indexLow + bytesPerRow.
        indexHi := indexHi - bytesPerRow.
    ].
    mask notNil ifTrue:[
        mask flipVertical
    ].
    "flush device info"
    self restored

    "
     (Image fromFile:'bitmaps/gifImages/garfield.gif') flipVertical inspect
    "

    "Modified: 24.4.1997 / 18:29:36 / cg"
!

hardAntiAliasedMagnifiedBy:scalePoint
    "return a new image magnified and antiAliased by scalePoint, aPoint.
     This converts into a depth24Image before doing the antiAlias-magnify.
     It is definitely slower than the non antiAliasing/integral magnification methods."

    ^ (Depth24Image fromImage:self)
        hardAntiAliasedMagnifiedBy:scalePoint

    "Modified: 2.6.1997 / 13:19:57 / cg"
    "Created: 2.6.1997 / 15:53:34 / cg"
!

hardMagnifiedBy:scalePoint
    "return a new image magnified by scalePoint, aPoint.
     This is the general magnification method, handling non-integral values.
     It is slower than the integral magnification method."

    |mX        
     mY        
     newWidth  "{ Class: SmallInteger }"
     newHeight "{ Class: SmallInteger }"
     w         "{ Class: SmallInteger }"
     h         "{ Class: SmallInteger }"
     newImage newBits bitsPerPixel newBytesPerRow newMask
     value 
     srcRow pixelArray|

    mX := scalePoint x.
    mY := scalePoint y.
    ((mX < 0) or:[mY < 0]) ifTrue:[^ nil].
    ((mX = 1) and:[mY = 1]) ifTrue:[^ self].

    newWidth := (width * mX) truncated.
    newHeight := (height * mY) truncated.

    bitsPerPixel := self depth.
    newBytesPerRow := ((newWidth * bitsPerPixel) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * newHeight).

    mask notNil ifTrue:[
        newMask := (mask magnifiedBy:scalePoint)
    ].

    newImage := self species new.
    newImage 
        width:newWidth 
        height:newHeight 
        photometric:photometric 
        samplesPerPixel:samplesPerPixel 
        bitsPerSample:bitsPerSample 
        colorMap:colorMap copy
        bits:newBits
        mask:newMask.

    "walk over destination image fetching pixels from source image"

    w := newWidth - 1.
    h := newHeight - 1.
    pixelArray := newImage pixelArraySpecies new:newWidth.

    0 to:h do:[:row |
        srcRow := (row // mY).
        0 to:w do:[:col |
            value := self pixelAtX:(col // mX) y:srcRow.
            pixelArray at:(col+1) put:value.
        ].
        newImage rowAt:row putAll:pixelArray.
    ].

    ^ newImage

    "
     |i|
     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     Time millisecondsToRun:[
         i := i hardMagnifiedBy:0.5@0.5
     ].
     i
    "

    "Modified: 24.4.1997 / 18:30:24 / cg"
!

hardRotated:degrees
    "return a new image from the old one, by rotating the image
     degrees clockwise. 
     Warning: the returned image will be larger than the original image."

    |p r a aN p1 p2 p3 p4 maxX minX maxY minY
     newImage 
     newWidth  "{ Class: SmallInteger }"
     newHeight "{ Class: SmallInteger }"
     newBytesPerRow newBits
     blackPixel halfW halfH radians m t bad
     bytesPerRow myDepth maskBits
     pX pY srcX srcY pix nX nY 
     sinRot cosRot sinPY cosPY|

    "/ placing the image at the origin,
    "/ compute the diagonal and angle.

    p := (width - 1 / 2) @ (height - 1 / 2).
    r := p r.
    a := p angle.

    "/ add the rotation 
    "/ (sight - subtract, we defined things clockwise ... 
    "/  ... in contrast to point which thinks counter-clockwise)

    aN := a - degrees.

    "/ compute new corner points
    p1 := Point r:r angle:aN.         "/ rotated topRight
    p2 := Point r:r angle:aN-a-a.     "/ rotated bottomRight
    p3 := p1 * -1.                    "/ rotated bottomLeft
    p4 := p2 * -1.                    "/ rotated topLeft

    "/ compute the boundary of the new image

    maxX := minX := p1 x.
    (t := p2 x) > maxX ifTrue:[
        maxX := t
    ] ifFalse:[
        t < minX ifTrue:[minX := t].
    ].
    (t := p3 x) > maxX ifTrue:[
        maxX := t
    ] ifFalse:[
        t < minX ifTrue:[minX := t].
    ].
    (t := p4 x) > maxX ifTrue:[
        maxX := t
    ] ifFalse:[
        t < minX ifTrue:[minX := t].
    ].

    maxY := minY := p1 y.
    (t := p2 y) > maxY ifTrue:[
        maxY := t
    ] ifFalse:[
        t < minY ifTrue:[minY := t].
    ].
    (t := p3 y) > maxY ifTrue:[
        maxY := t
    ] ifFalse:[
        t < minY ifTrue:[minY := t].
    ].
    (t := p4 y) > maxY ifTrue:[
        maxY := t
    ] ifFalse:[
        t < minY ifTrue:[minY := t].
    ].


    newWidth := (maxX - minX) rounded + 1.
    newHeight := (maxY - minY) rounded + 1.

    newImage := self species new.
    newImage width:newWidth.
    newImage height:newHeight.
    newBytesPerRow := newImage bytesPerRow.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * newHeight).
    newImage bits:newBits.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.
    newImage maskedPixelsAre0:maskedPixelsAre0.
    mask notNil ifTrue:[
        newImage mask:(mask rotated:degrees)
    ] ifFalse:[
        self isMask ifFalse:[
            self depth ~~ 1 ifTrue:[
                m := ImageMask width:width height:height.
                m bits:(maskBits := ByteArray new:(m bytesPerRow * height)).
                maskBits atAllPut:16rFF.
                newImage mask:(m rotated:degrees)
            ]
        ]
    ].

    maskedPixelsAre0 ifTrue:[
        blackPixel := 0.
    ] ifFalse:[
        blackPixel := self valueFromColor:Color black.
        blackPixel isNil ifTrue:[
            blackPixel := self valueFromColor:Color white.
            blackPixel isNil ifTrue:[
                blackPixel := 0.
            ]
        ]
    ].
    self isMask ifTrue:[
        blackPixel := 0.
    ].

    myDepth := self depth.
    newBits atAllPut:0.

    "/ now, walk over destination pixels,
    "/ fetching from source. 
    "/ (if we walked over the source, we could get holes
    "/  in the destination image ...)

    halfW := (width - 1) / 2.0.
    halfH := (height - 1) / 2.0.

    radians := degrees degreesToRadians.

    bytesPerRow := self bytesPerRow.

%{
    int __newHeight = __intVal(newHeight);
    int __newWidth = __intVal(newWidth);
    int __height = __intVal(__INST(height));
    int __width = __intVal(__INST(width));
    double __minX, __minY, __radians;
    double __halfW, __halfH;
    int __dstX, __dstY;
    int __depth = __intVal(myDepth);
    unsigned char *__srcBytes, *__dstBytes;
    unsigned char *__dstPtr, *__dstRowPtr, *__dstEndPtr;
    int __nSrcBytes, __nDstBytes;
    int __srcBytesPerRow = __intVal(bytesPerRow);
    int __dstBytesPerRow = __intVal(newBytesPerRow);
    int __dstMask, __blackPixel;
    double __sin, __cos;
    double sin(), cos();
#   define Float_PI 3.14159

    bad = true;
    if (1
     && __isFloat(minX) 
     && __isFloat(minY)
     && __isFloat(radians)
     && __isFloat(halfW)
     && __isFloat(halfH)
     && __isByteArray(newBits)
     && __isByteArray(__INST(bytes))) {
        __srcBytes = __ByteArrayInstPtr(__INST(bytes))->ba_element;
        __dstBytes = __ByteArrayInstPtr(newBits)->ba_element;
        __nSrcBytes = __byteArraySize(__INST(bytes));
        __nDstBytes = __byteArraySize(newBits);
        __blackPixel = __intVal(blackPixel);

        __radians = __floatVal(radians);
        __radians = -__radians; /* sigh: clock-wise */
        __sin = sin(__radians);
        __cos = cos(__radians);
        __minX = __floatVal(minX);
        __minY = __floatVal(minY);
        __halfW = __floatVal(halfW);
        __halfH = __floatVal(halfH);

        __dstRowPtr = __dstBytes;
        __dstEndPtr = __dstBytes + __nDstBytes;

#       define EARLY_OUT
#       define FAST_ADVANCE 5
#       define FAST_ADVANCE2

        switch (__depth) {
            case 8:
                for (__dstY = 0; __dstY < __newHeight; __dstY++) {
                    double __pY, __sinPY, __cosPY;
#ifdef EARLY_OUT
                    int didFetchInRow = 0;
#endif
                    __pY = (double)(__dstY + __minY);

                    __sinPY = __sin * __pY;
                    __cosPY = __cos * __pY;

                    __dstPtr = __dstRowPtr;
                    __dstRowPtr += __dstBytesPerRow;

                    for (__dstX = 0; __dstX < __newWidth; __dstX++) {
                        double __pX, __nX;
                        unsigned __pix;

                        /* translate X in destination (center to 0/0) */
                        __pX = (double)(__dstX + __minX);
                        /* rotate X */
                        __nX = (__cos * __pX) - __sinPY;

                        /* translate X in source (origin to 0/0) */
                        __nX = __nX + __halfW + 0.5;

                        /* inside ? */
                        if (__nX < 0) {
#ifdef EARLY_OUT
                            if (didFetchInRow) {
                                break;
                            }
#endif
#ifdef FAST_ADVANCE
                            if (__blackPixel == 0) {
                                do {
                                    /* try advance by FAST_ADVANCE pixels ... */
                                    __dstX += FAST_ADVANCE; __dstPtr += FAST_ADVANCE;
                                    if (__dstX >= __newWidth) {
                                        break;
                                    }
                                    __pX = (double)(__dstX + __minX);
                                    __nX = (__cos * __pX) - __sinPY;
                                    __nX = __nX + __halfW + 0.5;
                                } while (__nX < 0);
                                __dstX -= FAST_ADVANCE; __dstPtr -= FAST_ADVANCE;
                            }
#endif
                            __pix = __blackPixel;
                        } else {
                            int __srcX;

                            __srcX = (int)__nX;
                            /* inside ? */
                            if (__srcX >= __width) {
#ifdef EARLY_OUT
                                if (didFetchInRow) {
                                    break;
                                }
#endif
#ifdef FAST_ADVANCE2
                                if (__blackPixel == 0) {
                                    do {
                                        /* try advance by FAST_ADVANCE pixels ... */
                                        __dstX += FAST_ADVANCE; __dstPtr += FAST_ADVANCE;
                                        if (__dstX >= __newWidth) {
                                            break;
                                        }
                                        __pX = (double)(__dstX + __minX);
                                        __nX = (__cos * __pX) - __sinPY;
                                        __nX = __nX + __halfW + 0.5;
                                        __srcX = (int)__nX;
                                    } while (__srcX >= __width);
                                    __dstX -= FAST_ADVANCE; __dstPtr -= FAST_ADVANCE;
                                }
#endif
                                __pix = __blackPixel;
                            } else {
                                double __nY;

                                /* rotate Y */
                                __nY = (__sin * __pX) + __cosPY;
                                /* translate Y in source (origin to 0/0) */
                                __nY = __nY + __halfH + 0.5;

                                /* inside ? */
                                if (__nY < 0) {
#ifdef EARLY_OUT
                                    if (didFetchInRow) {
                                        break;
                                    }
#endif
#ifdef FAST_ADVANCE2
                                    if (__blackPixel == 0) {
                                        do {
                                            /* try advance by FAST_ADVANCE pixels ... */
                                            __dstX += FAST_ADVANCE; __dstPtr += FAST_ADVANCE;
                                            if (__dstX >= __newWidth) {
                                                break;
                                            }
                                            __pX = (double)(__dstX + __minX);
                                            __nY = (__sin * __pX) + __cosPY;
                                            __nY = __nY + __halfH + 0.5;
                                        } while (__nY < 0);
                                        __dstX -= FAST_ADVANCE; __dstPtr -= FAST_ADVANCE;
                                    }
#endif
                                    __pix = __blackPixel;
                                } else {
                                    int __srcY;

                                    __srcY = (int)__nY;
                                    /* inside ? */
                                    if (__srcY >= __height) {
#ifdef EARLY_OUT
                                        if (didFetchInRow) {
                                            break;
                                        }
#endif
#ifdef FAST_ADVANCE
                                        if (__blackPixel == 0) {
                                            do {
                                                /* try advance by FAST_ADVANCE pixels ... */
                                                __dstX += FAST_ADVANCE; __dstPtr += FAST_ADVANCE;
                                                if (__dstX >= __newWidth) {
                                                    break;
                                                }
                                                __pX = (double)(__dstX + __minX);
                                                __nY = (__sin * __pX) + __cosPY;
                                                __nY = __nY + __halfH + 0.5;
                                                __srcY = (int)__nY;
                                            } while (__srcY >= __height);
                                            __dstX -= FAST_ADVANCE; __dstPtr -= FAST_ADVANCE;
                                        }
#endif
                                        __pix = __blackPixel;
                                    } else {
                                        /* fetch source pixel */

                                        int idx;
#ifdef EARLY_OUT
                                        didFetchInRow = 1;
#endif
                                        idx = __srcY * __srcBytesPerRow + __srcX;
                                        if ((unsigned)idx < __nSrcBytes) {
                                            __pix = __srcBytes[idx];
                                        } else {
                                            __pix = __blackPixel;
                                        }
                                    }
                                }
                            }
                        }

                        if (__pix != 0) {
                            *__dstPtr = __pix;
                        }
                        __dstPtr++;
                    }
                }
                break;

            case 1:
                for (__dstY = 0; __dstY < __newHeight; __dstY++) {
                    double __pY, __sinPY, __cosPY;
#ifdef EARLY_OUT
                    int didFetchInRow = 0;
#endif
                    __pY = (double)(__dstY + __minY);

                    __sinPY = __sin * __pY;
                    __cosPY = __cos * __pY;

                    __dstPtr = __dstRowPtr;
                    __dstMask = 0x80;
                    __dstRowPtr += __dstBytesPerRow;

                    for (__dstX = 0; __dstX < __newWidth; __dstX++) {
                        double __pX, __nX;
                        int __pix;

                        /* translate X in destination (center to 0/0) */
                        __pX = (double)(__dstX + __minX);
                        /* rotate X */
                        __nX = (__cos * __pX) - __sinPY;

                        /* translate X in source (origin to 0/0) */
                        __nX = __nX + __halfW + 0.5;

                        /* inside ? */
                        if (__nX < 0) {
#ifdef EARLY_OUT
                            if (didFetchInRow) {
                                break;
                            }
#endif
#ifdef FAST_ADVANCE
                            if (__blackPixel == 0) {
                                do {
                                    /* try advance by 8 pixels ... */
                                    __dstX += 8; __dstPtr ++;
                                    if (__dstX >= __newWidth) {
                                        break;
                                    }
                                    __pX = (double)(__dstX + __minX);
                                    __nX = (__cos * __pX) - __sinPY;
                                    __nX = __nX + __halfW + 0.5;
                                } while (__nX < 0);
                                __dstX -= 8; __dstPtr--;
                            }
#endif
                            __pix = __blackPixel;
                        } else {
                            int __srcX;

                            __srcX = (int)__nX;
                            /* inside ? */
                            if (__srcX >= __width) {
#ifdef EARLY_OUT
                                if (didFetchInRow) {
                                    break;
                                }
#endif
#ifdef FAST_ADVANCE2
                                if (__blackPixel == 0) {
                                    do {
                                        /* try advance by 8 pixels ... */
                                        __dstX += 8; __dstPtr++;
                                        if (__dstX >= __newWidth) {
                                            break;
                                        }
                                        __pX = (double)(__dstX + __minX);
                                        __nX = (__cos * __pX) - __sinPY;
                                        __nX = __nX + __halfW + 0.5;
                                        __srcX = (int)__nX;
                                    } while (__srcX >= __width);
                                    __dstX -= 8; __dstPtr--;
                                }
#endif
                                __pix = __blackPixel;
                            } else {
                                double __nY;

                                /* rotate Y */
                                __nY = (__sin * __pX) + __cosPY;
                                /* translate Y in source (origin to 0/0) */
                                __nY = __nY + __halfH + 0.5;

                                /* inside ? */
                                if (__nY < 0) {
#ifdef EARLY_OUT
                                    if (didFetchInRow) {
                                        break;
                                    }
#endif
#ifdef FAST_ADVANCE2
                                    if (__blackPixel == 0) {
                                        do {
                                            /* try advance by 8 pixels ... */
                                            __dstX += 8; __dstPtr++;
                                            if (__dstX >= __newWidth) {
                                                break;
                                            }
                                            __pX = (double)(__dstX + __minX);
                                            __nY = (__sin * __pX) + __cosPY;
                                            __nY = __nY + __halfH + 0.5;
                                        } while (__nY < 0);
                                        __dstX -= 8; __dstPtr--;
                                    }
#endif
                                    __pix = __blackPixel;
                                } else {
                                    int __srcY;

                                    __srcY = (int)__nY;
                                    /* inside ? */
                                    if (__srcY >= __height) {
#ifdef EARLY_OUT
                                        if (didFetchInRow) {
                                            break;
                                        }
#endif
#ifdef FAST_ADVANCE
                                        if (__blackPixel == 0) {
                                            do {
                                                /* try advance by 8 pixels ... */
                                                __dstX += 8; __dstPtr++;
                                                if (__dstX >= __newWidth) {
                                                    break;
                                                }
                                                __pX = (double)(__dstX + __minX);
                                                __nY = (__sin * __pX) + __cosPY;
                                                __nY = __nY + __halfH + 0.5;
                                                __srcY = (int)__nY;
                                            } while (__srcY >= __height);
                                            __dstX -= 8; __dstPtr--;
                                        }
#endif
                                        __pix = __blackPixel;
                                    } else {
                                        /* fetch source pixel */

                                        int idx, pV;
#ifdef EARLY_OUT
                                        didFetchInRow = 1;
#endif
                                        idx = __srcY * __srcBytesPerRow + (__srcX >> 3);
                                        if ((unsigned)idx < __nSrcBytes) {
                                            pV = __srcBytes[idx];
                                            __pix = (pV & (0x80 >> (__srcX & 7))) ? 1 : 0;
                                        } else {
                                            __pix = __blackPixel;
                                        }
                                    }
                                }
                            }
                        }

                        /* store pixel */
                        if (__pix != 0) {
                            *__dstPtr |= __dstMask;
                        }
                        __dstMask >>= 1;
                        if (__dstMask == 0) {
                            __dstMask = 0x80;
                            __dstPtr++;
                        }
                    }
                }
                break;

            case 24:
                for (__dstY = 0; __dstY < __newHeight; __dstY++) {
                    double __pY, __sinPY, __cosPY;
#ifdef EARLY_OUT
                    int didFetchInRow = 0;
#endif
                    __pY = (double)(__dstY + __minY);

                    __sinPY = __sin * __pY;
                    __cosPY = __cos * __pY;

                    __dstPtr = __dstRowPtr;
                    __dstRowPtr += __dstBytesPerRow;

                    for (__dstX = 0; __dstX < __newWidth; __dstX++) {
                        double __pX, __nX;
                        unsigned __pix;

                        /* translate X in destination (center to 0/0) */
                        __pX = (double)(__dstX + __minX);
                        /* rotate X */
                        __nX = (__cos * __pX) - __sinPY;

                        /* translate X in source (origin to 0/0) */
                        __nX = __nX + __halfW + 0.5;

                        /* inside ? */
                        if (__nX < 0) {
#ifdef EARLY_OUT
                            if (didFetchInRow) {
                                break;
                            }
#endif
                            __pix = __blackPixel;
                        } else {
                            int __srcX;

                            __srcX = (int)__nX;
                            /* inside ? */
                            if (__srcX >= __width) {
#ifdef EARLY_OUT
                                if (didFetchInRow) {
                                    break;
                                }
#endif
                                __pix = __blackPixel;
                            } else {
                                double __nY;

                                /* rotate Y */
                                __nY = (__sin * __pX) + __cosPY;
                                /* translate Y in source (origin to 0/0) */
                                __nY = __nY + __halfH + 0.5;

                                /* inside ? */
                                if (__nY < 0) {
#ifdef EARLY_OUT
                                    if (didFetchInRow) {
                                        break;
                                    }
#endif
                                    __pix = __blackPixel;
                                } else {
                                    int __srcY;

                                    __srcY = (int)__nY;
                                    /* inside ? */
                                    if (__srcY >= __height) {
#ifdef EARLY_OUT
                                        if (didFetchInRow) {
                                            break;
                                        }
#endif
                                        __pix = __blackPixel;
                                    } else {
                                        /* fetch source pixel */

                                        int idx;
#ifdef EARLY_OUT
                                        didFetchInRow = 1;
#endif
                                        idx = __srcY * __srcBytesPerRow + __srcX + __srcX + __srcX;
                                        if ((unsigned)idx < __nSrcBytes) {
                                            __pix = __srcBytes[idx];
                                            __pix = (__pix<<8) | __srcBytes[idx+1];
                                            __pix = (__pix<<8) | __srcBytes[idx+2];
                                        } else {
                                            __pix = __blackPixel;
                                        }
                                    }
                                }
                            }
                        }

                        /* store pixel */
                        if (__pix != 0) {
                            __dstPtr[0] = (__pix >> 16 & 0xFF);
                            __dstPtr[1] = (__pix >> 8) & 0xFF;
                            __dstPtr[2] = __pix & 0xFF;
                        }
                        __dstPtr += 3;
                    }
                }
                break;

            default:
                for (__dstY = 0; __dstY < __newHeight; __dstY++) {
                    double __pY, __sinPY, __cosPY;
#ifdef EARLY_OUT
                    int didFetchInRow = 0;
#endif
                    __pY = (double)(__dstY + __minY);

                    __sinPY = __sin * __pY;
                    __cosPY = __cos * __pY;

                    __dstPtr = __dstRowPtr;
                    __dstMask = 0x80;
                    __dstRowPtr += __dstBytesPerRow;

                    for (__dstX = 0; __dstX < __newWidth; __dstX++) {
                        double __pX, __nX;
                        OBJ __pix;

                        /* translate X in destination (center to 0/0) */
                        __pX = (double)(__dstX + __minX);
                        /* rotate X */
                        __nX = (__cos * __pX) - __sinPY;

                        /* translate X in source (origin to 0/0) */
                        __nX = __nX + __halfW + 0.5;

                        /* inside ? */
                        if (__nX < 0) {
#ifdef EARLY_OUT
                            if (didFetchInRow) {
                                break;
                            }
#endif
                            __pix = blackPixel;
                        } else {
                            int __srcX;

                            __srcX = (int)__nX;
                            /* inside ? */
                            if (__srcX >= __width) {
#ifdef EARLY_OUT
                                if (didFetchInRow) {
                                    break;
                                }
#endif
                                __pix = blackPixel;
                            } else {
                                double __nY;

                                /* rotate Y */
                                __nY = (__sin * __pX) + __cosPY;
                                /* translate Y in source (origin to 0/0) */
                                __nY = __nY + __halfH + 0.5;

                                /* inside ? */
                                if (__nY < 0) {
#ifdef EARLY_OUT
                                    if (didFetchInRow) {
                                        break;
                                    }
#endif
                                    __pix = blackPixel;
                                } else {
                                    int __srcY;

                                    __srcY = (int)__nY;
                                    /* inside ? */
                                    if (__srcY >= __height) {
#ifdef EARLY_OUT
                                        if (didFetchInRow) {
                                            break;
                                        }
#endif
                                        __pix = blackPixel;
                                    } else {
                                        /* fetch source pixel */

                                        static struct inlineCache valAt = _ILC2;
#ifdef EARLY_OUT
                                        didFetchInRow = 1;
#endif
                                        __pix = (*valAt.ilc_func)(self,
                                                              @symbol(pixelAtX:y:),
                                                              nil, &valAt,
                                                              __MKSMALLINT(__srcX),
                                                              __MKSMALLINT(__srcY));
                                    }
                                }
                            }
                        }

                        /* store pixel */
                        {
                            static struct inlineCache atPutVal = _ILC3;

                            if (__pix != __MKSMALLINT(0)) {
                                (*atPutVal.ilc_func)(newImage,
                                                      @symbol(pixelAtX:y:put:),
                                                      nil, &atPutVal,
                                                      __MKSMALLINT(__dstX),
                                                      __MKSMALLINT(__dstY),
                                                      __pix
                                                     );
                            }
                        }
                    }
                }
                break;
        }

        bad = false;
    }
%}.

    bad ifTrue:[
        "/ should not happen
        self primitiveFailed

"/        sinRot := radians negated sin.
"/        cosRot := radians negated cos.
"/
"/        0 to:newHeight-1 do:[:dstY |
"/            pY := (dstY + minY).
"/            sinPY := (sinRot * pY).
"/            cosPY := (cosRot * pY).
"/
"/            0 to:newWidth-1 do:[:dstX |
"/
"/                "/ translate center to origin
"/                pX := (dstX + minX).
"/
"/                nX := (cosRot * pX) - sinPY.
"/
"/                "/ translate in source
"/                srcX := nX + halfW.
"/                srcX := srcX rounded.
"/                "/ inside ?
"/
"/                (srcX >= 0 and:[srcX < width]) ifTrue:[
"/                    nY := (sinRot * pX) + cosPY.
"/                    srcY := nY + halfH.
"/                    srcY := srcY rounded.
"/
"/                    "/ inside ?
"/                    (srcY >= 0 and:[srcY < height]) ifTrue:[
"/                        pix := self pixelAtX:srcX y:srcY
"/                    ] ifFalse:[
"/                        pix := blackPixel.        
"/                    ].
"/                ] ifFalse:[
"/                    pix := blackPixel.        
"/                ].
"/                pix ~~ blackPixel ifTrue:[
"/                    newImage pixelAtX:dstX y:dstY put:pix.        
"/                ]
"/            ].
"/        ].
    ].

    ^ newImage

    "
     |i|

     i := Image fromFile:'bitmaps/xpmBitmaps/BOOK.xpm'.
     i inspect.
     (i rotated:45) inspect.
     (i rotated:90) inspect.
     (i rotated:91) inspect.
     (i rotated:95) inspect.
    "
    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     i := Depth24Image fromImage:i.
     (i rotated:200) inspect
    "
    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     Transcript showCR:(
        Time millisecondsToRun:[
           i rotated:45.
        ]
     ).
    "
    "
     |v i rI rot|

     i := Image fromFile:'bitmaps/xpmBitmaps/BOOK.xpm'.
     v := View new extent:(i width max:100)@(i height max:100).
     v openAndWait.
     rot := 0.
     [true] whileTrue:[
        rI := i rotated:rot.
        rI := rI on:v device.
        v clear.
        v displayForm:rI x:v width//2-(rI width//2) y:v height//2-(rI height // 2).
        rot := rot + 5.
        rI close.
     ]
    "
    "
     |v i rI rot|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     v := View new extent:400@400.
     v openAndWait.
     rot := 0.
     [true] whileTrue:[
        rI := i rotated:rot.
        rI := rI on:v device.
        v clear.
        v displayForm:rI x:v width//2-(rI width//2) y:v height//2-(rI height // 2).
        rot := rot + 5.
        rI close.
     ]
    "

    "Modified: 22.4.1997 / 17:36:37 / cg"
!

lightened
    "return a new image which is slightly brighter than the receiver.
     The receiver must be a palette image (currently).
     CAVEAT: this only works with palette images (i.e. not for rgb or greyScale).
     CAVEAT: Need an argument, which specifies by how much it should be lighter."

     ^ self 
        copyWithColorMapProcessing:[:clr | clr lightened]

    "
     (Image fromFile:'bitmaps/gifImages/claus.gif') inspect
     (Image fromFile:'bitmaps/gifImages/claus.gif') lightened inspect
     (Image fromFile:'bitmaps/gifImages/claus.gif') darkened inspect
     (Image fromFile:'bitmaps/gifImages/claus.gif') darkened darkened inspect
    "

    "Modified: 24.4.1997 / 18:31:23 / cg"
!

magnifiedBy:scale
    "return a new image magnified by scalePoint, aPoint.
     If non-integral magnify is asked for, pass the work on to 'hardMagnifyBy:'
     while simple (integral) magnifications are handled here."

    |scalePoint mX mY
     magX      "{ Class: SmallInteger }"   "new version of stc can find this out itself..."
     magY      "{ Class: SmallInteger }"
     srcOffset "{ Class: SmallInteger }"
     dstOffset "{ Class: SmallInteger }"
     w         "{ Class: SmallInteger }"
     h         "{ Class: SmallInteger }"
     first
     newWidth newHeight newImage newBits newMask
     bitsPerPixel newBytesPerRow oldBytesPerRow|

    scalePoint := scale asPoint.
    mX := scalePoint asPoint x.
    mY := scalePoint asPoint y.
    ((mX <= 0) or:[mY <= 0]) ifTrue:[^ nil].
    ((mX = 1) and:[mY = 1]) ifTrue:[^ self].

    ((mX isMemberOf:SmallInteger) and:[mY isMemberOf:SmallInteger]) ifFalse:[
        ^ self hardMagnifiedBy:scalePoint
    ].

    bitsPerPixel := self depth.
    oldBytesPerRow := ((width * bitsPerPixel) + 7) // 8.

    w := width.
    h := height.
    magX := mX.
    magY := mY.

    newWidth := w * mX.
    newHeight := h * mY.
    newBytesPerRow := ((newWidth * bitsPerPixel) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * newHeight).

    mask notNil ifTrue:[
        newMask := (mask magnifiedBy:scalePoint)
    ].

    newImage := self species new.
    newImage 
        width:newWidth 
        height:newHeight 
        photometric:photometric 
        samplesPerPixel:samplesPerPixel 
        bitsPerSample:bitsPerSample 
        colorMap:colorMap copy
        bits:newBits
        mask:newMask.

    mX = 1 ifTrue:[
        "expand rows only"
        srcOffset := 1.
        dstOffset := 1.

        1 to:h do:[:row |
            1 to:mY do:[:i |
                newBits replaceFrom:dstOffset 
                        to:(dstOffset + oldBytesPerRow - 1)
                        with:bytes 
                        startingAt:srcOffset.
                dstOffset := dstOffset + newBytesPerRow
            ].
            srcOffset := srcOffset + oldBytesPerRow.
        ].
    ] ifFalse:[
        "expand cols"
        (mX > 1) ifTrue:[
            dstOffset := 1.
            srcOffset := 1.
            1 to:h do:[:row |
                self magnifyRowFrom:bytes 
                     offset:srcOffset  
                     into:newBits 
                     offset:dstOffset 
                     factor:mX.

                first := dstOffset.
                dstOffset := dstOffset + newBytesPerRow.
                " and copy for row expansion "
                2 to:mY do:[:i |
                    newBits replaceFrom:dstOffset 
                            to:(dstOffset + newBytesPerRow - 1)
                            with:newBits 
                            startingAt:first.
                    dstOffset := dstOffset + newBytesPerRow
                ].
                srcOffset := srcOffset + oldBytesPerRow.
            ].
        ]
    ].
    ^ newImage

    "
     ((Image fromFile:'bitmaps/gifImages/claus.gif') magnifiedBy:1@2)
    "

    "Modified: 24.4.1997 / 18:32:04 / cg"
!

magnifiedPreservingRatioTo:anExtent 
    "return a new image magnified to fit into anExtent,
     preserving the receivers width/height ratio.
     (i.e. not distorting the image).
     See also #magnifiedTo: and #magnifiedBy:"

    |rX rY|

    rX := anExtent x / self width.
    rY := anExtent y / self height.
    ^ self magnifiedBy:(rX min:rY)

    "
     ((Image fromFile:'bitmaps/gifImages/garfield.gif') magnifiedPreservingRatioTo:100@100)

    in contrast to:

     ((Image fromFile:'bitmaps/gifImages/garfield.gif') magnifiedTo:100@100)
    "

    "Modified: 22.4.1997 / 12:33:46 / cg"
!

magnifiedTo:anExtent 
    "return a new image magnified to have the size specified by extent.
     This may distort the image if the arguments ratio is not the images ratio.
     See also #magnifiedPreservingRatioTo: and #magnifiedBy:"

    ^ self magnifiedBy:(anExtent / self extent)

    "
     ((Image fromFile:'bitmaps/garfield.gif') magnifiedTo:100@100)

    in contrast to:

     ((Image fromFile:'bitmaps/garfield.gif') magnifiedPreservingRatioTo:100@100)
    "
!

negative
    "return a new image which is a negative of the receiver.
     The receiver must be a palette image (currently)."

    |newImage|

    colorMap isNil ifTrue:[
        photometric == #blackIs0 ifTrue:[
            newImage := self copy.
            newImage photometric:#whiteIs0.
            ^ newImage
        ].
        photometric == #whiteIs0 ifTrue:[
            newImage := self copy.
            newImage photometric:#blackIs0.
            ^ newImage
        ].
        ^ nil
    ].

     ^ self 
        copyWithColorMapProcessing:[:clr | 
                |newColor|

                newColor := Color
                                redByte:(255 - clr redByte)
                                greenByte:(255 - clr greenByte)
                                blueByte:(255 - clr blueByte)
        ]

    "
     (Image fromFile:'bitmaps/gifImages/claus.gif') inspect
     (Image fromFile:'bitmaps/gifImages/claus.gif') negative inspect
    "

    "Created: 20.6.1997 / 13:13:41 / cg"
    "Modified: 20.6.1997 / 13:17:50 / cg"
!

rotated:degrees
    "return a new image from the old one, by rotating the image
     degrees clockwise.
     Notice, that the resulting image has a different extent.
     If rotation is heavily used, the workHorse methods
     (#easyRotateBitsInto:angle: and #hardRotateBitsInto:angle) may
     be redefined in concrete image subclasses."

    |nW nH newImage newBits newBytesPerRow d|

    d := degrees.
    [d < 0] whileTrue:[d := d + 360].
    d >= 360 ifTrue:[d := d \\ 360].
    d := d truncated.
    d = 0 ifTrue:[^ self].
    ((d ~= 90) and:[(d ~= 270) and:[d ~= 180]]) ifTrue:[
        ^ self hardRotated:d
    ].

    d = 180 ifTrue:[
        nW := width.
        nH := height.
    ] ifFalse:[
        nW := height.
        nH := width.
    ].

    newImage := self species new.
    newImage width:nW.
    newImage height:nH.
    newBytesPerRow := newImage bytesPerRow.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * nH).
    newImage bits:newBits.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.
    mask notNil ifTrue:[
        newImage mask:(mask rotated:degrees)
    ].

    self easyRotateBitsInto:newImage angle:d.

    ^ newImage

    "
     |i|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     (i rotated:45) inspect.
     (i rotated:90) inspect.
     (i rotated:180) inspect.
     (i rotated:270) inspect.
    "
    "
     |i|

     i := Image fromFile:'bitmaps/xpmBitmaps/BOOK.xpm'.
     i inspect.
     (i rotated:90) inspect.
     (i rotated:180) inspect.
     (i rotated:270) inspect.
    "

    "Modified: 24.4.1997 / 18:33:42 / cg"
!

withPixelFunctionApplied:pixelFunctionBlock
    "return a new image from the old one, by applying a pixel processor
     on the pixel colors.
     Notice: this method is very slow - either apply pixel values
     (#withPixelFunctionAppliedToPixels:) or redefine this method in
     a concrete subclass.
     (read `Behond photography, by Gerard J. Holzmann;
           ISBM 0-13-074410-7)
     See blurred / oilPointed as examples ...)"

    |w  "{Class: SmallInteger }"
     h  "{Class: SmallInteger }"
     newImage newBits newBytesPerRow|

    newBytesPerRow := ((width * self depth) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * height).

    newImage := self species new.
    newImage bits:newBits.
    newImage width:width.
    newImage height:height.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.

    w := width - 1.
    h := height - 1.

    0 to:h do:[:y |
        0 to:w do:[:x |
            newImage colorAtX:x y:y put:(pixelFunctionBlock
                                                value:self
                                                value:(self colorAtX:x y:y)
                                                value:x
                                                value:y)
        ]
    ].
    ^ newImage

    "
     |i black|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.

     black := Color black.
     (i withPixelFunctionApplied:[:oldImage :oldColor :x :y |
                        ((x between:100 and:200) 
                        and:[y between:100 and:200]) ifTrue:[
                            oldColor
                        ] ifFalse:[
                            black.
                        ]
                     ]) inspect.
    "

    "Modified: 24.4.1997 / 18:36:59 / cg"
!

withPixelFunctionAppliedToPixels:pixelFunctionBlock
    "return a new image from the old one, by applying a pixel processor
     on the pixel values.
     (read `Behond photography, by Gerard J. Holzmann;
           ISBM 0-13-074410-7)
     See blurred / oilPointed as examples ...)"

    |newImage newBits newBytesPerRow|

    newBytesPerRow := ((width * self depth) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * height).

    newImage := self species new.
    newImage bits:newBits.
    newImage width:width.
    newImage height:height.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.

    self applyPixelValuesTo:pixelFunctionBlock into:newImage.
    ^ newImage

    "oil painting effect:

     |i w h|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     w := i width - 1.
     h := i height - 1.
     (i withPixelFunctionAppliedToPixels:[:oldImage :oldPixel :x :y |
                        |b p max xMin xMax yMin yMax|

                        b := Bag identityNew:10.
                        xMin := x-3 max:0.
                        xMax := x+3 min:w.
                        yMin := y-3 max:0.
                        yMax := y+3 min:h.
                        xMin to:xMax do:[:tx|
                          yMin to:yMax do:[:ty|
                            b add:(oldImage pixelAtX:tx y:ty)
                          ]
                        ].
                        max := 0.
                        b contents keysAndValuesDo:[:pixel :n |
                            n > max ifTrue:[
                                p := pixel.
                                max := n
                            ]
                        ].
                        p
                     ]) inspect.
    "

    "fisheye effect:

     |i w h w2 h2 R white|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     w := i width - 1.
     h := i height - 1.
     w2 := w // 2.
     h2 := h // 2.
     R := w2.
     white := i valueFromColor:Color white.
     (i withPixelFunctionAppliedToPixels:[:oldImage :oldPixel :x :y |
                        |p r a nR nP nX nY|

                        p := (x-w2)@(y-h2).
                        r := p r.
                        a := p theta.
                        nR := r * r / R.
                        nP := Point r:nR theta:a.
                        nX := ((nP x+w2) rounded max:0) min:w.
                        nY := ((nP y+h2) rounded max:0) min:h.
                        (nX > w or:[nX < 0]) ifTrue:[
                            white
                        ] ifFalse:[
                            (nY > h or:[nY < 0]) ifTrue:[
                                white
                            ] ifFalse:[
                                oldImage pixelAtX:nX y:nY
                            ]
                        ]
                     ]) inspect.
    "
    "fisheye effect:

     |i w h w2 h2 R white|

     i := Image fromFile:'bitmaps/gifImages/claus.gif'.
     i inspect.
     w := i width - 1.
     h := i height - 1.
     w2 := w // 2.
     h2 := h // 2.
     R := w2.
     white := i valueFromColor:Color white.
     (i withPixelFunctionAppliedToPixels:[:oldImage :oldPixel :x :y |
                        |p r a nR nP nX nY|

                        p := (x-w2)@(y-h2).
                        r := p r.
                        a := p theta.
                        nR := r * r / R.
                        nP := Point r:nR theta:a.
                        nX := (nP x+w2) rounded.
                        nY := (nP y+h2) rounded.
                        (nX > w or:[nX < 0]) ifTrue:[
                            white
                        ] ifFalse:[
                            (nY > h or:[nY < 0]) ifTrue:[
                                white
                            ] ifFalse:[
                                oldImage pixelAtX:nX y:nY
                            ]
                        ]
                     ]) inspect.
    "

    "Created: 24.4.1997 / 18:37:17 / cg"
    "Modified: 24.4.1997 / 18:40:02 / cg"
! !

!Image methodsFor:'inspecting'!

inspectorClass
    "redefined to launch an ImageInspector
     (instead of the default InspectorView)."

    (width notNil and:[height notNil]) ifTrue:[
        ^ ImageInspectorView
    ].
    ^ super inspectorClass

    "Modified: 10.6.1996 / 18:23:55 / cg"
! !

!Image methodsFor:'instance release'!

close
    "release device resources; destroy any device-resources"

    deviceForm notNil ifTrue:[
        deviceForm destroy.
        deviceForm := nil.
    ].
    monoDeviceForm notNil ifTrue:[
        monoDeviceForm destroy.
        monoDeviceForm := nil.
    ].
    fullColorDeviceForm notNil ifTrue:[
        fullColorDeviceForm destroy.
        fullColorDeviceForm := nil.
    ].

    device := nil.
    mask notNil ifTrue:[
        mask close.
    ].
    Lobby unregister:self.

    "Modified: 21.6.1996 / 19:08:19 / cg"
    "Created: 24.4.1997 / 11:55:20 / cg"
!

disposed
    "some Image has been collected - nothing to do"
!

quickRelease
    "release device resources"

    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.

    "Modified: 16.1.1997 / 19:33:01 / cg"
!

release
    "release device resources"

    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.
    mask notNil ifTrue:[
        mask release.
    ].
    Lobby unregister:self.
    super release.

    "Modified: 11.6.1997 / 13:20:04 / cg"
!

restored
    "flush device specifics after a snapin or binary restore"

    self release
! !

!Image methodsFor:'obsolete'!

magnifyBy:scale
    "obsolete: has been renamed to magnifiedBy: for ST-80 compatibility
     and name consistency ..."

    self obsoleteMethodWarning.
    ^ self magnifiedBy:scale
! !

!Image methodsFor:'pixel copying'!

copyFrom:anImage x:srcX y:srcY toX:dstX y:dstY width:w height:h
    "replace a rectangular area by pixels from another image.
     The sources colors must be present in the destinations
     colorMap - otherwise, an error will be reported.

     WARNING:
       This implementation is a very slow fallback general algorithm
       (the loop over the source pixels is very slow).
       If this method is used heavily, you may want to redefine it in 
       concrete subclasses for the common case of of copying from an Image 
       with the same depth & palette.
       If you do heavy image processing, specialized versions are even req'd 
       for other cases, rewriting the inner loops as inline C-code."

    |dX dY|

    dX := srcX-dstX.
    dY := srcY-dstY.

    ((photometric == anImage photometric)
     and:[self bitsPerPixel == anImage bitsPerPixel
     and:[colorMap = anImage colorMap]]) ifTrue:[
        "/ can loop over values
        anImage valuesFromX:srcX  y:srcY 
                        toX:srcX+w-1 y:srcY+h-1  
                         do:[:x :y :pixelValue |
            self pixelAtX:x-dX y:y-dY put:pixelValue.
        ]
    ] ifFalse:[
        "/ must loop over colors - horribly slow
        anImage colorsFromX:srcX  y:srcY 
                        toX:srcX+w-1 y:srcY+h-1  
                         do:[:x :y :clr |
            self colorAtX:x-dX y:y-dY put:clr.
        ]
    ].
    mask notNil & anImage mask notNil
    ifTrue:
    [
        mask copyFrom:anImage mask x:srcX y:srcY toX:dstX y:dstY width:w height:h
    ]
    "
     |i1 i8 i4|

     i8 := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     i8 inspect.
     i1 := Image fromFile:'bitmaps/SBrowser.xbm'.
     i1 inspect.

     i4 := Depth4Image fromImage:i8.
     i4 copyFrom:i1 x:0 y:0 toX:20 y:20 width:20 height:20.
     i4 inspect.
    "

    "Created: 20.9.1995 / 10:14:01 / claus"
    "Modified: 20.9.1995 / 10:25:31 / claus"
    "Modified: 21.6.1997 / 13:15:31 / cg"
!

subImageIn:aRectangle
    "create and return a new image consisting of a subArea of myself"

    ^ self class fromSubImage:self in:aRectangle

    "
     |i|

     i := Image fromFile:'bitmaps/garfield.gif'.
     i inspect.
     (i subImageIn:(300@160 corner:340@200)) inspect
    "

    "Created: 20.9.1995 / 01:24:20 / claus"
! !

!Image methodsFor:'printing & storing'!

storeOn:aStream
    "append a printed representation of the receiver to aStream,
     from which a copy of it can be reconstructed."

    aStream nextPutAll:'(' , self class name , ' new)'.
    aStream nextPutAll:' width: '. width storeOn:aStream.
    aStream nextPutAll:'; height: '. height storeOn:aStream.
    aStream nextPutAll:'; photometric:('. photometric storeOn:aStream.
    aStream nextPutAll:'); bitsPerSample:('. bitsPerSample storeOn:aStream.
    aStream nextPutAll:'); samplesPerPixel:('. samplesPerPixel storeOn:aStream. 
    aStream nextPutAll:'); bits:(ByteArray fromPackedString:'. bytes asPackedString storeOn:aStream. aStream nextPutAll:') '.

    colorMap notNil 
    ifTrue:
    [
        |colorMapArray|
        colorMapArray := OrderedCollection new.
        colorMap do: [:clr| colorMapArray add: clr redByte; add: clr greenByte; add: clr blueByte].
        aStream nextPutAll:'; colorMapFromArray:'.
        colorMapArray asByteArray storeOn:aStream.
    ].  
    mask notNil
    ifTrue:
    [
        aStream nextPutAll:'; mask:('.
        mask storeOn:aStream.
        aStream nextPutAll:')'.
    ].
    aStream nextPutAll:'; yourself'
! !

!Image methodsFor:'private'!

bestSupportedImageFormatFor:aDevice
    "scan through the image formats as supported by aDevice,
     and return the best to use when the receiver is to be represented
     on it. The best format is the one with the same number or more bits per
     pixel. Here, the smallest format found which can represent enough pixels is taken."

    |bestDeviceDepth bestDeviceBitsPerPixel bestInfo maxInfo myDepth "maxDepth" maxBitsPerPixel |

    myDepth := self bitsPerPixel.
    maxBitsPerPixel := 0.

    aDevice supportedImageFormats do:[:entry |
        |deviceImageDepth deviceImageBitsPerPixel|

        deviceImageDepth := entry at:#depth.
        deviceImageBitsPerPixel := entry at:#bitsPerPixel.

        "/ for now, ignore all depth's which are neither 1 nor the
        "/ devices depth.
        "/ (actually, many devices can handle other pixMap formats,
        "/ but I dont know (yet) how to pass the correct color info
        
        ((deviceImageDepth == 1) or:[deviceImageDepth == aDevice depth]) ifTrue:[

            deviceImageBitsPerPixel > maxBitsPerPixel ifTrue:[
                maxInfo := entry.
    "/            maxBitsPerPixel := deviceImageBitsPerPixel.
    "/            maxDepth := deviceImageDepth.
            ].

            deviceImageDepth >= myDepth ifTrue:[
                deviceImageDepth == myDepth ifTrue:[
                    "/ take the better one ...
                    (bestDeviceDepth isNil
                     or:[(bestDeviceBitsPerPixel ~~ bestDeviceDepth)
                        and:[deviceImageDepth == deviceImageBitsPerPixel]]) ifTrue:[
                        bestInfo := entry.
                        bestDeviceDepth := deviceImageDepth.
                        bestDeviceBitsPerPixel := deviceImageBitsPerPixel.
                    ]
                ] ifFalse:[
                    "/ take the next-larger depth
                    (bestDeviceDepth isNil
                     or:[deviceImageBitsPerPixel < bestDeviceBitsPerPixel]) ifTrue:[
                        bestInfo := entry.
                        bestDeviceDepth := deviceImageDepth.
                        bestDeviceBitsPerPixel := deviceImageBitsPerPixel.
                    ]
                ]    
            ].
        ].
    ].

    bestDeviceDepth isNil ifTrue:[
        maxBitsPerPixel == 0 ifTrue:[
            "/
            "/ oops - nothing appropriate
            "/
            maxInfo notNil ifTrue:[
                ^ maxInfo
            ].
            bestDeviceDepth := bestDeviceBitsPerPixel := aDevice depth.
            bestInfo := IdentityDictionary new.
            bestInfo at:#depth put:bestDeviceDepth.
            bestInfo at:#bitsPerPixel put:bestDeviceBitsPerPixel.
            bestInfo at:#padding put:32.
        ] ifFalse:[
            bestInfo := maxInfo.
        ]
    ].
    ^ bestInfo

    "Modified: / 7.2.1998 / 11:23:24 / cg"
!

colormapFromImage:anImage
    "setup the receivers colormap from another image.
     Color precision may be lost, if conversion is from a higher depth
     image. This does not convert any pixel values; it is  non-public helper
     for fromImage:/fromSubImake:"

    samplesPerPixel == 3 ifTrue:[
        photometric := #rgb
    ] ifFalse:[
        photometric := anImage photometric.
        photometric == #palette ifTrue:[
            colorMap := anImage colorMap copy.
            "
             must compress the colormap, if source image has higher depth
             than myself. 
            "
            anImage bitsPerPixel > self bitsPerPixel ifTrue:[
                "
                 get used colors are extracted into our colorMap
                 (the at-put below will set the pixelValue according the
                 new colorIndex
                "
                colorMap := anImage usedColors asArray.
                colorMap size > (1 bitShift:self bitsPerPixel) ifTrue:[
                    'Image [warning]: possibly too many colors in image' errorPrintCR
                ]
            ]
        ]
    ].

    "Created: 20.9.1995 / 00:58:42 / claus"
    "Modified: 10.1.1997 / 17:52:21 / cg"
!

greyByteMapForRange:range
    "return a collection to map from pixelValues to greyLevels
     in the range 0..range. 
     Range must be < 256 (for elements to fit into a ByteArray).
     The values are rounded towards the nearest pixel."

    |d clr val greyMap r
     n "{ Class: SmallInteger }"|

    r := range.
    r == 256 ifTrue:[
        r := 255
    ].

    photometric == #palette ifTrue:[
        n := colorMap size.
        greyMap := ByteArray new:n.

        1 to:n do:[:i |
            (clr := colorMap at:i) isNil ifTrue:[
                "/ an unused color
                val := 0.
            ] ifFalse:[
                val := (r * clr brightness) rounded
            ].
            greyMap at:i put:val
        ].
    ] ifFalse:[
        d := self bitsPerPixel.
        n := 1 bitShift:d.
        n >= 4096 ifTrue:[
            self error:'size not supported - too large'.
            ^ nil
        ].
        greyMap := ByteArray new:n.


        photometric == #rgb ifTrue:[
            1 to:n do:[:i |
                greyMap at:i put:(r * (self colorFromValue:i-1) brightness) rounded
            ].
        ] ifFalse:[
            1 to:n do:[:i |
                greyMap at:i put:(r / (n-1) * (i-1)) rounded
            ].
            photometric == #blackIs0 ifTrue:[
                "/ we are done
            ] ifFalse:[
                photometric == #whiteIs0 ifTrue:[
                    greyMap reverse
                ] ifFalse:[
                    self error:'invalid format'.
                    ^ nil
                ]
            ]
        ].
    ].

    ^ greyMap

    "
     Depth8Image new greyByteMapForRange:256

     Depth8Image new greyByteMapForRange:64
     Depth4Image new greyByteMapForRange:64

     Depth4Image new greyByteMapForRange:1
    "

    "Created: 8.6.1996 / 08:34:14 / cg"
    "Modified: 1.3.1997 / 17:24:45 / cg"
!

greyMapForRange:range
    "return a collection to map from pixelValues to greyLevels
     in the range 0..range. The brightness values are not rounded."

    |d r clr val
     n "{Class: SmallInteger }"
     n2 "{Class: SmallInteger }"
     greyArray|

    d := self bitsPerPixel.
    n := 1 bitShift:d.
    n >= 4096 ifTrue:[
        self error:'size not supported - too large'.
        ^ nil
    ].

    greyArray := Array new:n.

    photometric == #palette ifTrue:[
        n2 := colorMap size.
        1 to:n2 do:[:i |
            (clr := colorMap at:i) isNil ifTrue:[
                "/ an unused color
                val := 0.
            ] ifFalse:[
                val := range * clr brightness
            ].
            greyArray at:i put:val
        ].
        n2 < n ifTrue:[
            greyArray from:n2+1 to:n put:0
        ]
    ] ifFalse:[
        photometric == #rgb ifTrue:[
            1 to:n do:[:i |
                greyArray at:i put:(range * (self colorFromValue:(i-1)) brightness)
            ]
        ] ifFalse:[
            r := range asFloat.
            1 to:n do:[:i |
                greyArray at:i put:(r / (n-1) * (i-1))
            ].
            photometric == #blackIs0 ifTrue:[
                "/ we are done
            ] ifFalse:[
                photometric == #whiteIs0 ifTrue:[
                    greyArray reverse
                ] ifFalse:[
                    self error:'invalid format'.
                    ^ nil
                ]
            ]
        ]
    ].
    ^ greyArray

    "
     Depth8Image new greyMapForRange:64
     Depth4Image new greyMapForRange:64

     Depth16Image new greyMapForRange:1
     Depth4Image new greyMapForRange:1
     Depth2Image new greyMapForRange:1
    "

    "Modified: 1.3.1997 / 15:48:49 / cg"
!

magnifyRowFrom:srcBytes offset:srcStart pixels:oldPixels 
          into:dstBytes offset:dstStart factor:mX

    "magnify a single pixel row - can only magnify by integer factors,
     can only magnify 1,2,4,8 and 24 bit-per-pixel images. But this is done fast."

    ^ self subclassResponsibility
!

makeDeviceGrayPixmapOn:aDevice depth:depth fromArray:bits
    "given the bits of a grey/color bitmap, 8-bit padded and
     pixels interpreted as greyValues, 0 is black,  
     create a device form for it"

    |f|

    f := Form width:width height:height depth:depth on:aDevice.
    f isNil ifTrue:[^ nil].

    f bits:bits.
    f initGC.

    (aDevice blackpixel ~~ 0) ifTrue:[
        "/ have to invert bits
        f function:#copyInverted
    ].
    aDevice 
	drawBits:bits 
	depth:depth
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) 
        x:0 y:0 
        width:width height:height 
        with:(f gcId).
    ^ f

    "Created: 10.6.1996 / 20:10:31 / cg"
!

makeDeviceMonochromeBitmapOn:aDevice fromArray:monoBits
    "given the bits of a monochrome bitmap, 8-bit padded and
     0-bits as black, create a device form for it"

"/    monoBits invert.
    ^ (Form width:width height:height fromArray:monoBits on:aDevice)
        colorMap:(Array with:Color black with:Color white).

    "Created: 10.6.1996 / 20:18:09 / cg"
    "Modified: 17.4.1997 / 01:07:38 / cg"
!

makeDevicePixmapOn:aDevice depth:depth fromArray:bits
    "given the bits of a grey/color bitmap, 8-bit padded and
     pixels interpreted as in the devices colormap, 
     create a device form for it"

    |f|

    f := Form width:width height:height depth:depth on:aDevice.
    f isNil ifTrue:[^ nil].

    f bits:bits.
    f initGC.

    aDevice 
	drawBits:bits 
	depth:depth
	padding:8
        width:width height:height
        x:0 y:0
        into:(f id) 
        x:0 y:0 
        width:width height:height 
        with:(f gcId).
    ^ f

    "Created: 10.6.1996 / 17:56:08 / cg"
    "Modified: 10.6.1996 / 20:11:27 / cg"
! !

!Image methodsFor:'queries'!

averageColor
    "return the average color of the image.
     This usually only makes sense for textures and patterns
     (i.e. to compute shadow & light colors for viewBackgrounds).
     Notice, that for the above purpose, it is usually ok to process 
     a subImage - i.e. use Image>>averageColorIn: on a smaller rectangle"

    ^ self averageColorIn:(0@0 corner:(width-1)@(height-1))
!

averageColorIn:aRectangle
    "return the images average color in an area.
     The implementation below is slow - so you may want to
     create tuned versions for DepthXImage if you plan to do
     heavy image processing ... 
     (also, creating tuned versions of the enumeration messages helps a lot)"

    |x0 "{ Class:SmallInteger }"
     y0 "{ Class:SmallInteger }"
     x1 "{ Class:SmallInteger }"
     y1 "{ Class:SmallInteger }"
     sumRed sumGreen sumBlue n|

    sumRed := sumGreen := sumBlue := 0.    
    y0 := aRectangle top.
    y1 := aRectangle bottom.
    x0 := aRectangle left.
    x1 := aRectangle right.

    self colorsFromX:x0 y:y0 toX:x1 y:y1 do:[:x :y :colorAtXY |
       sumRed := sumRed + colorAtXY red.
       sumGreen := sumGreen + colorAtXY green.
       sumBlue := sumBlue + colorAtXY blue.
    ].
    n := (x1 - x0 + 1) * (y1 - y0 + 1).    
    ^ Color red:(sumRed / n) green:(sumGreen / n) blue:(sumBlue / n)
!

bitsPerPixel
    "return the number of bits per pixel"

    ^ (bitsPerSample inject:0 into:[:sum :i | sum + i])
!

bitsPerRow
    "return the number of bits in one scanline of the image"

    ^  width * (self bitsPerPixel).
!

blueBitsOf:pixel
    "if the receiver is an rgb-image:
     return the blue bits of a pixelValue.
     This has to be redefined by subclasses."

    |blueBits|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        blueBits := bitsPerSample at:3.
        blueBits == 0 ifTrue:[^ 0].

        ^ pixel bitAnd:(1 bitShift:blueBits)-1
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 09:44:21 / cg"
    "Modified: 10.6.1996 / 14:59:44 / cg"
!

blueComponentOf:pixel
    "if the receiver is an rgb-image:
     return the blue component (0 .. 100) of a pixelValue.
     This has to be redefined by subclasses."

    |blueBits  "{ Class: SmallInteger }"
     s         "{ Class: SmallInteger }"|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        blueBits := bitsPerSample at:3.
        blueBits == 0 ifTrue:[^ 0].

        s := (1 bitShift:blueBits) - 1.

        ^ 100.0 / s * (pixel bitAnd:(1 bitShift:blueBits)-1)
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 08:45:22 / cg"
    "Modified: 10.6.1996 / 14:55:39 / cg"
!

bounds
    "return my bounds (added to make images usable as VisualComponents)"

    ^ Rectangle left:0 top:0 width:width height:height

    "Created: 13.5.1996 / 10:27:06 / cg"
!

brightness
    "return the (average) brightness of the image.
     This usually only makes sense for textures and patterns
     (i.e. to compute shadow & light colors for viewBackgrounds).
     Notice, that for the above purpose, only a subimage is inspected here"

    ^ (self averageColorIn:(0@0 corner:7@7)) brightness

    "Modified: 8.6.1996 / 08:46:46 / cg"
!

bytesPerRow
    "return the number of bytes in one scanline of the image"

    |bitsPerRow bytesPerRow|

    bitsPerRow := width * (self bitsPerPixel).
    bytesPerRow := bitsPerRow // 8.
    ((bitsPerRow \\ 8) ~~ 0) ifTrue:[
        bytesPerRow := bytesPerRow + 1
    ].
    ^ bytesPerRow
!

bytesPerRowPaddedTo:padding
    "return the number of bytes in one scanline of the image,
     if scanlines are to be padded to padding-bits."

    ^ self class
        bytesPerRowForWidth:width depth:(self bitsPerPixel) padding:padding
!

center
    "for compatibility with GC protocol - return the centerPoint"

    ^ (width // 2) @ (height // 2)

    "Created: 22.10.1997 / 23:52:40 / cg"
!

colorFromValue:pixelValue
    "given a pixel value, return the corresponding color.
     Pixel values start with 0."

    |p maxPixel r g b|

    p := photometric.
    p isNil ifTrue:[
        colorMap notNil ifTrue:[
            p := #palette
        ] ifFalse:[
"/            'Image [warning]: no photometric - assume greyscale' infoPrintCR
            p := #blackIs0
        ]
    ].

    p == #blackIs0 ifTrue:[
        maxPixel := (1 bitShift:self bitsPerPixel) - 1.
        ^ Color gray:(pixelValue * (100 / maxPixel)).
    ].

    p == #whiteIs0 ifTrue:[
        maxPixel := (1 bitShift:self bitsPerPixel) - 1.
        ^ Color gray:100 - (pixelValue * (100 / maxPixel)).
    ].

    p == #palette ifTrue:[
        pixelValue >= colorMap size ifTrue:[
            ^ Color black
        ].
        ^ colorMap at:(pixelValue + 1)
    ].

    p == #rgb ifTrue:[
        r := self redComponentOf:pixelValue.
        g := self greenComponentOf:pixelValue.
        b := self blueComponentOf:pixelValue.
        ^ Color red:r green:g blue:b
    ].

    self error:'invalid photometric'

    "Created: 8.6.1996 / 08:46:18 / cg"
    "Modified: 28.1.1997 / 12:05:22 / cg"
!

greenBitsOf:pixel
    "if the receiver is an rgb-image:
     return the green bits of a pixelValue.
     This has to be redefined by subclasses."

    |blueBits greenBits|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        greenBits := bitsPerSample at:2.
        greenBits == 0 ifTrue:[^ 0].
        blueBits := bitsPerSample at:3.

        ^ (pixel bitShift:blueBits negated) bitAnd:(1 bitShift:greenBits)-1
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 09:44:37 / cg"
    "Modified: 10.6.1996 / 14:59:35 / cg"
!

greenComponentOf:pixel
    "if the receiver is an rgb-image:
     return the green component (0..100)  of a pixelValue.
     This has to be redefined by subclasses."

    |greenBits "{ Class: SmallInteger }"
     blueBits  "{ Class: SmallInteger }"
     s         "{ Class: SmallInteger }"|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        greenBits := bitsPerSample at:2.
        greenBits == 0 ifTrue:[^ 0].
        blueBits := bitsPerSample at:3.

        s := (1 bitShift:greenBits) - 1.

        ^ 100.0 / s * ((pixel bitShift:blueBits negated) bitAnd:(1 bitShift:greenBits)-1)
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 08:45:34 / cg"
    "Modified: 10.6.1996 / 14:55:24 / cg"
!

heightOn:aGC
    "return my height, if displayed on aGC;
     since my height is independent of the device (the number of pixels),
     return the pixel-height"

    ^ height

    "Created: 12.5.1996 / 21:35:06 / cg"
    "Modified: 13.5.1996 / 10:26:36 / cg"
!

isImage
    "return true, if the receiver is some kind of image;
     true is returned here - the method is redefined from Object."

    ^ true
!

isImageOrForm
    "return true, if the receiver is some kind of image or form;
     true is returned here - the method is redefined from Object."

    ^ true
!

isMask
    ^ false

    "Created: 22.4.1997 / 14:12:02 / cg"
!

pixelArraySpecies
    "return the kind of pixel-value container in rowAt:/rowAt:put: methods"

    self depth > 8 ifTrue:[
        ^ Array
    ].
    ^ ByteArray

    "Created: 6.3.1997 / 15:24:05 / cg"
    "Modified: 24.4.1997 / 15:34:34 / cg"
!

realColorMap
    "return a collection usable as a real colormap of the image.
     For palette images, this is the internal colormap; 
     for other photometrics (which do not have a real colormap), synthesize one.
     This is different from #colorMap, which returns nil for non palette images."

    |d nEntries "{ Class: SmallInteger }"
     colorArray|

    photometric == #palette ifTrue:[
        ^ colorMap asArray
    ].

    d := self depth.
    d > 12 ifTrue:[
        self error:'deep palette images not supported'.
        ^ nil.
    ].

    nEntries := 1 bitShift:d.

    colorArray := Array new:nEntries.
    1 to:nEntries do:[:idx |
        colorArray at:idx put:(self colorFromValue:(idx-1)).
    ].

    ^ colorArray

    "Created: 11.7.1996 / 20:20:35 / cg"
    "Modified: 11.7.1996 / 20:49:21 / cg"
!

redBitsOf:pixel
    "if the receiver is an rgb-image:
     return the red component of a pixelValue.
     This has to be redefined by subclasses."

    |redBits greenBits blueBits|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        redBits := bitsPerSample at:1.
        redBits == 0 ifTrue:[^ 0].
        greenBits := bitsPerSample at:2.
        blueBits := bitsPerSample at:3.

        ^ (pixel bitShift:(greenBits + blueBits) negated)
           bitAnd:(1 bitShift:redBits)-1
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 09:44:51 / cg"
    "Modified: 10.6.1996 / 14:59:05 / cg"
!

redComponentOf:pixel
    "if the receiver is an rgb-image:
     return the red component (0..100) of a pixelValue.
     This has to be redefined by subclasses."

    |redBits   "{ Class: SmallInteger }"
     greenBits "{ Class: SmallInteger }"
     blueBits  "{ Class: SmallInteger }"
     s         "{ Class: SmallInteger }"|

    samplesPerPixel >= 3 ifTrue:[
        "/ assume that the red bits are the leftMost bits

        redBits := bitsPerSample at:1.
        redBits == 0 ifTrue:[^ 0].

        greenBits := bitsPerSample at:2.
        blueBits := bitsPerSample at:3.

        s := (1 bitShift:redBits) - 1.

        ^ 100.0 / s * 
          ((pixel bitShift:(greenBits + blueBits) negated)
           bitAnd:(1 bitShift:redBits)-1)
    ].

    self subclassResponsibility

    "Created: 8.6.1996 / 08:45:30 / cg"
    "Modified: 14.6.1996 / 17:34:00 / cg"
!

usedColors
    "return a collection of colors used in the receiver."

    |usedValues max|

    photometric == #rgb ifTrue:[
        "/ too many to be returned here (think of the mass of
        "/ data to be returned by a 24bit image ... ;-)

        self error:'query not allowed for this image'.
        ^ nil
    ].

    usedValues := self usedValues asArray.
    photometric == #palette ifTrue:[
        ^ usedValues collect:[:val | (colorMap at:val+1 ifAbsent:[Color black])]
    ].

    "/ (photometric == #blackIs0 or:[photometric == #whiteIs0])

    max := (1 bitShift:self depth) - 1.
    ^ usedValues collect:[:val | (Color gray:(100 * val / max ))]

    "
     (Image fromFile:'bitmaps/garfield.gif') usedColors
     (Image fromFile:'bitmaps/SBrowser.xbm') usedColors
     (Image fromFile:'ttt.tiff') usedColors  
    "

    "Modified: 8.6.1996 / 09:04:56 / cg"
!

usedValues
    "return a collection of color values used in the receiver.
     Notice, that the interpretation of the pixels depends on the photometric
     of the image.
     This is a general and therefore slow implementation; subclasses
     may want to redefine this method for more performance."

    |set|

    set := IdentitySet new.
    self valuesFromX:0 y:0 toX:(self width-1) y:(self height-1) do:[:x :y :pixel |
        set add:pixel 
    ].
    ^ set

    "
     (Image fromFile:'bitmaps/garfield.gif') usedValues
     (Image fromFile:'bitmaps/SBrowser.xbm') usedValues
     (Image fromFile:'ttt.tiff') usedValues  
    "
!

valueFromColor:color
    "given a color, return the corresponding pixel value.
     Non-representable colors return nil."

    |pixel maxPixel redBits greenBits blueBits r g b|

    photometric == #whiteIs0 ifTrue:[
        maxPixel := (1 bitShift:self bitsPerPixel) - 1.
        ^ maxPixel - (color brightness * maxPixel) rounded.
    ].

    photometric == #blackIs0 ifTrue:[
        maxPixel := (1 bitShift:self bitsPerPixel) - 1.
        ^ (color brightness * maxPixel) rounded.
    ].

    photometric == #palette ifTrue:[
        pixel := colorMap indexOf:color.
        pixel == 0 ifTrue:[
            "
             the color to be stored is not in the images colormap
            "
            ^ nil
        ].
        ^ pixel - 1
    ].

    photometric == #rgb ifTrue:[
        samplesPerPixel >= 3 ifTrue:[
            redBits := bitsPerSample at:1.
            greenBits := bitsPerSample at:2.
            blueBits := bitsPerSample at:3.
        
            "/ map r/g/b to 0..255
            r := (color red / 100.0 * ((1 bitShift:redBits)-1)) rounded.
            g := (color green / 100.0 * ((1 bitShift:greenBits)-1)) rounded.
            b := (color blue / 100.0 * ((1 bitShift:blueBits)-1)) rounded.
            pixel := (((r bitShift:greenBits) + g) bitShift:blueBits) + b.
            ^ pixel
        ]
    ].

    self error:'format not supported'.
    ^ nil

    "Modified: 15.6.1996 / 09:38:09 / cg"
!

widthOn:aGC
    "return my width, if displayed on aGC;
     since my width is independent of the device (the number of pixels),
     return the pixel-width"

    ^ width

    "Created: 12.5.1996 / 21:34:59 / cg"
    "Modified: 13.5.1996 / 10:26:42 / cg"
! !

!Image methodsFor:'saving on file'!

saveOn:aFileName
    "save the image in a aFileName. The suffix of the filename
     controls the format. Currently, not all formats may be supported
     (see ImageReader subclasses implementing save:onFile:).
     May raise a signal, if the image cannot be written by the
     reader."

    |suffix readerClass|

    "/
    "/ from the extension, get the imageReader class
    "/ (which should know how to write images as well)
    "/
    suffix := aFileName asFilename suffix.
    readerClass := MIMETypes imageReaderForSuffix:suffix.
    readerClass notNil ifTrue:[
        ^ self saveOn:aFileName using:readerClass
    ].

    "/
    "/ no known extension - could ask user for the format here.
    "/ currently default to tiff format.
    "/
    'Image [warning]: unknown extension - cannot figure out format - using tiff' errorPrintCR.

    ^ self saveOn:aFileName using:TIFFReader

    "
     |image|

     image := Image fromFile:'bitmaps/RCube.tiff'.
     image saveOn:'myImage.tiff'.
     image saveOn:'myImage.xbm'.
     image saveOn:'myImage.xpm'.
     image saveOn:'myImage.xwd'.
    "

    "Modified: 30.6.1997 / 22:06:34 / cg"
!

saveOn:aFileName using:readerClass
    "save the receiver using the representation class.
     May raise a signal, if the image cannot be written by the
     reader."

    ^ readerClass save:self onFile:aFileName

    "
     |anImage|

     anImage := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     anImage saveOn:'myImage.tiff' using:TIFFReader.
    "

    "
     |anImage|

     anImage := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     anImage saveOn:'myImage.xbm' using:XBMReader.
    "

    "
     |anImage|

     anImage := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     Image cannotRepresentImageSignal handle:[:ex |
        self warn:'cannot save the image in this format'
     ] do:[
        anImage saveOn:'myImage.xbm' using:XBMReader.
     ]
    "

    "
     |anImage|

     anImage := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     anImage saveOn:'myImage.xpm' using:XPMReader.
    "

    "
     |anImage|

     anImage := Image fromFile:'bitmaps/gifImages/garfield.gif'.
     anImage saveOn:'myImage.gif' using:GIFReader.
    "

    "Modified: 10.4.1997 / 17:49:26 / cg"
! !

!Image methodsFor:'screen capture'!

from:aDrawable in:aRectangle
    "read an image from aDrawable.
     This may be a device Form, a view or the rootView.
     If its a view or rootView, it must be completely visible (or have
     the backingStore option turned on). Otherwise, only the clipped contents
     is returned. This is a common helper for form-to-image conversion,
     and to read hardcopy images from the screen."

    | visType 
     x        "{ Class: SmallInteger }"
     y        "{ Class: SmallInteger }"
     w        "{ Class: SmallInteger }"
     h        "{ Class: SmallInteger }"
     dstIndex "{ Class: SmallInteger }" 
     srcIndex "{ Class: SmallInteger }" 
     srcRow   "{ Class: SmallInteger }"
     dstRow   "{ Class: SmallInteger }"
     inData tmpData usedPixels mapSize 
     map bytesPerLine 
     byteOrder bitOrder spaceBitsPerPixel
     info bitsPerPixelIn bytesPerLineIn 
     maskR "{ Class: SmallInteger }"
     maskG "{ Class: SmallInteger }"
     maskB "{ Class: SmallInteger }"
     shR "{ Class: SmallInteger }"
     shG "{ Class: SmallInteger }"
     shB "{ Class: SmallInteger }"
     shR2 "{ Class: SmallInteger }"
     shG2 "{ Class: SmallInteger }"
     shB2 "{ Class: SmallInteger }"
     r "{ Class: SmallInteger }"
     g "{ Class: SmallInteger }"
     b "{ Class: SmallInteger }"
     word "{ Class: SmallInteger }"
     lword 
     device ddepth isMSB|

    depth := self depth.
    bitsPerPixel := self bitsPerPixel.

    device := aDrawable graphicsDevice.

    (aDrawable isForm and:[aDrawable depth == 1]) ifTrue:[
        "/ a monochrome bitmap ?
        visType := #StaticGray.
        ddepth := 1.
    ] ifFalse:[
        "
         get some attributes of the display device
        "
        visType := device visualType.
        ddepth := device depth.
    ].

    "/ kludge for 15bit XFree server
    ddepth == 15 ifTrue:[
        ddepth := 16
    ].

    (visType == #StaticGray) ifTrue:[
        (device blackpixel == 0) ifTrue:[
            photometric := #blackIs0
        ] ifFalse:[
            photometric := #whiteIs0
        ].
        samplesPerPixel := 1.
        bitsPerPixel := ddepth.
        bitsPerSample := Array with:bitsPerPixel.
    ] ifFalse:[
        ((visType == #PseudoColor) or:[(visType == #StaticColor) or:[visType == #GrayScale]]) ifTrue:[
            photometric := #palette.
            samplesPerPixel := 1.
            bitsPerPixel := ddepth.
            bitsPerSample := Array with:bitsPerPixel.
        ] ifFalse:[
            ((visType == #TrueColor) or:[visType == #DirectColor]) ifTrue:[
                photometric := #rgb.
                samplesPerPixel := 3.
"/                bitsPerPixel := depth.
"/                bitsPerSample := Array with:device bitsRed
"/                                       with:device bitsGreen
"/                                       with:device bitsBlue
                bitsPerPixel := 24.
                bitsPerSample := #(8 8 8).
            ] ifFalse:[
                self error:'screen visual not supported'.
                ^ nil
            ]
        ]
    ].

    "
     dont know yet, how the display pads; assume worst case, 
     offering enough space for 32 bit padding
    "
    w := width := aRectangle width.
    h := height := aRectangle height.
    x := aRectangle left.
    y := aRectangle top.

    "
     a kludge: we dont know in advance how much we are going to need
     (its too late when info is present ...)
    "
    spaceBitsPerPixel := bitsPerPixel.
    (bitsPerPixel > 8) ifTrue:[
        spaceBitsPerPixel := 16.
        (bitsPerPixel > 16) ifTrue:[
            spaceBitsPerPixel := 32.
            (bitsPerPixel > 32) ifTrue:[
                spaceBitsPerPixel := bitsPerPixel.
            ]
        ]
    ].

    bytesPerLine := (w * spaceBitsPerPixel + 31) // 32 * 4.
    inData := ByteArray uninitializedNew:(bytesPerLine * height).

    "
     get the pixels
    "
    aDrawable isForm ifTrue:[
        info := device getBitsFromPixmapId:aDrawable id x:x y:y width:w height:h into:inData. 
    ] ifFalse:[
        info := device getBitsFromViewId:aDrawable id x:x y:y width:w height:h into:inData. 
    ].

    bitsPerPixelIn := info at:#bitsPerPixel.
    "
     Note: bit- and byte ordering doesn't matter for palette images!!
    "
    photometric ~~ #palette ifTrue:[
        "/
        "/ check if byteorder is what I like (msbFirst)
        "/

        "/ now done on-the-fly (see isMSB arg)

"/        byteOrder := info at:#byteOrder.
"/        byteOrder ~~ #msbFirst ifTrue:[
"/            bitsPerPixelIn == 16 ifTrue:[
"/                "/ must swap bytes
"/                inData swapBytes
"/            ] ifFalse:[
"/                bitsPerPixelIn == 32 ifTrue:[
"/                    "/ must swap longs
"/                    inData swapLongs
"/                ]
"/            ]
"/        ].

        "/
        "/ check if bitorder is what I like (msbFirst)
        "/
        "/ mhmh - thats not needed

"/        bitOrder := info at:#bitOrder.
"/        bitOrder ~~ #msbFirst ifTrue:[
"/            inData 
"/                expandPixels:8 
"/                width:(inData size)
"/                height:1 
"/                into:inData
"/                mapping:(ImageReader reverseBits).
"/        ].
    ].

    "/
    "/ check if bitorder is what I like (msbFirst)
    "/
    "/ mhmh - thats not needed

    bitsPerPixelIn < 8 ifTrue:[
        bitOrder := info at:#bitOrder.
        bitOrder ~~ #msbFirst ifTrue:[
            inData 
                expandPixels:8 
                width:(inData size)
                height:1 
                into:inData
                mapping:(ImageReader reverseBits "TODO: reverseBitsForDepth:bitsPerPixelIn").
        ].
    ].

    "
     check, if the devices padding is different ..
     or if the bitsPerPixels are different
    "
    bytesPerLineIn := (info at:#bytesPerLine).                    "what I got"
    bytesPerLine := (w * bitsPerPixel + 7) // 8.                  "what I want"

    ((bytesPerLine ~~ bytesPerLineIn) 
    or:[bitsPerPixelIn ~~ bitsPerPixel]) ifTrue:[
        tmpData := inData.
        inData := ByteArray uninitializedNew:(bytesPerLine * height).

        srcRow := 1.
        dstRow := 1.

        bitsPerPixelIn ~~ bitsPerPixel ifTrue:[
            "/ for now, only 32 -> 24 is supported

            maskR := (1 bitShift:device bitsRed) - 1.
            maskG := (1 bitShift:device bitsGreen) - 1.
            maskB := (1 bitShift:device bitsBlue) - 1.
            shR := device shiftRed negated.
            shG := device shiftGreen negated.
            shB := device shiftBlue negated.
            shR2 := (8 - device bitsRed).
            shG2 := (8 - device bitsGreen).
            shB2 := (8 - device bitsBlue).
            isMSB := (byteOrder == #msbFirst).

            ((bitsPerPixelIn == 32) and:[bitsPerPixel == 24]) ifTrue:[
                "/ 'reformatting 32->24...' printNL.
                1 to:h do:[:hi |
                    srcIndex := srcRow.
                    dstIndex := dstRow.

                    1 to:w do:[:wi |
                        lword := tmpData doubleWordAt:srcIndex MSB:isMSB.
                        r := (lword bitShift:shR) bitAnd:maskR.
                        g := (lword bitShift:shG) bitAnd:maskG.
                        b := (lword bitShift:shB) bitAnd:maskB.

                        inData at:dstIndex   put:r.
                        inData at:dstIndex+1 put:g.
                        inData at:dstIndex+2 put:b.
                        srcIndex := srcIndex + 4.
                        dstIndex := dstIndex + 3.
                    ].
                    dstRow := dstRow + bytesPerLine.
                    srcRow := srcRow + bytesPerLineIn
                ]
            ] ifFalse:[
                ((bitsPerPixelIn == 16) and:[bitsPerPixel == 24]) ifTrue:[
                    "/ 'reformatting 16->24...' printNL.
                    1 to:h do:[:hi |
                        srcIndex := srcRow.
                        dstIndex := dstRow.

                        1 to:w do:[:wi |
                            word := tmpData wordAt:srcIndex MSB:isMSB.
                            r := (word bitShift:shR) bitAnd:maskR.
                            g := (word bitShift:shG) bitAnd:maskG.
                            b := (word bitShift:shB) bitAnd:maskB.

                            inData at:dstIndex   put:(r bitShift:shR2).
                            inData at:dstIndex+1 put:(g bitShift:shG2).
                            inData at:dstIndex+2 put:(b bitShift:shB2).
"/ word print.
"/ ' -> ' print.
"/ (r bitShift:shR2) print.
"/ ' ' print. (g bitShift:shG2) print.
"/ ' ' print. (b bitShift:shB2) printCR.

                            srcIndex := srcIndex + 2.
                            dstIndex := dstIndex + 3.
                        ].
                        dstRow := dstRow + bytesPerLine.
                        srcRow := srcRow + bytesPerLineIn
                    ]
                ] ifFalse:[
                    ('Image [warning]: unsupported depth combination: ' , bitsPerPixelIn printString , ' -> ' ,
                                                        bitsPerPixel printString) errorPrintCR.
                    ^ nil
                ]
            ].
        ] ifFalse:[
            "
             repad in the buffer
            "
            1 to:h do:[:hi |
                inData replaceFrom:dstRow to:(dstRow + bytesPerLine - 1)
                              with:tmpData startingAt:srcRow.
                dstRow := dstRow + bytesPerLine.
                srcRow := srcRow + bytesPerLineIn
            ]
        ]
    ] ifFalse:[
        (bytesPerLine * height) ~~ inData size ifTrue:[
            tmpData := inData.
            inData := ByteArray uninitializedNew:(bytesPerLine * height).
            inData replaceFrom:1 to:bytesPerLine * height with:tmpData startingAt:1
        ]
    ].
    bytes := inData.

    "info printNL."

    "
      if not #palette we are done, the pixel values are the rgb/grey values
    "
    photometric == #palette ifTrue:[
        "
         what we have now are the color numbers - still need the r/g/b values.
         find out, which colors are in the picture
        "
        usedPixels := inData usedValues.
        mapSize := usedPixels max + 1.

        "get the palette"
        map := Array new:mapSize.
        usedPixels do:[:colorIndex |
            |i|

            i := colorIndex + 1.
            device getRGBFrom:colorIndex into:[:r :g :b |
                map at:i put:(Color red:r green:g blue:b)
            ]
        ].
        colorMap := map.
    ].

    "Modified: / 9.1.1998 / 21:32:36 / stefan"
    "Modified: / 7.2.1998 / 18:23:07 / cg"
!

fromScreen:aRectangle
    "read an image from the display screen.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    ^ self 
        fromScreen:aRectangle 
        on:Screen current
        grab:true

    "Modified: 26.3.1997 / 10:43:34 / cg"
!

fromScreen:aRectangle on:aDevice
    "read an image from aDevices display screen.
     Since I have no other displays, only the MonoChrome, StaticGrey 
     and PseudoColor cases have been tested ... 
     (especially True- and DirectColor may be wrong).
     Late note: 24bit rgb now also works.
     WARNING: this temporarily grabs the display
              it may not work from within a buttonMotion
              (use #fromScreen:on:grab: with a false grabArg then)."

    ^ self
        fromScreen:aRectangle 
        on:aDevice 
        grab:true

    "
     Image fromScreen:((0 @ 0) corner:(100 @ 100)) on:Display
     Image fromScreen:((0 @ 0) corner:(500 @ 500)) on:Display
    "

    "Modified: 24.4.1997 / 18:25:13 / cg"
!

fromScreen:aRectangle on:aDevice grab:doGrab
    "read an image from aDevices display screen.
     If the doGrab argument is true, the display
     is grabbed (i.e. blocked for others) and a camera cursor is
     shown while the readout is done.
     Since I have no other displays, only the MonoChrome, StaticGrey 
     and PseudoColor cases have been tested ... 
     (especially True- and DirectColor may be wrong).
     Late note: 24bit rgb now also works.
     WARNING: with doGrab true, this temporarily grabs the display
              and it may not work from within a buttonMotion
              (use with a false grabArg then)."

    |curs cid rootView|

    doGrab ifTrue:[
        curs := Cursor sourceForm:(Image fromFile:'bitmaps/Camera.xbm')
                         maskForm:(Image fromFile:'bitmaps/Camera_m.xbm')
                          hotSpot:16@16.
        curs notNil ifTrue:[
            cid := (curs onDevice:aDevice) id
        ].
    ].

    "
     actually have to grabServer ... but thats not yet available
    "
    rootView := aDevice rootView.
    doGrab ifTrue:[
        aDevice setActivePointerGrab:rootView.
        aDevice grabPointerIn:rootView id
                   withCursor:cid 
                   pointerMode:#async 
                   keyboardMode:#sync 
                   confineTo:nil.
    ].

    "
     get the pixels
    "
    [
        self from:rootView in:aRectangle.
    ] valueNowOrOnUnwindDo:[
        doGrab ifTrue:[
            aDevice ungrabPointer.
        ]
    ]

    "Created: 26.3.1997 / 10:36:30 / cg"
    "Modified: 20.5.1997 / 16:56:39 / cg"
!

photometricFromScreen:aDevice
    "read aDevices display photometric and set my colormap for it.
     This must be used after an images bits have been read from the screen
     or from an offScreen bitmap, for correct pixel interpretation."

    |depth visType bitsPerPixel|

    "
     get some attributes of the display device
    "
    visType := aDevice visualType.
    depth := aDevice depth.

    "/ kludge for 15bit XFree server
    depth == 15 ifTrue:[
        depth := 16
    ].

    (visType == #StaticGray) ifTrue:[
        (aDevice blackpixel == 0) ifTrue:[
            photometric := #blackIs0
        ] ifFalse:[
            photometric := #whiteIs0
        ].
        samplesPerPixel := 1.
        bitsPerPixel := depth.
        bitsPerSample := Array with:bitsPerPixel.
        "
         were done, the pixel values are the grey values
        "
        ^ self
    ].

    ((visType == #TrueColor) or:[visType == #DirectColor]) ifTrue:[
        photometric := #rgb.
        samplesPerPixel := 3.

        "/ for now - only support 24bit TrueColor
        depth ~~ 24 ifTrue:[
            'IMAGE: unsupported display depth' errorPrintCR.
        ].
"/                bitsPerPixel := depth.
"/                bitsPerSample := Array with:aDevice bitsRed
"/                                       with:aDevice bitsGreen
"/                                       with:aDevice bitsBlue
        bitsPerPixel := 24.
        bitsPerSample := #(8 8 8).
        "
         were done, the pixel values are the rgb values
        "
        ^ self
    ].

    ((visType ~~ #PseudoColor) 
    and:[(visType ~~ #StaticColor) 
    and:[visType ~~ #GrayScale]]) ifTrue:[
        self error:'screen visual not supported'.
        ^ nil
    ].

    photometric := #palette.
    samplesPerPixel := 1.
    bitsPerPixel := depth.
    bitsPerSample := Array with:bitsPerPixel.

    "
     still need the pixels r/g/b values ...
    "
    colorMap := aDevice colorMap

    "Modified: 11.7.1996 / 11:11:34 / cg"
! !

!Image class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libview/Image.st,v 1.245 1998-04-27 15:22:39 cg Exp $'
! !
Image initialize!
